%{
/* This file is part of mailfrom filter.
   Copyright (C) 2005, 2006 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <fnmatch.h>
#include <netdb.h>
#include "mailfrom.h"

NODE *alloc_node(enum node_type type, const struct locus *locus);
int string_to_stat(char *, mf_status *);
void set_poll_arg(struct poll_data *poll, int kw, NODE *expr);
static int time_multiplier(const char *, unsigned *, unsigned *);
static int codegen(prog_counter_t *pc, NODE *node);
NODE *reverse(NODE *in);
 
static NODE *root_node[smtp_state_count];
prog_counter_t entry_point[smtp_state_count];
 
int regex_flags;        /* Should default to REG_NOSUB ? */
unsigned error_count;   /* Number of detected errors */
size_t variable_count = 1; /* Offset of the next-to-be-defined variable
			      in the data segment */
static enum smtp_state state_tag; /* State tag of the currently processed
				     PROG */

static struct function *func;   /* The currently compiled function */
static prog_counter_t jump_pc;  /* Pointer to the chain of jmp instructions */

/* State handlers and their positional parameters */ 
struct state_parms {
	int cnt;                  /* Number or positional parameters */
	instr_rettype types[2];   /* Their data types */
} state_parms[] = {
	{ 0,  }, /* smtp_state_none */
	{ 0,  }, /* smtp_state_connect */
	{ 1, { rettype_string } },  /* smtp_state_helo */
	{ 1, { rettype_string } },  /* smtp_state_envfrom */
	{ 1, { rettype_string } },  /* smtp_state_envrcpt */
	{ 2, { rettype_string, rettype_string } }, /* smtp_state_header */
	{ 0,  }, /* smtp_state_eoh */
	{ 2, { rettype_string, rettype_number } }, /* smtp_state_body */
	{ 0, }, /* smtp_state_eom */
};

struct parmtype {
	struct parmtype *next;
	instr_rettype type;
};

#define	synt_none        0
#define	synt_handler     1
#define	synt_catch       2
#define	synt_function    3

static int stie_in;
static int allow_unquoted_strings;
 
static int
parmcount_none()
{
	return 0;
}

static instr_rettype
parmtype_none(int n)
{
	return rettype_unspecified;
}
 
static int
parmcount_handler()
{
	return state_parms[state_tag].cnt;
}

static instr_rettype
parmtype_handler(int n)
{
	return state_parms[state_tag].types[n-1];
}

static int
parmcount_catch()
{
	return 1;
}

static instr_rettype
parmtype_catch(int n)
{
	return rettype_number;
}

static int
parmcount_function()
{
	return func->parmcount;
}

static instr_rettype
parmtype_function(int n)
{
	return func->parmtype[n-1];
}

struct parminfo {
	int (*parmcount)(void);
	instr_rettype (*parmtype)(int n);
} parminfo[] = {
	{ parmcount_none, parmtype_none },
	{ parmcount_handler, parmtype_handler },
	{ parmcount_catch, parmtype_catch },
	{ parmcount_function, parmtype_function }
};

#define PARMCOUNT() parminfo[stie_in].parmcount()
#define PARMTYPE(n) parminfo[stie_in].parmtype(n)

instr_rettype
string_to_type(char *s)
{
	if (strcmp(s, "n") == 0)
		return rettype_number;
	else if (strcmp(s, "s") == 0)
		return rettype_string;
	else
		return rettype_unspecified;
}

const char *
type_to_string(instr_rettype t)
{
	switch (t) {
	case rettype_number:
		return "n";
	case rettype_string:
		return "s";
	default:
		return "?";
	}
}

static void jump_fixup(prog_counter_t pos, prog_counter_t endpos);
static NODE *create_on_node(NODE *sel, struct case_stmt *cases);
%}

%error-verbose

%union {
	char *string;
	struct {
		NODE *head;
 		NODE *tail;
	} stmtlist;
	NODE *node;
	struct return_node ret;
	struct poll_data poll;
	struct locus locus;
	struct pollarg {
		int kw;
		NODE *expr;
	} pollarg;
	struct {
		NODE *head;
		NODE *tail;
		size_t count;
	} arglist;
	long number;
	const struct builtin *builtin;
	struct variable *var;
	enum smtp_state state;
	struct {
		struct locus locus;
		int qualifier;
	} matchtype;
	instr_rettype type;
	struct typelist {
		struct parmtype *head, *tail;
		int count;
	} typelist;
	int tie_in;
	struct function *function;
	struct {
		struct valist *head, *tail;
	} valist_list;
	struct valist *valist;
	struct {
		struct case_stmt *head, *tail;
	} case_list ;
	struct case_stmt *case_stmt;
};

%token <locus> ACT_ACCEPT ACT_REJECT ACT_TEMPFAIL ACT_CONTINUE ACT_DISCARD
%token <locus> ADD REPLACE DELETE
%token <locus> PROG IF FI ELSE ELIF 
%token <locus> ON HOST FOR FROM AS DO DONE POLL MATCHES FNMATCHES
%token <locus> MXMATCHES MXFNMATCHES
%token <locus> WHEN RATE NEXT SET CATCH KW_ECHO RETURNS RETURN FUNC
%token <locus> SWITCH CASE DEFAULT
%token <string> STRING CODE XCODE
%token <string> SYMBOL VARIABLE IDENTIFIER
%token <number> ARG NUMBER
%token <builtin> BUILTIN
%token <locus> OR AND EQ NE LT LE GT GE NOT  
%token <function> FUNCTION

%left OR
%left AND
%left EQ NE MATCHES FNMATCHES
%left LT LE GT GE
%left '+' '-'
%left '*' '/' 
%left NOT UMINUS BUILTIN

%type <node> stmt condition action if_cond else_cond on_cond cond atom
             funcall expr asgn catch return case_cond
%type <stmtlist> stmtlist
%type <ret> triplet maybe_triplet
%type <poll> pollstmt pollarglist
%type <pollarg> pollarg
%type <number> number
%type <arglist> arglist 
%type <var> variable
%type <string> string
%type <state> state_ident
%type <matchtype> matches fnmatches
%type <type> retdecl
%type <typelist> typelist typedecl
%type <function> fundecl
%type <valist> value
%type <valist_list> valist catchlist
%type <case_list> cond_branches branches 
%type <case_stmt> cond_branch branch
%type <locus> on 

%expect 70

%%

input      : decllist
           ;

decllist   : decl
           | decllist decl
           ;

decl       : PROG state_ident DO stmtlist DONE
             {
		     root_node[$2] = $4.head;
		     if (codegen(&entry_point[$2], $4.head))
			     error_count++;
		     else
			     milter_enable_state($2);
		     stie_in = synt_none;
	     }
           | FUNC fundecl DO stmtlist DONE
             {
		     prog_counter_t pc;

		     func->node = $4.head;
		     /* Assign the entry point early to properly
			handle recursive functions */
		     func->entry = code_get_counter();

		     /* Entry code */
		     if (func->exmask) {
			     code_instr(instr_saveex);
			     code_immediate((void*)func->exmask);
		     }

		     if (codegen(&pc, $4.head))
			     error_count++;
		     else {
			     jump_fixup(jump_pc, code_get_counter());

			     /* Exit code */
			     if (func->exmask) 
				     code_instr(instr_restex);
			     code_instr(instr_return);
		     }
		     
		     stie_in = synt_none;
		     func = NULL;
	     }
           | SET IDENTIFIER string
             {
		     struct variable *var = variable_install($2);
		     if (var->off == 0)
			     var->off = variable_count++;
		     var->type = rettype_string;
		     declare_string_variable(var, $3);
	     }
	   | SET IDENTIFIER number
             {
		     struct variable *var = variable_install($2);
		     if (var->off == 0)
			     var->off = variable_count++;
		     var->type = rettype_number;
		     declare_numeric_variable(var, $3);
	     }
           ;

fundecl    : IDENTIFIER '(' typedecl ')' retdecl
 	     {
		     instr_rettype *ptypes = NULL;
		     
		     if ($3.count) {
			     int i;
			     struct parmtype *p;
			     
			     ptypes = emalloc($3.count*sizeof *ptypes);
			     for (i = 0, p = $3.head; p; i++) {
				     struct parmtype *next = p->next;
				     ptypes[i] = p->type;
				     free(p);
				     p = next;
			     }
		     }

		     $$ = func = function_install($1, $3.count, ptypes, $5,
						  get_locus());
		     stie_in = synt_function;
	     }

typedecl   : /* empty */
             {
		     $$.count = 0;
	     }
           | typelist
           ;

typelist   : IDENTIFIER
             {
		     $$.head = emalloc(sizeof *$$.head);
		     $$.head->next = NULL;
		     if (($$.head->type = string_to_type($1))
			 == rettype_unspecified) {
			     parse_error("unknown type specification: %s", $1);
			     error_count++;
		     }
		     $$.tail = $$.head;
		     $$.count = 1;
	     }
           | typelist ',' IDENTIFIER
             {
		     struct parmtype *p = emalloc(sizeof *p);
		     p->next = NULL;
		     if ((p->type = string_to_type($3))
			 == rettype_unspecified) {
			     parse_error("unknown type specification: %s", $3);
			     error_count++;
		     }
		     $$.tail->next = p;
		     $$.tail = p;
		     $$.count++;
	     }
           ;

retdecl    : /* empty */
             {
		     $$ = rettype_unspecified;
	     }
           | RETURNS IDENTIFIER
             {
		     if (($$ = string_to_type($2)) == rettype_unspecified) {
			     parse_error("unknown type specification: %s", $2);
			     error_count++;
		     }
	     }
           ;

state_ident: IDENTIFIER
             {
		     $$ = string_to_state($1);
		     if ($$ == smtp_state_none) {
			     parse_error("Unknown smtp state tag: %s", $1);
			     error_count++;
		     }
		     state_tag = $$;
		     stie_in = synt_handler;
	     }
           ;

stmtlist   : stmt
             {
		     $1->next = NULL;
		     $$.head = $$.tail = $1;
	     }
           | stmtlist stmt
             {
		     $1.tail->next = $2;
		     $1.tail = $2;
		     $$ = $1;
	     }
           ;

stmt       : condition
           | action
           | asgn
           | catch
           | return
           ;

asgn       : SET IDENTIFIER expr
             {
		     struct variable *var;
		     $$ = alloc_node(node_type_asgn, &$1);
		     var = variable_install($2);
		     if (var->off == 0)
			     var->off = variable_count++;
		     $$->v.asgn.var = var;
		     $$->v.asgn.node = $3;
	     }
           ;

action     : ACT_ACCEPT maybe_triplet
             {
		     if ($2.code || $2.xcode || $2.message)
			     parse_error("warning: arguments are ignored for accept");
		     $$ = alloc_node(node_type_result, &$1);
		     $$->v.ret = $2;
		     $$->v.ret.stat = SMFIS_ACCEPT;
	     }
           | ACT_REJECT maybe_triplet
             {
		     if  ($2.code && $2.code[0] != '5')
			     parse_error("reject code should be 5xx");
		     if  ($2.xcode && $2.xcode[0] != '5')
			     parse_error("reject extended code should be 5.x.x");
		     $$ = alloc_node(node_type_result, &$1);
		     $$->v.ret = $2;
		     $$->v.ret.stat = SMFIS_REJECT;
	     }
           | ACT_TEMPFAIL maybe_triplet
             {
		     if  ($2.code && $2.code[0] != '4')
			     parse_error("tempfail code should be 4xx");
		     if  ($2.xcode && $2.xcode[0] != '4')
			     parse_error("tempfail extended code should be 4.x.x");
		     $$ = alloc_node(node_type_result, &$1);
		     $$->v.ret = $2;
		     $$->v.ret.stat = SMFIS_TEMPFAIL;
	     }
           | ACT_CONTINUE
             {
		     $$ = alloc_node(node_type_result, &$1);
		     memset(&$$->v.ret, 0, sizeof $$->v.ret);
		     $$->v.ret.stat = SMFIS_CONTINUE;
	     }
           | ACT_DISCARD
             {
		     $$ = alloc_node(node_type_result, &$1);
		     memset(&$$->v.ret, 0, sizeof $$->v.ret);
		     $$->v.ret.stat = SMFIS_DISCARD;
	     }
           | NEXT
             {
		     $$ = alloc_node(node_type_next, &$1);
	     }
           | ADD string string
             {
		     $$ = alloc_node(node_type_header, &$1);
		     $$->v.hdr.opcode = header_add;
		     $$->v.hdr.name = $2;
		     $$->v.hdr.value = $3;
	     }
           | REPLACE string string
             {
		     $$ = alloc_node(node_type_header, &$1);
		     $$->v.hdr.opcode = header_replace;
		     $$->v.hdr.name = $2;
		     $$->v.hdr.value = $3;
	     }
           | DELETE string
             {
		     $$ = alloc_node(node_type_header, &$1);
		     $$->v.hdr.opcode = header_delete;
		     $$->v.hdr.name = $2;
	     }
           | KW_ECHO expr
             {
		     $$ = alloc_node(node_type_echo, &$1);
		     $$->v.node = $2;
	     }
           ;

maybe_triplet: /* empty */
             {
		     memset(&$$, 0, sizeof $$);
	     }
           | triplet
	   ;

triplet    : CODE
             {
		     $$.code = $1;
		     $$.xcode = NULL;
		     $$.message = NULL;
	     }
           | CODE XCODE
             {
		     $$.code = $1;
		     $$.xcode = $2;
		     $$.message = NULL;
	     }
           | CODE XCODE expr
             {
		     $$.code = $1;
		     $$.xcode = $2;
		     $$.message = $3;
	     }
           | CODE expr
             {
		     $$.code = $1;
		     $$.xcode = NULL;
		     $$.message = $2;
	     }
           ;

condition  : if_cond
           | case_cond 
           | on_cond
           ;

if_cond    : IF cond stmtlist else_cond FI
             {
		     $$ = alloc_node(node_type_if, &$1);
		     $$->v.cond.cond = $2;
		     $$->v.cond.if_true = $3.head;
		     $$->v.cond.if_false = $4;
	     }
           ;

else_cond  : /* empty */
             {
		     $$ = NULL;
	     }
           | ELIF cond stmtlist else_cond
             {
		     $$ = alloc_node(node_type_if, &$1);
		     $$->v.cond.cond = $2;
		     $$->v.cond.if_true = $3.head;
		     $$->v.cond.if_false = $4;
	     }
           | ELSE stmtlist
             {
		     $$ = $2.head;
	     }
           ;

case_cond  : SWITCH expr DO cond_branches DONE
             {
		     struct case_stmt *defcase = NULL, *pcase, *prev;
		     
		     $$ = alloc_node(node_type_switch, &$1);
		     $$->v.switch_stmt.node = $2;
		     
		     /* Make sure there is only one default case and
			place it at the beginning of the list */
		     pcase = $4.head;
		     if (!pcase->valist) {
			     defcase = pcase;
			     $4.head = $4.head->next;
		     }
		     prev = pcase;
		     pcase = pcase->next;
		     while (pcase) {
			     if (!pcase->valist) {
				     if (defcase) {
					     parse_error_locus(&pcase->locus,
					       "duplicate default statement");
					     parse_error_locus(&defcase->locus,
					       "previously defined here");
					     YYERROR;
				     }
				     defcase = pcase;
				     prev->next = pcase->next;
			     } else
				     prev = pcase;
			     pcase = pcase->next;
		     }

		     if (defcase) {
			     defcase->next = $4.head;
			     $4.head = defcase;
		     }

		     $$->v.switch_stmt.cases = $4.head;
	     }
           ;

cond_branches: cond_branch
               {
		       $$.head = $$.tail = $1;
	       }
           | cond_branches cond_branch
               {
		       $$.tail->next = $2;
		       $$.tail = $2;
	       }
           ;

cond_branch: CASE valist ':' stmtlist 
             {
		     $$ = emalloc(sizeof *$$);
		     $$->next = NULL;
		     $$->locus = $1;
		     $$->valist = $2.head;
		     $$->node = $4.head;
	     }
           | DEFAULT ':' stmtlist
             {
		     $$ = emalloc(sizeof *$$);
		     $$->next = NULL;
		     $$->locus = $1;
		     $$->valist = NULL;
		     $$->node = $3.head;
	     }		     
           ;

valist     : value
             {
		     $$.head = $$.tail = $1;
	     }
           | valist OR value
             {
		     $1.tail->next = $3;
		     $1.tail = $3;
		     $$ = $1;
	     }
           ;

value      : string
             {
		     $$ = emalloc(sizeof *$$);
		     $$->next = NULL;
		     $$->value.type = rettype_string;
		     $$->value.v.string = $1;
	     }
           | number
             {
		     $$ = emalloc(sizeof *$$);
		     $$->next = NULL;
		     $$->value.type = rettype_number;
		     $$->value.v.number = $1;
	     }
           ;

cond       : expr
           | expr EQ expr
             {
		     $$ = alloc_node(node_type_bin, &$2);
		     $$->v.bin.opcode = bin_eq;
		     $$->v.bin.arg[0] = $1;
		     $$->v.bin.arg[1] = $3;
	     }
           | expr NE expr
             {
		     $$ = alloc_node(node_type_bin, &$2);
		     $$->v.bin.opcode = bin_ne;
		     $$->v.bin.arg[0] = $1;
		     $$->v.bin.arg[1] = $3;
	     }
           | expr LT expr
             {
		     $$ = alloc_node(node_type_bin, &$2);
		     $$->v.bin.opcode = bin_lt;
		     $$->v.bin.arg[0] = $1;
		     $$->v.bin.arg[1] = $3;
	     }
	   | expr LE expr
             {
		     $$ = alloc_node(node_type_bin, &$2);
		     $$->v.bin.opcode = bin_le;
		     $$->v.bin.arg[0] = $1;
		     $$->v.bin.arg[1] = $3;
	     }
	   | expr GT expr
             {
		     $$ = alloc_node(node_type_bin, &$2);
		     $$->v.bin.opcode = bin_gt;
		     $$->v.bin.arg[0] = $1;
		     $$->v.bin.arg[1] = $3;
	     }
	   | expr GE expr
             {
		     $$ = alloc_node(node_type_bin, &$2);
		     $$->v.bin.opcode = bin_ge;
		     $$->v.bin.arg[0] = $1;
		     $$->v.bin.arg[1] = $3;
	     }
           | expr matches expr 
             {
		     $$ = alloc_node(node_type_bin, &$2.locus);
		     $$->v.bin.opcode = bin_match;
		     $$->v.bin.qualifier = $2.qualifier;
		     $$->v.bin.arg[0] = $1;
		     $$->v.bin.arg[1] = $3;
	     }
           | expr fnmatches expr
             {
		     $$ = alloc_node(node_type_bin, &$2.locus);
		     $$->v.bin.opcode = bin_fnmatch;
		     $$->v.bin.qualifier = $2.qualifier;
		     $$->v.bin.arg[0] = $1;
		     $$->v.bin.arg[1] = $3;
	     }
           | cond OR cond
             {
		     $$ = alloc_node(node_type_bin, &$2);
		     $$->v.bin.opcode = bin_or;
		     $$->v.bin.arg[0] = $1;
		     $$->v.bin.arg[1] = $3;
	     }
           | cond AND cond
             {
		     $$ = alloc_node(node_type_bin, &$2);
		     $$->v.bin.opcode = bin_and;
		     $$->v.bin.arg[0] = $1;
		     $$->v.bin.arg[1] = $3;
	     }
           | NOT cond
             {
		     $$ = alloc_node(node_type_un, &$1);
		     $$->v.un.opcode = unary_not;
		     $$->v.un.arg = $2;
	     }
           | '(' cond ')'
             {
		     $$ = $2;
	     }
           | RATE error
             {
		     parse_error("`rate' statement is no longer supported, use function instead;");
		     parse_error("refer to FIXME-xref for the detailed discussion");
	     }
           ;

matches    : MATCHES
             {
		     $$.locus = $1;
		     $$.qualifier = 0;
	     }
           | MXMATCHES
             {
		     $$.locus = $1;
		     $$.qualifier = QUALIFIER_MX;
	     }
           ;

fnmatches  : FNMATCHES
             {
		     $$.locus = $1;
		     $$.qualifier = 0;
	     }
           | MXFNMATCHES
             {
		     $$.locus = $1;
		     $$.qualifier = QUALIFIER_MX;
	     }
           ;

number     : NUMBER
           | CODE
             {
		     char *p;
		     $$ = strtol($1, &p, 10);
		     if (*p) {
			     /* should not happen */
			     parse_error("Invalid number (near `%s')", p);
			     YYERROR;
		     }
	     }
           ; 

expr       : atom
           | expr atom  
             {
		     $$ = alloc_node(node_type_concat, get_locus());
		     $$->v.concat.arg[0] = $1;
		     $$->v.concat.arg[1] = $2;
	     }
           | expr '+' expr
             {
		     $$ = alloc_node(node_type_bin, get_locus());
		     $$->v.bin.opcode = bin_add;
		     $$->v.bin.arg[0] = $1;
		     $$->v.bin.arg[1] = $3;
	     }		     
	   | expr '-' expr
             {
		     $$ = alloc_node(node_type_bin, get_locus());
		     $$->v.bin.opcode = bin_sub;
		     $$->v.bin.arg[0] = $1;
		     $$->v.bin.arg[1] = $3;
	     }		     
	   | expr '*' expr
             {
		     $$ = alloc_node(node_type_bin, get_locus());
		     $$->v.bin.opcode = bin_mul;
		     $$->v.bin.arg[0] = $1;
		     $$->v.bin.arg[1] = $3;
	     }		     
	   | expr '/' expr
             {
		     $$ = alloc_node(node_type_bin, get_locus());
		     $$->v.bin.opcode = bin_div;
		     $$->v.bin.arg[0] = $1;
		     $$->v.bin.arg[1] = $3;
	     }		     
	   | '-' expr %prec UMINUS
             {
		     $$ = alloc_node(node_type_un, get_locus());
		     $$->v.un.opcode = unary_minus;
		     $$->v.un.arg = $2;
	     }
           | '(' expr ')'
             {
		     $$ = $2;
	     }
           ;
 
atom       : string
             {
		     $$ = alloc_node(node_type_string, get_locus());
		     $$->v.string = $1;
	     }
           | SYMBOL
             {
		     $$ = alloc_node(node_type_symbol, get_locus());
		     $$->v.string = $1;
	     }
           | number
             {
		     $$ = alloc_node(node_type_number, get_locus());
		     $$->v.number = $1;
	     }
           | variable
             {
		     $$ = alloc_node(node_type_variable, get_locus());
		     $$->v.variable = $1;
	     }
           | ARG
             {
		     if ($1 > PARMCOUNT()) { 
			     parse_error("Argument number too high");
			     error_count++;
		     }
		     $$ = alloc_node(node_type_arg, get_locus());
		     $$->v.number = $1;
	     }
           | funcall
	   ;

funcall    : BUILTIN '(' arglist ')'
             {
		     if ($3.count < $1->parmcount) {
			     parse_error("Too few arguments in call to `%s'",
					 $1->name);
			     YYERROR;
		     } else if ($3.count > $1->parmcount) {
			     parse_error("Too many arguments in call to `%s'",
					 $1->name);
			     YYERROR;
		     } else {
			     $$ = alloc_node(node_type_builtin, get_locus());
			     $$->v.builtin.builtin = $1;
			     $$->v.builtin.args = $3.head;
		     }
	     }
	   | BUILTIN '(' ')'
             {
		     if ($1->parmcount) {
			     parse_error("Too few arguments in call to `%s'",
					 $1->name);
			     YYERROR;
		     } else {
 			     $$ = alloc_node(node_type_builtin, get_locus());
			     $$->v.builtin.builtin = $1;
			     $$->v.builtin.args = NULL;
		     }
	     }
	   | BUILTIN expr
             {
		     if ($1->parmcount > 1) {
			     parse_error("Too few arguments in call to `%s'",
					 $1->name);
			     YYERROR;
		     } else if ($1->parmcount == 0) {
			     parse_error("Too many arguments in call to `%s'",
					 $1->name);
			     YYERROR;
		     } else {
			     $$ = alloc_node(node_type_builtin, get_locus());
			     $$->v.builtin.builtin = $1;
			     $$->v.builtin.args = $2;
		     }
	     }
           | RATE '(' arglist ')'
             {
		     $$ = alloc_node(node_type_builtin, get_locus());
		     $$->v.builtin.builtin = builtin_lookup("rate");
		     if (!$$->v.builtin.builtin) {
			     parse_error("Function `%s' is not defined",
					 "rate");
			     YYERROR;
		     }
		     if ($3.count < $$->v.builtin.builtin->parmcount) {
			     parse_error("Too few arguments in call to `%s'",
					 "rate");
			     YYERROR;
		     } else if ($3.count > $$->v.builtin.builtin->parmcount) {
			     parse_error("Too many arguments in call to `%s'",
					 "rate");
			     YYERROR;
		     }
		     $$->v.builtin.args = $3.head;
	     }
           | FUNCTION '(' arglist ')'
             {
		     if ($3.count < $1->parmcount) {
			     parse_error("Too few arguments in call to `%s'",
					 $1->name);
			     YYERROR;
		     } else if ($3.count > $1->parmcount) {
			     parse_error("Too many arguments in call to `%s'",
					 $1->name);
			     YYERROR;
		     } else {
			     $$ = alloc_node(node_type_call, get_locus());
			     $$->v.call.func = $1;
			     $$->v.call.args = reverse($3.head);
		     }
	     }
           | FUNCTION '(' ')'
             {
		     if ($1->parmcount) {
			     parse_error("Too few arguments in call to `%s'",
					 $1->name);
			     YYERROR;
		     } else {
 			     $$ = alloc_node(node_type_call, get_locus());
			     $$->v.call.func = $1;
			     $$->v.call.args = NULL;
		     }
	     }
           | FUNCTION expr
             {
		     if ($1->parmcount > 1) {
			     parse_error("Too few arguments in call to `%s'",
					 $1->name);
			     YYERROR;
		     } else if ($1->parmcount == 0) {
			     parse_error("Too many arguments in call to `%s'",
					 $1->name);
			     YYERROR;
		     } else {
			     $$ = alloc_node(node_type_call, get_locus());
			     $$->v.call.func = $1;
			     $$->v.call.args = $2;
		     }
	     }
           ;

string     : STRING
           | IDENTIFIER
             {
		     if (!allow_unquoted_strings)
			     parse_error("warning: unquoted identifier `%s'",
					 $1);
	     }
           ;

arglist    : expr
             {
		     $1->next = NULL;
		     $$.head = $$.tail = $1;
		     $$.count = 1;
	     }
           | arglist ',' expr
             {
		     $1.tail->next = $3;
		     $1.tail = $3;
		     $1.count++;
		     $$ = $1;
	     }
           ;

variable   : VARIABLE
             {
		     $$ = variable_lookup($1);
		     if (!$$) {
			     parse_error("Variable %s is not defined", $1);
			     YYERROR;
		     }
	     }

catch       : CATCH catchlist DO
                 { $<tie_in>$ = stie_in; stie_in = synt_catch; }
                 stmtlist DONE
              {
		      int i;
		      struct valist *p;

		      stie_in = $<tie_in>4;
		      $$ = alloc_node(node_type_catch, &$1);
		      for (i = 0, p = $2.head; p; p = p->next)
			      i++;
		      $$->v.catch.count = i;
		      $$->v.catch.exmask = 0;
		      for (i = 0, p = $2.head; p; p = p->next, i++) {
			      if (p->value.type != rettype_number) {
				      parse_error_locus(&$1,
							"expected numeric value, but found `%s'",
							p->value.v.string);
				      error_count++;
				      continue;
			      }
			      $$->v.catch.exmask |= EXMASK(p->value.v.number);
			      if (stie_in == synt_function)
				      func->exmask |= EXMASK(p->value.v.number);
		      }
		      $$->v.catch.node = $5.head;
	      }
            ;

catchlist   : '*'
              {
		      int i;

		      $$.head = $$.tail = NULL;
		      for (i = 0; i < mf_status_count; i++) {
			      struct valist *p = emalloc(sizeof *p);
			      p->next = NULL; 
			      p->value.type = rettype_number;
			      p->value.v.number = i;
			      if (!$$.head)
				      $$.head = $$.tail = p;
			      else {
				      $$.tail->next = p;
				      $$.tail = p;
			      }
		      }
	      }
            | valist
            ; 

return      : RETURN
              {
		      $$ = alloc_node(node_type_return, &$1);
		      $$->v.node = NULL;
	      }
            | RETURN expr
              {
		      $$ = alloc_node(node_type_return, &$1);
		      $$->v.node = $2;
	      }
            ;

/* *************************** */
/*         ON statement        */
/* *************************** */

on_cond    : on pollstmt do branches DONE
             {
		     NODE *sel, *np;
		     NODE *head = NULL, *tail;
		     const struct builtin *bi;
		     
		     /* Build argument list */
		     head = tail = $2.email;

		     if ($2.ehlo)
			     np = $2.ehlo;
		     else {
			     np = alloc_node(node_type_variable, get_locus());
			     np->v.variable = variable_lookup("ehlo_domain");
		     }
		     tail->next = np;
		     tail = np;

		     if ($2.mailfrom)
			     np = $2.mailfrom;
		     else {
			     np = alloc_node(node_type_variable, get_locus());
			     np->v.variable =
				     variable_lookup("mailfrom_address");
		     }
		     tail->next = np;
		     tail = np;

		     if ($2.client_addr) {
			     np = $2.client_addr;
			     tail->next = np;
			     tail = np;
			     bi = builtin_lookup("strictpoll");
		     } else
			     bi = builtin_lookup("stdpoll");
		     
		     sel = alloc_node(node_type_builtin, &$1);
		     sel->v.builtin.builtin = bi;
		     sel->v.builtin.args = head;

		     $$ = create_on_node(sel, $4.head);
	     }
           | on funcall do branches DONE
             {
		     $$ = create_on_node($2, $4.head);
	     }
           ;

on         : ON
             {
		     onblock(1);
	     }
           ;

do         : DO
             {
		     onblock(0);
	     }
           ;

pollstmt   : POLL expr
             {
		     struct pollarg arg;
		     
		     arg.kw = FOR;
		     arg.expr = $2;
		     memset(&$$, 0, sizeof $$);
		     set_poll_arg(&$$, arg.kw, arg.expr);
	     }
           | POLL expr pollarglist
             {
		     struct pollarg arg;
		     
		     arg.kw = FOR;
		     arg.expr = $2;
		     set_poll_arg(&$3, arg.kw, arg.expr);
		     $$ = $3;
	     }
           | POLL pollarglist
             {
		     $$ = $2;
	     }
           ;

pollarglist: pollarg
             {
		     memset(&$$, 0, sizeof $$);
		     set_poll_arg(&$$, $1.kw, $1.expr);
	     }
           | pollarglist pollarg
             {
		     set_poll_arg(&$1, $2.kw, $2.expr);
		     $$ = $1;
	     }
           ;

pollarg    : FOR expr
             {
		     $$.kw = FOR;
		     $$.expr = $2;
	     }
           | HOST expr
             {
		     $$.kw = HOST;
		     $$.expr = $2;
	     }
           | AS expr
             {
		     $$.kw = AS;
		     $$.expr = $2;
	     }
           | FROM expr
             {
		     $$.kw = FROM;
		     $$.expr = $2;
	     }
           ;


branches   : branch
             {
		     $$.head = $$.tail = $1;
	     }
           | branches branch
             {
		     $1.tail->next = $2;
		     $1.tail = $2;
		     $$ = $1;
	     }
           ;

branch     : WHEN { allow_unquoted_strings = 1; } valist ':'
                  { allow_unquoted_strings = 0; } stmtlist
             {
		     struct valist *p;

		     for (p = $3.head; p; p = p->next) {
			     if (p->value.type == rettype_string) {
				     mf_status status;
				     if (string_to_stat(p->value.v.string,
							&status)) {
				        parse_error("Unknown return code: %s",
						    p->value.v.string);
					YYERROR;
				     }
				     p->value.type = rettype_number;
				     p->value.v.number = status;
			     }
		     }
		     
		     $$ = emalloc(sizeof *$$);
		     $$->next = NULL;
		     $$->locus = $1;
		     $$->valist = $3.head;
		     $$->node = $6.head;
	     }
           ;

%%

int
yyerror(char *s)
{
	parse_error("%s", s);
	error_count++;
}


int
parse_config(char *name, int ydebug, int ldebug)
{
	yydebug = ydebug;
	if (source(name, ldebug))
		return -1;
	stie_in = synt_none;
	code_immediate(NULL); /* Reserve 0 slot */
	return yyparse() + error_count;
}

NODE *
alloc_node(enum node_type type, const struct locus *locus)
{
	NODE *node = malloc(sizeof(*node));
	if (!node) {
		yyerror("Not enough memory");
		abort();
	}
	node->type = type;
	node->locus = *locus;
	node->next = NULL;
	return node;
}


static struct status_tab {
	mf_status status;
	char *name;
} status_tab[] = {
	{ mf_success, "success" },
	{ mf_not_found, "not_found" },
	{ mf_failure, "failure" }, 
	{ mf_temp_failure, "temp_failure" },

	{ mf_ston_conv, "ston_conv" },
        { mf_divzero, "divzero" },
        { mf_regcomp, "regcomp" },
        { mf_invip, "invip" },
        { mf_invcidr, "invcidr" },
        { mf_invtime, "invtime" },
        { mf_dbfailure, "dbfailure" },
	{ 0, NULL }
};

int
string_to_stat(char *str, mf_status *status)
{
	struct status_tab *sp;
	for (sp = status_tab; sp->name; sp++)
		if (strcmp(str, sp->name) == 0) {
			*status = sp->status;
			return 0;
		}
	return 1;
}

const char *
mf_status_str(mf_status stat)
{
	struct status_tab *sp;
	for (sp = status_tab; sp->name; sp++)
		if (sp->status == stat)
			return sp->name;
	return "UNKNOWN";
}


/* Print parse tree */

static void print_node_list(NODE *node, int indent);
static void print_node(NODE *node, int indent);
void print_stat(sfsistat stat);
static int dbg_setreply(void *data, char *code, char *xcode, char *message);
static void dbg_setheader(void *data, struct header_node *hdr);

static void
print_level(int level)
{
	level *= 2;
	printf("%*.*s", level, level, "");
}

static void
print_bin_op(enum bin_opcode opcode)
{
	char *p;
	switch (opcode) {
	case bin_and:
		p = "AND";
		break;
		
	case bin_or:
		p = "OR";
		break;
		
	case bin_eq:
		p = "EQ";
		break;
		
	case bin_ne:
		p = "NE";
		break;
		
	case bin_lt:
		p = "LT";
		break;
		
	case bin_le:
		p = "LE";
		break;
		
	case bin_gt:
		p = "GT";
		break;
		
	case bin_ge:
		p = "GE";
		break;
		
	case bin_match:
		p = "MATCH";
		break;
		
	case bin_fnmatch:
		p = "FNMATCH";
		break;

	case bin_add:
		p = "ADD";
		break;
		
	case bin_sub:
		p = "SUB";
		break;
		
	case bin_mul:
		p = "MUL";
		break;
		
	case bin_div:
		p = "DIV";
		break;
		
	default:
		p = "UNKNOWN_OP";
	}
	printf("%s", p);
}

static void
print_node(NODE *node, int level)
{
	switch (node->type) {
	case node_type_string:
		print_level(level);
		printf("CONSTANT: %s\n", node->v.string);
		break;
		
	case node_type_symbol:
		print_level(level);
		printf("SYMBOL: %s\n", node->v.string);
		break;
		
	case node_type_number:
		print_level(level);
		printf("NUMBER: %ld\n", node->v.number);
		break;

	case node_type_variable:
		print_level(level);
		printf("VARIABLE %s (%lu)\n", node->v.asgn.var->name,
		       (unsigned long) node->v.asgn.var->off);
		break;

	case node_type_arg:
		print_level(level);
		printf("ARG %lu\n", node->v.number);
		break;

	case node_type_asgn:
		print_level(level);
		printf("SET %s\n", node->v.asgn.var->name);
		print_node(node->v.asgn.node, level + 1);
		break;
		
	case node_type_if:
		print_level(level);
		printf("COND: \n");
		print_node(node->v.cond.cond, level);
		print_level(level);
		printf("IFTRUE\n");
		print_node_list(node->v.cond.if_true, level+1);
		print_level(level);
		printf("IFFALSE\n");
		print_node_list(node->v.cond.if_false, level+1);
		break;

	case node_type_switch:
	{
		struct case_stmt *pcase;
		
		print_level(level);
		printf("SWITCH: \n");
		print_node(node->v.switch_stmt.node, level+1);
		for (pcase = node->v.switch_stmt.cases;
		     pcase; pcase = pcase->next) {
			
			print_level(level+1);
			if (pcase->valist) {
				struct valist *vp;

				printf("CASE ");
				for (vp = pcase->valist; vp; vp = vp->next)
					switch (vp->value.type) {
					case rettype_string:
						printf("\"%s\" ",
						       vp->value.v.string);
						break;
						
					case rettype_number:
						printf("%ld ",
						       vp->value.v.number);
						break;
						
					default:
						abort();
					}
			} else {
				printf("DEFAULT");
			}
			putchar('\n');

			print_level(level+1);
			printf("ACTION\n");
			print_node_list(pcase->node, level+2);
		}
		break;
	}
	
	case node_type_bin:
		print_level(level);
		print_bin_op(node->v.bin.opcode);
		if (node->v.bin.opcode == bin_match
		    || node->v.bin.opcode == bin_fnmatch) {
			if (node->v.bin.qualifier & QUALIFIER_MX)
				printf(",MX");
			if (node->v.bin.qualifier & QUALIFIER_REGFREE)
				printf(",REGFREE");
		}
		printf("\n");
		print_node(node->v.bin.arg[0], level+1);
		print_node(node->v.bin.arg[1], level+1);
		break;

	case node_type_un:
		print_level(level);
		switch (node->v.un.opcode) {
		case unary_not:
			printf("NOT\n");
			break;
		case unary_minus:
			printf("NEG\n");
			break;
		default:
			abort();
		}
		print_node(node->v.un.arg, level+1);
		break;

	case node_type_next:
		print_level(level);
		printf("next\n");
		break;
		
	case node_type_result:
		if (node->v.ret.code) {
		 	char *s = NULL;
			int expr = 0;
			
			print_level(level);
			if (node->v.ret.message) {
				if (node->v.ret.message->type == node_type_string)
					s = node->v.ret.message->v.string;
				else {
					expr = 1;
					s = "(expression)";
				}
			} 
			dbg_setreply(NULL, node->v.ret.code,
				     node->v.ret.xcode,
				     s);
			if (expr) 
				print_node(node->v.ret.message, level+1);
		}
		print_level(level);
		print_stat(node->v.ret.stat);
		printf("\n");
		break;
		
	case node_type_header:
		print_level(level);
		dbg_setheader(NULL, &node->v.hdr);
		break;

	case node_type_builtin:
	{
		print_level(level);
		printf("BUILTIN %s\n", node->v.builtin.builtin->name);
		print_node_list(node->v.builtin.args, level+1);
		break;
	}

	case node_type_concat:
		print_level(level);
		printf("CONCAT:\n");
		print_node(node->v.concat.arg[0], level+1);
		print_node(node->v.concat.arg[1], level+1);
		break;

	case node_type_push:
		print_level(level);
		printf("PUSH %p:\n", node->v.val);
		break;

	case node_type_regcomp:
		print_level(level);
		printf("REGCOMP:\n");
		print_node(node->v.node, level+1);
		break;

	case node_type_catch:
	{
		int i;
		
		print_level(level);
		printf("CATCH ");
		for (i = 0; i < mf_status_count; i++)
			if (node->v.catch.exmask & EXMASK(i))
				printf("%s ", mf_status_str(i));
		printf("\n");
		print_node_list(node->v.catch.node, level+1);
		printf("END CATCH\n");
		break;
	}

	case node_type_echo:
		print_level(level);
		printf("ECHO:\n");
		print_node(node->v.node, level+1);
		break;

	case node_type_return:
		print_level(level);
		printf("RETURN:\n");
		print_node(node->v.node, level+1);
		break;
		

	case node_type_call:
		print_level(level);
		printf("CALL %s\n", node->v.call.func->name);
		print_node_list(node->v.call.args, level+1);
		break;
		
	default:
		abort();
	}		

}	

static void
print_node_list(NODE *node, int level)
{
	for (; node; node = node->next)
		print_node(node, level);
}

int
function_enumerator(void *sym, void *data)
{
	int i;
	struct function *f = sym;
	
	printf("function %s (", f->name);
	for (i = 0; i < f->parmcount; i++) {
		printf("%s", type_to_string(f->parmtype[i]));
		if (i < f->parmcount-1)
			putchar(',');
	}
	putchar(')');
	if (f->rettype != rettype_unspecified)
		printf(" returns %s", type_to_string(f->rettype));
	printf(":\n");
	print_node_list(f->node, 0);
	printf("END function %s\n", f->name);
	return 0;
}
		
void
print_config()
{
	enum smtp_state tag;
	printf("State handlers:\n");
	printf("---------------\n");
	for (tag = smtp_state_first; tag < smtp_state_count; tag++) {
		if (root_node[tag]) {
			printf("%s:\n", state_to_string(tag));
			print_node_list(root_node[tag], 0);
			putchar('\n');
		}
	}
	printf("User functions:\n");
	printf("---------------\n");
	symbol_enumerate(SYM_FUNC, function_enumerator, NULL);
}


/* Code generation */
int errors;

#define __code_cat3__(a,b,c) a ## b ## c

#define CODE_BINARY(ltype, rtype, op, node) do {                              \
	if (ltype != rtype) {                                                 \
		if (code_cast(&node->locus, rtype, ltype))                    \
			return rettype_unspecified;                           \
	}                                                                     \
	switch (ltype) {                                                      \
	case rettype_number:                                                  \
		code_instr(__code_cat3__(instr_,op,n));                       \
		break;                                                        \
	case rettype_string:                                                  \
		code_instr(__code_cat3__(instr_,op,s));                       \
		break;                                                        \
	default:                                                              \
		parse_error_locus(&node->locus,                               \
				"Invalid argument type in binary operation"); \
		break;                                                        \
	}                                                                     \
} while (0)

int
code_cast(struct locus *locus, instr_rettype fromtype, instr_rettype totype)
{
	if (fromtype == totype) {
		if (totype == rettype_string)
			code_instr(instr_svalidate);
		
		return 0;
	}
	
	switch (fromtype) {
	case rettype_unspecified:
		parse_error_locus(locus,
				  "Expression evaluates to unspecified data type");
		errors = 1;
		return 1;

	case rettype_string:
		code_instr(instr_svalidate);
		code_instr(instr_ston);
		break;
		
	case rettype_number:
		code_instr(instr_ntos);
		break;
	}
	return 0;
}

static void traverse_tree(NODE *node);
static instr_rettype code_node(NODE *node);

static void
jump_fixup(prog_counter_t pos, prog_counter_t endpos)
{
	while (pos) {
		prog_counter_t next = (prog_counter_t)code_peek(pos);
		code_put(pos, (void*)(endpos - pos - 1));
		pos = next;
	}
}

/* The following code is generated for switch statements:
   
	popreg      ; Pop the result of the previous instruction into the reg
	xlat        ; Look up in the table and replace reg with the new value
	            ; If the switch selector is of string type, xlats is
		    ; coded instead
	N           ; Number of elements in the xlat table
	key1        ; First key 
	off1        ; Corresponding jump offset, relative to L0
	key2        ; Second key
	off2        ; Corresponding jump offset, relative to L0 
	.
	.
	.
	keyN
	offN
	bnz L1      ; If xlat failed, jump to the default case
L0:	jreg        ; Jump to the selected branch 
L1:	
	...         ; Default case
L0+off1:
        ...         ; First branch
	jmp L2
L0+off2:
        ...         ; Second branch
	jmp L2
	.
	.
	.
L0+offN:
        ...         ; Nth branch
L2:	
*/
static void
code_switch_branches(NODE *node, instr_rettype type)
{
	prog_counter_t start, refpos, jmppos;
	struct case_stmt *pcase;
	size_t tabsize = 0;

	/* Compute the table size and convert all key values
	   to type */
	for (pcase = node->v.switch_stmt.cases;
	     pcase;
	     pcase = pcase->next) {
		struct valist *vp;
		long v;
		char *p;

		for (vp = pcase->valist; vp; vp = vp->next) {
			tabsize++;
			if (vp->value.type != type) {
				char buf[NUMERIC_BUFSIZE_BOUND];

				/* FIXME: The following code is very crude */
				switch (type) {
				case rettype_number:
					v = strtol(vp->value.v.string, &p, 0);
					if (*p) {
						parse_error_locus(&pcase->locus,
						"cannot convert %s to number",
							vp->value.v.string);
						errors = 1;
						return;
					}
					vp->value.v.number = v;
					break;

				case rettype_string:
					snprintf(buf, sizeof buf, "%ld",
						 vp->value.v.number);
					vp->value.v.string = strdup(buf);
					break;

				default:
					abort();
				}
			}
		}
	}

	code_instr(instr_popreg);
	code_instr(type == rettype_number ? instr_xlat : instr_xlats);
	start =	code_reserve(2*tabsize+1);
	code_instr(instr_bnz);
	code_immediate((void*)1);
	refpos = code_instr(instr_jreg);
		
	/* Generate code for the branches */
	jmppos = 0;
	code_put(start, (void*)tabsize);
	start++;
	for (pcase = node->v.switch_stmt.cases;
	     pcase;
	     pcase = pcase->next) {
		struct valist *vp;
		prog_counter_t pos;
		
		pos = code_get_counter() - refpos;
		traverse_tree(pcase->node);
		if (pcase->next) {
			code_instr(instr_jmp);
			jmppos = code_immediate((void*)jmppos);
		}
		for (vp = pcase->valist; vp; vp = vp->next) {
			code_put(start, (void*)vp->value.v.string);
			code_put(start+1, (void*)pos);
			start += 2;
		}
	}

	/* Fix up jump offsets */
	jump_fixup(jmppos, code_get_counter());
}
	
static void
code_switch(NODE *node)
{
	instr_rettype type;
	
	type = code_node(node->v.switch_stmt.node);
	if (type == rettype_unspecified) {
		errors = 1;
		return;
	}
	code_switch_branches(node, type);
}
		
#define LOCUS_EQ(a,b) \
 (strcmp((a)->file, (b)->file) == 0 && (a)->line == (b)->line)

static instr_rettype
code_node(NODE *node)
{
	instr_rettype ltype, rtype;
	prog_counter_t pos1, pos2, endpos;
	static struct locus *old_locus;

#define MARK_LOCUS()                                                 \
	do if (!old_locus || !LOCUS_EQ(old_locus, &node->locus)) {   \
		old_locus = &node->locus;                            \
		code_instr(instr_locus);                             \
		code_immediate(node->locus.file);                    \
		code_immediate((void*)node->locus.line);             \
	} while (0)

	switch (node->type) {
	case node_type_string:
		MARK_LOCUS();
		code_instr(instr_push_immediate);
		code_immediate(node->v.string);
		return rettype_string;

	case node_type_number:
		MARK_LOCUS();
		code_instr(instr_push_immediate);
		code_immediate((void*)node->v.number);
		return rettype_number;
		
	case node_type_symbol:
		MARK_LOCUS();
		code_instr(instr_symbol);
		code_immediate(node->v.string);
		return rettype_string;

	case node_type_variable:
		MARK_LOCUS();
		code_instr(instr_variable);
		code_immediate((void*)node->v.variable->off);
		return node->v.variable->type;

	case node_type_arg:
		MARK_LOCUS();
		code_instr(instr_arg);
		code_immediate((void*)node->v.number);
		return PARMTYPE(node->v.number);
		
	case node_type_asgn:
		ltype = code_node(node->v.asgn.node);
		node->v.variable->type = ltype;
		code_instr(instr_asgn);
		code_immediate((void*)node->v.variable->off);
		return ltype;
			
	case node_type_if:
		code_node(node->v.cond.cond);
		MARK_LOCUS();
		code_instr(instr_bz);
		pos1 = code_immediate(NULL);
		traverse_tree(node->v.cond.if_true);
		if (node->v.cond.if_false) {
			code_instr(instr_jmp);
			pos2 = code_immediate(NULL);
			traverse_tree(node->v.cond.if_false);
			endpos = code_get_counter ();
			code_put(pos1, (void *)(pos2 - pos1));
			code_put(pos2, (void *)(endpos - pos2 - 1));
		} else 
			code_put(pos1, (void *)(code_get_counter () - pos1 - 1));
		break;

	case node_type_bin:
		ltype = code_node(node->v.bin.arg[0]);

		switch (node->v.bin.opcode) {
		case bin_and:
			MARK_LOCUS();
			/* cond1                        
			   if not true goto X
			   cond2
			   if true goto Y
			X: push 0
			   goto Z
			Y: push 1
                        Z: */
			if (code_cast(&node->locus, ltype, rettype_number))
				return rettype_unspecified;
			code_instr(instr_bz);
			pos1 = code_immediate(NULL);
			rtype = code_node(node->v.bin.arg[1]);
			if (code_cast(&node->locus, rtype, rettype_number))
				return rettype_unspecified;
			code_instr(instr_bnz);
			pos2 = code_immediate((void *)4);
			code_instr(instr_push_immediate);
			code_immediate((void *)0);
			code_instr(instr_jmp);
			code_immediate((void *)2);
			code_instr(instr_push_immediate);
			code_immediate((void*)1);
			code_put(pos1, (void *)(pos2 - pos1));
			return rettype_number;
			
		case bin_or:
			MARK_LOCUS();
			/* cond1                        
			   if true goto X
			   cond2
			   if not true goto Y
			X: push 1
			   goto Z
			Y: push 0
                        Z: */

			if (code_cast(&node->locus, ltype, rettype_number))
				return rettype_unspecified;
			code_instr(instr_bnz);
			pos1 = code_immediate(NULL);
			rtype = code_node(node->v.bin.arg[1]);
			if (code_cast(&node->locus, rtype, rettype_number))
				return rettype_unspecified;
			code_instr(instr_bz);
			pos2 = code_immediate((void *)4);
			code_instr(instr_push_immediate);
			code_immediate((void *)1);
			code_instr(instr_jmp);
			code_immediate((void *)2);
			code_instr(instr_push_immediate);
			code_immediate((void *)0);
			code_put(pos1, (void *)(pos2 - pos1));
			return rettype_number;

		case bin_eq:
			rtype = code_node(node->v.bin.arg[1]);
			CODE_BINARY(ltype, rtype, eq, node);
			return rettype_number;
			
		case bin_ne:
			rtype = code_node(node->v.bin.arg[1]);
			CODE_BINARY(ltype, rtype, ne, node);
			return rettype_number;

		case bin_lt:
			rtype = code_node(node->v.bin.arg[1]);
			CODE_BINARY(ltype, rtype, lt, node);
			return rettype_number;
			
		case bin_le:
			rtype = code_node(node->v.bin.arg[1]);
			CODE_BINARY(ltype, rtype, le, node);
			return rettype_number;

		case bin_gt:
			rtype = code_node(node->v.bin.arg[1]);
			CODE_BINARY(ltype, rtype, gt, node);
			return rettype_number;

		case bin_ge:
			rtype = code_node(node->v.bin.arg[1]);
			CODE_BINARY(ltype, rtype, ge, node);
			return rettype_number;
			
		case bin_match:
			if (code_cast(&node->locus, ltype, rettype_string))
				return rettype_unspecified;
			rtype = code_node(node->v.bin.arg[1]);
			if (code_cast(&node->locus, rtype, rettype_string))
				return rettype_unspecified;
			code_instr(instr_push_immediate);
			code_immediate((void*) (node->v.bin.qualifier &
						 QUALIFIER_REGFREE));
			MARK_LOCUS();
			if (node->v.bin.qualifier & QUALIFIER_MX)
				code_instr(instr_regex_mx);
			else
				code_instr(instr_regex);
			return rettype_number;

		case bin_fnmatch:
			if (code_cast(&node->locus, ltype, rettype_string))
				return rettype_unspecified;
			rtype = code_node(node->v.bin.arg[1]);
			if (code_cast(&node->locus, rtype, rettype_string))
				return rettype_unspecified;
			MARK_LOCUS();
			if (node->v.bin.qualifier & QUALIFIER_MX)
				code_instr(instr_fnmatch_mx);
			else
				code_instr(instr_fnmatch);
			return rettype_number;

		case bin_add:
			if (code_cast(&node->locus, ltype, rettype_number))
				return rettype_unspecified;
			rtype = code_node(node->v.bin.arg[1]);
			if (code_cast(&node->locus, rtype, rettype_number))
				return rettype_unspecified;
			MARK_LOCUS();
			code_instr(instr_add);
			return rettype_number;
			
		case bin_sub:
			if (code_cast(&node->locus, ltype, rettype_number))
				return rettype_unspecified;
			rtype = code_node(node->v.bin.arg[1]);
			if (code_cast(&node->locus, rtype, rettype_number))
				return rettype_unspecified;
			MARK_LOCUS();
			code_instr(instr_sub);
			return rettype_number;
			
		case bin_mul:
			if (code_cast(&node->locus, ltype, rettype_number))
				return rettype_unspecified;
			rtype = code_node(node->v.bin.arg[1]);
			if (code_cast(&node->locus, rtype, rettype_number))
				return rettype_unspecified;
			MARK_LOCUS();
			code_instr(instr_mul);
			return rettype_number;
			
		case bin_div:
			if (code_cast(&node->locus, ltype, rettype_number))
				return rettype_unspecified;
			rtype = code_node(node->v.bin.arg[1]);
			if (code_cast(&node->locus, rtype, rettype_number))
				return rettype_unspecified;
			MARK_LOCUS();
			code_instr(instr_div);
			return rettype_number;
			
		default:
			break;
		}
		return rettype_unspecified;

	case node_type_regcomp:
		ltype = code_node(node->v.node);
		if (code_cast(&node->locus, ltype, rettype_string))
			return rettype_unspecified;
		code_instr(instr_regcomp);
		return rettype_string;

	case node_type_push:
		code_instr(instr_push_immediate);
		code_immediate(node->v.val);
		return rettype_string; /* FIXME */
		
	case node_type_un:
		ltype = code_node(node->v.un.arg);
		if (code_cast(&node->locus, ltype, rettype_number))
			return rettype_unspecified;
		MARK_LOCUS();
		switch (node->v.un.opcode) {
		case unary_not:
			code_instr(instr_not);
			break;
			
		case unary_minus:
			code_instr(instr_neg);
			break;

		default:
			abort();
		}
		return rettype_number;

	case node_type_next:
		MARK_LOCUS();
		code_instr(instr_next);
		break;

	case node_type_result:
		if (node->v.ret.message)
			code_node(node->v.ret.message);
		else {
			code_instr(instr_push_immediate);
			code_immediate(NULL);
		}
		MARK_LOCUS();
		code_instr(instr_result);
		code_immediate((void*)node->v.ret.stat);
		code_immediate(node->v.ret.code);
		code_immediate(node->v.ret.xcode);
		code_instr(NULL);
		break;

	case node_type_header:	
		MARK_LOCUS();
		code_instr(instr_header);
		code_immediate((void*)node->v.hdr.opcode);
		code_immediate(node->v.hdr.name);
		code_immediate(node->v.hdr.value);
		break;

	case node_type_builtin:
	{
		NODE *p;
		int i;

		for (i = 0, p = node->v.builtin.args; p; i++, p = p->next) {
			ltype = code_node(p);
			if (code_cast(&node->locus, ltype,
				      node->v.builtin.builtin->parmtype[i]))
				errors = 1;
		}
		MARK_LOCUS();
		code_instr(instr_builtin);
		code_immediate(node->v.builtin.builtin->name);
		code_immediate(node->v.builtin.builtin->handler);
		return node->v.builtin.builtin->rettype;
	}

	case node_type_call:
	{
		NODE *p;
		int i;

		for (i = 0, p = node->v.call.args; p; i++, p = p->next) {
			ltype = code_node(p);
			if (code_cast(&node->locus, ltype,
				      node->v.call.func->parmtype[i]))
				errors = 1;
		}
		MARK_LOCUS();
		code_instr(instr_funcall);
		code_immediate(node->v.call.func->name);
		code_immediate((void*)node->v.call.func->entry);
		code_instr(instr_adjust);
		code_immediate((void*)node->v.call.func->parmcount);
		if (node->v.call.func->rettype != rettype_unspecified)
			code_instr(instr_pushreg);
		return node->v.call.func->rettype;
	}
		
	case node_type_concat:
		ltype = code_node(node->v.concat.arg[0]);
		if (code_cast(&node->v.concat.arg[0]->locus, ltype,
			      rettype_string))
			return rettype_unspecified;
		rtype = code_node(node->v.concat.arg[1]);
		if (code_cast(&node->v.concat.arg[1]->locus, rtype, rettype_string))
			return rettype_unspecified;
		code_instr(instr_concat);
		return rettype_string;

	case node_type_catch:
	{
		int old_tie_in;
		prog_counter_t ctr;
		
		MARK_LOCUS();
		code_instr(instr_catch);
		pos1 = code_immediate(NULL);
		code_immediate((void*)node->v.catch.count);
		code_immediate((void*)node->v.catch.exmask);
		old_tie_in = stie_in;
		stie_in = synt_catch;
		ctr = jump_pc;
		jump_pc = 0;

		traverse_tree(node->v.catch.node);
		jump_fixup(jump_pc, code_get_counter());
		
		jump_pc = ctr;
		stie_in = old_tie_in;

		code_instr(instr_retcatch);
		if (stie_in == synt_function) {
			code_instr(instr_adjust);
			code_immediate((void*)1);
			code_instr(instr_jmp);
			jump_pc = code_immediate((void*)jump_pc);
		} else {
			code_instr(instr_push_immediate);
                        code_immediate(NULL);
			code_instr(instr_result);
			code_immediate(SMFIS_CONTINUE);
			code_immediate(NULL);
			code_immediate(NULL);
			code_immediate(NULL);
		}

		endpos = code_get_counter ();
		code_put(pos1, (void *)(endpos - pos1));

		break;
	}

	case node_type_echo:
		ltype = code_node(node->v.node);
		if (code_cast(&node->locus, ltype, rettype_string))
			return rettype_unspecified;
		MARK_LOCUS();
		code_instr(instr_echo);
		break;

	case node_type_return:
		if (!func) {
			parse_error_locus(&node->locus,
					  "`return' outside of a function");
			errors = 1;
		} else if (func->rettype == rettype_unspecified) {
			if (node->v.node) {
				parse_error_locus(&node->locus,
				       "`return' with a value, in function "
				       "returning void");
				errors = 1;
			}
			MARK_LOCUS();
			code_instr(instr_jmp);
			jump_pc = code_immediate((void*)jump_pc);
		} else if (node->v.node == NULL) {
			parse_error_locus(&node->locus,
					"`return' with no value, in function "
					"returning non-void");
			errors = 1;
		} else {
			ltype = code_node(node->v.node);
			if (code_cast(&node->v.node->locus, ltype,
				      func->rettype))
				return rettype_unspecified;
			MARK_LOCUS();
			code_instr(instr_popreg);
			code_instr(instr_jmp);
			jump_pc = code_immediate((void*)jump_pc);
			return func->rettype;
		}
		break;

	case node_type_switch:
		code_switch(node);
		break;
		
	default:
		abort();
	}
	return rettype_unspecified;
}

static void
traverse_tree(NODE *node)
{
	for (; node; node = node->next)
		code_node(node);
}

static void optimize(NODE *node);

static void
optimize_node(NODE *node)
{
	int i;
	NODE *p;
	
	switch (node->type) {
	case node_type_bin:
		optimize(node->v.bin.arg[0]);
		optimize(node->v.bin.arg[1]);
		if (node->v.bin.opcode == bin_match) {
			if (node->v.bin.arg[1]->type == node_type_string) {
				int rc;
				regex_t *regex;
				NODE *p = alloc_node(node_type_push,
						     &node->locus);

				regex = emalloc(sizeof *regex);
				rc = regcomp(regex,
					     node->v.bin.arg[1]->v.string,
					     regex_flags);
				if (rc) {
					char errbuf[512];
					regerror(rc,
						 regex, errbuf,
						 sizeof(errbuf));
					parse_error_locus(&node->locus,
						   "Cannot compile regex: %s",
							  errbuf);
					error_count++;
					free(regex);
					break;
				}
				p->v.val = regex;
				node->v.bin.arg[1] = p;
			} else {
				NODE *p = alloc_node(node_type_regcomp,
						     &node->locus);
				p->v.node = node->v.bin.arg[1];
				node->v.bin.arg[1] = p;
				node->v.bin.qualifier |= QUALIFIER_REGFREE;
			}
		}
		break;
		
	case node_type_if:
		optimize(node->v.cond.cond);
		optimize(node->v.cond.if_true);
		optimize(node->v.cond.if_false);
		break;

	case node_type_un:
		optimize(node->v.un.arg);
		break;

	case node_type_asgn:
		optimize(node->v.asgn.node);
		break;

	case node_type_result:
		optimize(node->v.ret.message);
		break;

	case node_type_builtin:
		for (i = 0, p = node->v.builtin.args; p; i++, p = p->next) 
			optimize(p);
		if (strcmp(node->v.builtin.builtin->name, "interval") == 0) {
			if (node->v.builtin.args->type ==
			    node_type_string) {
				time_t t;
				char *endp;
				
				if (parse_time_interval(
					    node->v.builtin.args->v.string,
					    &t, &endp)) {
					parse_error_locus(&node->locus,
					"unrecognized time format (near `%s')",
							  endp);
					error_count++;
					break;
				}
				/* Replace this node */
				node->type = node_type_number;
				node->v.number = t;
			}
		}
		break;

	case node_type_concat:
		optimize(node->v.concat.arg[0]);
		optimize(node->v.concat.arg[1]);
		break;

	case node_type_catch:
		optimize(node->v.catch.node);
		break;

	case node_type_echo:
	case node_type_return:
		optimize(node->v.node);
		break;

	case node_type_switch:
	{
		struct case_stmt *pcase;
		
		optimize(node->v.switch_stmt.node);
		for (pcase = node->v.switch_stmt.cases;
		     pcase; pcase = pcase->next) 
			optimize(pcase->node);
		break;
	}
			
	default:
		break;
	}
}

static void
optimize(NODE *node)
{
	for (; node; node = node->next)
		optimize_node(node);
}

static int
codegen(prog_counter_t *pc, NODE *node)
{
	*pc = code_get_counter();
	errors = 0;
	jump_pc = 0;
	optimize(node);
	traverse_tree(node);
	code_instr(NULL);
	return errors;
}



#define skip_ws(p) while (*(p) && isspace(*(p))) (p)++;
			  
static char *
getword(char **p, char **buf)
{
	int len;
	
	skip_ws(*p);
	if (!**p)
		return NULL;
	for (len = 0; (*p)[len] && !isspace((*p)[len]); len++)
		;
	*buf = malloc(len+1);
	if (*buf) {
		memcpy(*buf, *p, len);
		(*buf)[len] = 0;
	}
	*p += len;
	return *buf;
}

enum regex_mode { regex_enable, regex_disable, regex_set };

static void
pragma_regex(char *text)
{
	char *p, *buf;
	int bit;
	enum regex_mode mode = regex_set;
	
	while (p = getword(&text, &buf)) {
		switch (p[0]) {
		case '+':
			mode = regex_enable;
			p++;
			break;

		case '-':
			mode = regex_disable;
			p++;
			break;

		case '=':
			mode = regex_set;
			p++;
			break;
		}

		if (strcmp (p, "extended") == 0)
			bit = REG_EXTENDED;
		else if (strcmp (p, "icase") == 0)
			bit = REG_ICASE;
		else if (strcmp (p, "newline") == 0)
			bit = REG_NEWLINE;
		else {
			parse_error("Unknown regexp flag: %s", p);
			free(buf);
			return;
		}

		switch (mode) {
		case regex_disable:
			regex_flags &= ~bit;
			break;
		case regex_enable:
			regex_flags |= bit;
			break;
		case regex_set:
			regex_flags = bit;
			break;
		}
		free(buf);
	}
}

static void
pragma_option(char *text)
{
	char *name;
	char *value;
	char *p;
	int len;
	
	if (!getword(&text, &name)) {
		parse_error("Invalid pragma: expected option name");
		return;
	}

#define EXPIRE_SUFFIX 	"expire-interval"
#define EXPIRE_SUFFIX_LEN (sizeof(EXPIRE_SUFFIX)-1)

#define TIMEOUT_SUFFIX  "timeout"
#define TIMEOUT_SUFFIX_LEN (sizeof(TIMEOUT_SUFFIX)-1)
	
	if ((strlen(name) >= EXPIRE_SUFFIX_LEN
	     && strcmp(name + strlen(name) - EXPIRE_SUFFIX_LEN,
		       EXPIRE_SUFFIX) == 0)
	    || (strlen(name) >= TIMEOUT_SUFFIX_LEN
		&& strcmp(name + strlen(name) - TIMEOUT_SUFFIX_LEN,
			  TIMEOUT_SUFFIX) == 0)) {
		/* Special handling for expire and timeout intervals */
		value = strdup(text);
		text += strlen(text);
	} else if (!getword(&text, &value)) {
		parse_error("Invalid pragma: expected option value");
		free(name);
		return;
	}

	skip_ws(text);
	if (*text) {
		parse_error("Invalid pragma: junk after the arguments");
		/* Try to handle it anyway */
	}

	p = value;
	len = strlen(p);
	if ((*p == '"' || *p == '\'') && p[len-1] == *p) {
		p[len-1] = 0;
		p++;
	}
		
	if (set_option(name, p, 0)) {
		if (errno == ENOENT)
			parse_error("Unknown pragma option");
		else
			parse_error("Invalid pragma");
	}
	
	free(name);
	free(value);
}

static void
pragma_stacksize(char *text)
{
	char *p;
	stack_size = strtoul(text, &p, 0);
	if (*p) {
		for (; *p; p++) 
			if (!(isascii(*p) && isspace(*p))) {
				parse_error("Invalid pragma: junk after stack size");
				return;
			}
	}
}

static int
get_interval(char *text, time_t *pinterval)
{
	char *endp;
	time_t interval;

	while (*text && isspace(*text))
		text++;
	if (!*text) {
		parse_error("Invalid pragma: unexpected end of line");
		return 1;
	}
	if (parse_time_interval(text, &interval, &endp)) {
		parse_error("Invalid pragma: unrecognized time format "
			    "(near `%s')",
			    endp);
		return 1;
	}
	*pinterval = interval;
	return 0;
}

/* #pragma database FMTNAME file FILENAME
   #pragma database FMTNAME expire-interval INTERVAL
   #pragma database cache positive-expire-interval INTERVAL
   #pragma database cache negative-expire-interval INTERVAL
*/
static void
pragma_database(char *text)
{
	char *fmtname;
	char *kw;
	struct db_format *fmt;
	
	if (!getword(&text, &fmtname)) {
		parse_error("Invalid pragma: expected database name");
		return;
	}
	if ((fmt = db_format_lookup(fmtname)) == NULL) {
		parse_error("unknown database format: %s", fmtname);
		return;
	}
	if (!getword(&text, &kw)) {
		parse_error("Invalid pragma: expected keyword");
		return;
	}
	if (strcmp(kw, "file") == 0) {
		char *value;
		char *p;
		int len;
		
		if (!getword(&text, &value)) {
			parse_error("Invalid pragma: expected file name");
			return;
		}
		p = value;
		len = strlen(p);
		if ((*p == '"' || *p == '\'') && p[len-1] == *p) {
			p[len-1] = 0;
			p++;
		}
		fmt->dbname = strdup(p);
	} else if (strcmp(kw, "expire-interval") == 0) {
		get_interval(text, &fmt->expire_interval);
	} else if (strcmp(kw, "positive-expire-interval") == 0) {
		if (fmt != cache_format) {
			parse_error("positive-expire-interval is valid only for cache database");
			return;
		}
		get_interval(text, &fmt->expire_interval);
	} else if (strcmp(kw, "negative-expire-interval") == 0) {
		if (fmt != cache_format) {
			parse_error("negative-expire-interval is valid only for cache database");
			return;
		}
		get_interval(text, &negative_expire_interval);
	} else {
		parse_error("Invalid pragma: unrecognized keyword: %s", kw);
		return;
	}
}

void
parse_pragma(char *text)
{
	char *buf;

	while (*text != '#')
		text++;
	++text;
	skip_ws(text);
	text += 6; /* "pragma" */

	if (!getword(&text, &buf))
		parse_error("Empty pragma");
	else if (strcmp(buf, "regex") == 0)
		pragma_regex(text);
	else if (strcmp(buf, "database") == 0)
		pragma_database(text);
	else if (strcmp(buf, "option") == 0)
		pragma_option(text);
	else if (strcmp(buf, "stacksize") == 0)
		pragma_stacksize(text);
	else
		parse_error("Unknown pragma");
	free(buf);
}


/* Test run */
struct sfsistat_tab {
	char *name;
	sfsistat stat;
} sfsistat_tab[] = {
	{ "accept", SMFIS_ACCEPT },
	{ "continue", SMFIS_CONTINUE },
	{ "discard", SMFIS_DISCARD },
	{ "reject", SMFIS_REJECT },
	{ "tempfail", SMFIS_TEMPFAIL },
	{ NULL }
};

const char *
sfsistat_str(sfsistat stat)
{
	struct sfsistat_tab *p;
	for (p = sfsistat_tab; p->name; p++)
		if (p->stat == stat) 
			return p->name;
	return NULL;
}

void
print_stat(sfsistat stat)
{
	struct sfsistat_tab *p;
	for (p = sfsistat_tab; p->name; p++)
		if (p->stat == stat) {
			printf("%s", p->name);
			return;
		}
	printf("%d", stat);
}

const char *
header_command_str(enum header_opcode opcode)
{
	switch (opcode) {
	case header_add:
		return "ADD HEADER ";
		
	case header_replace:
		return "REPLACE HEADER ";
		break;

	case header_delete:
		return "DELETE HEADER ";
		break;
	}
	return "UNKNOWN HEADER COMMAND ";
}

static int
dbg_setreply(void *data, char *code, char *xcode, char *message)
{
	if (code) {
		printf("SET REPLY %s", code);
		if (xcode)
			printf(" %s", xcode);
		if (message)
			printf(" %s", message);
		printf("\n");
	}
	return 0;
}

static void
dbg_setheader(void *data, struct header_node *hdr)
{
	printf("SET HEADER %s: %s\n", hdr->name, hdr->value);
}

void
mailfromd_test(int argc, char **argv)
{
	int i, rc;
	mu_list_t dict = NULL;
	eval_environ_t env;
	char *p, *end;
	long n;
	sfsistat status;
	char *args[9] = {0,0,0,0,0,0,0,0,0};
		
	dict_init(&dict);
	env = create_environment(NULL,
				 dict_getsym, dbg_setreply, dbg_setheader,
				 dict);
	env_init(env);
	for (i = 0; i < argc; i++) {
		if (p = strchr(argv[i], '=')) {
			char *ident = argv[i];
			*p++ = 0;
			if (*ident == '%') 
				env_set_variable(env, ident+1, p);
			else if (isdigit(*ident) && *ident != 0
				 && ident[1] == 0) 
				args[*ident - '0' - 1] = p;
			else 
				dict_install(dict, argv[i], p);
		}
	}

	for (i = state_parms[test_state].cnt; i--; ) {
		switch (state_parms[test_state].types[i]) {
		case rettype_string:
			env_push_string(env, args[i] ? args[i] : "");
			break;
		case rettype_number:
			n = strtol(p, &end, 0);
			if (*end) 
				mu_error("$%d is not a number", i+1);
			env_push_number(env, n);
			break;
		default:
			abort();
		}
	}
	env_make_frame(env);
	rc = xeval(env, test_state);
	env_leave_frame(env, state_parms[test_state].cnt);
	status = environment_get_status(env);
	printf("State %s: ", state_to_string(test_state));
	print_stat(status);
	printf("\n");
	destroy_environment(env);
}

void
set_poll_arg(struct poll_data *poll, int kw, NODE *expr)
{
	switch (kw) {
	case FOR:
		poll->email = expr;
		break;
		
	case HOST:
		poll->client_addr = expr;
		break;
		
	case AS:
		poll->mailfrom = expr;
		break;
		
	case FROM:
		poll->ehlo = expr;
		break;
		
	default:
		abort();
	}
}
		
static int
time_multiplier(const char *str, unsigned *m, unsigned *plen)
{
	static struct timetab {
		char *name;
		unsigned mul;
	} tab[] = {
		{ "seconds", 1 },
		{ "minutes", 60 },
		{ "hours", 60*60 },
		{ "days",  24*60*60 },
		{ "weeks", 7*24*60*60 },
		{ "months", 31*7*24*60*60 },
		{ NULL }
	};
	struct timetab *p;
	int slen;

	for (slen = 0; str[slen]; slen++)
		if (isspace(str[slen]))
			break;

	for (p = tab; p->name; p++) {
		if (p->name[0] == tolower(str[0])) {
			int nlen = strlen(p->name);

			if (nlen > slen)
				nlen = slen;
			
			if (strncasecmp(p->name, str, nlen) == 0) {
				*m = p->mul;
				if (plen)
					*plen = nlen;
				return 0;
			}
		}
	}
	return 1;
}

int
parse_time_interval(char *str, time_t *pint, char **endp)
{
	int rc = 0;
	time_t interval = 0;
	
	while (*str) {
		char *p;
		unsigned long n;
		unsigned mul, len;

		while (*str && isspace(*str))
			str++;

		if (!isdigit(*str) && time_multiplier(str, &mul, &len) == 0) {
			n = 1;
			str += len;
		} else {
			n = strtoul(str, &p, 10);
			if (*p && !isspace(*p)) {
				str = p;
				rc = 1;
				break;
			}

			while (*p && isspace(*p))
				p++;

			str = p;
			if (*str) {
				if (rc = time_multiplier(str, &mul, &len))
					break;
				str += len;
			} else
				mul = 1;
		}
		interval += n*mul;
	}

	if (rc && endp)
		*endp = str;
	*pint = interval;
	return rc;
}

int
convert_rate(const char *arg, double *rate)
{
	double count;
	char *p;

	count = strtod(arg, &p);
	if (*p) {
		time_t div;
		char *endp;

		while (*p && isspace(*p))
			p++;
		if (isalpha(*p)) {
			if (strcmp(p, "per") == 0 && isspace(p[3]))
				p += 3;
		} else if (ispunct(*p))
			p++;

		if (parse_time_interval(p, &div, &endp)) {
			mu_error("unrecognized time format (near `%s')",
				 endp);
			return 1;
		}
		count /= div;
	}
	*rate = count;
	return 0;
}
		
static struct tagtable {
	char *name;
	enum smtp_state tag;
} tagtable[] = {
	{ "connect", smtp_state_connect },
	{ "helo", smtp_state_helo },  
	{ "envfrom", smtp_state_envfrom }, 
	{ "envrcpt", smtp_state_envrcpt }, 
	{ "header", smtp_state_header }, 
	{ "eoh", smtp_state_eoh }, 
	{ "body", smtp_state_body }, 
	{ "eom", smtp_state_eom },
};

enum smtp_state
string_to_state(char *name)
{
	struct tagtable *p;

	for (p = tagtable; p < tagtable + sizeof tagtable/sizeof tagtable[0];
	     p++)
		if (strcasecmp (p->name, name) == 0)
			return p->tag;

	return smtp_state_none;
}

const char *
state_to_string(enum smtp_state state)
{
	if (state-1 < sizeof tagtable/sizeof tagtable[0])
		return tagtable[state-1].name;
	abort();
}


struct variable *
builtin_variable_install(const char *name, instr_rettype type)
{
	struct variable *var = variable_install(name);
	var->type = type;
	var->off = variable_count++;
	return var;
}

static NODE *
_reverse(NODE *list, NODE **root)
{
	NODE *next;

	if (list->next == NULL) {
		*root = list;
		return list;
	}
	next = _reverse(list->next, root);
	next->next = list;
	list->next = NULL;
	return list;
}		

NODE *
reverse(NODE *in)
{
	NODE *root;
	_reverse(in, &root);
	return root;
}


static NODE *
create_on_node(NODE *sel, struct case_stmt *cases)
{
	struct case_stmt *cur;
	unsigned mask = 0;
	char nbuf[NUMERIC_BUFSIZE_BOUND];
	struct function *fp;
	const char *base_name;
	char *fname;
	size_t size;
	int count = 0;
	int argc;
	NODE *args;
	instr_rettype *ptypes, rettype;
	NODE *call_node, *retnode;
		     
	/* Create a list of exceptions to catch */
	for (cur = cases; cur; cur = cur->next) {
		struct valist *vp;
		for (vp = cur->valist; vp; vp = vp->next) {
			mask |= EXMASK(vp->value.v.number);
			count++;
		}
	}

	/* Check if we already have a wrapper function */
	snprintf(nbuf, sizeof nbuf, "%u", mask);
	switch (sel->type) {
	case node_type_builtin:
		base_name = sel->v.builtin.builtin->name;
		argc = sel->v.builtin.builtin->parmcount;
		args = reverse(sel->v.builtin.args);
		ptypes = sel->v.builtin.builtin->parmtype;
		rettype = sel->v.builtin.builtin->rettype;
		break;
	case node_type_call:
		base_name = sel->v.call.func->name;
		argc = sel->v.call.func->parmcount;
		args = sel->v.call.args;
		ptypes = sel->v.call.func->parmtype;
		rettype = sel->v.call.func->rettype;
		break;
	default:
		abort();
	}

	size = 2 + strlen(base_name) + 1 + strlen(nbuf) + 1;
	fname = emalloc(size);
	strcpy(fname, "$");
	strcat(fname, base_name);
	strcat(fname, "_");
	strcat(fname, nbuf);
	strcat(fname, "$");
	
	fp = function_lookup(fname);
	if (!fp) {
		/* Build and register the function */
		NODE *np, *catch_node;
		NODE *head, *tail;
		prog_counter_t pc;
		int i;
		struct function *save_func = func;
		int save_stie_in = stie_in;
		prog_counter_t save_jump_pc = jump_pc;
		
		/* Create the catch statement */
		np = alloc_node(node_type_catch, get_locus());
		np->v.catch.count = count;
		np->v.catch.exmask = 0;
		for (cur = cases; cur; cur = cur->next) {
			struct valist *vp;
			for (i = 0, vp = cur->valist;
			     vp; vp = vp->next, i++)
				np->v.catch.exmask |= EXMASK(vp->value.v.number);
		}
		
		catch_node = np;
		
		np = alloc_node(node_type_return, get_locus());
		np->v.node = alloc_node(node_type_arg, get_locus());
		np->v.node->v.number = 1;
		catch_node->v.catch.node = np;

		/* Build call arguments */
		head = tail = NULL;
		for (i = 1; i <= argc; i++) {
			np = alloc_node(node_type_arg, get_locus());
			np->v.number = i;
			if (!head)
				head = tail = np;
			else {
				tail->next = np;
				tail = np;
			}
		}
				     
		np = alloc_node(node_type_return, get_locus());
		np->v.node = sel; /* FIXME: create a copy */
		switch (np->v.node->type) {
		case node_type_builtin:
			np->v.node->v.builtin.args = head;
			break;
		case node_type_call:
			np->v.node->v.call.args = head;
			break;
		default:
			abort();
		}
		
		catch_node->next = np;
		
		fp = function_install(fname, argc, ptypes,
				      rettype,
				      get_locus());
		fp->exmask = mask;
		fp->node = catch_node;
		fp->entry = code_get_counter();
		
		/* Entry code */
		code_instr(instr_saveex);
		code_immediate((void*)fp->exmask);
		
		func = fp;
		stie_in = synt_function;
		jump_pc = 0;
		
		if (codegen(&pc, fp->node))
			error_count++;
		else {
			jump_fixup(jump_pc, code_get_counter());
			
			/* Exit code */
			code_instr(instr_restex);
			code_instr(instr_return);
		}
		func = save_func;
		stie_in = save_stie_in;
		jump_pc = save_jump_pc;
	}
	free(fname);
	
	call_node = alloc_node(node_type_call, get_locus());
	call_node->v.call.func = fp;
	call_node->v.call.args = args;
	
	retnode = alloc_node(node_type_switch, get_locus());
	retnode->v.switch_stmt.node = call_node;
	retnode->v.switch_stmt.cases = cases;

	return retnode;
}

