/* This file is part of mailfrom filter.
   Copyright (C) 2005, 2006 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#define obstack_chunk_alloc malloc
#define obstack_chunk_free free
#include <obstack.h>
#include <syslog.h>
#include <signal.h>
#include <pwd.h>
#include <grp.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pthread.h>

#include <mailutils/mailutils.h>
#include <mailutils/argp.h>
#include <mailutils/daemon.h>
#include <mailutils/locker.h>
#include <libmilter/mfapi.h>

#include "mailfrom.h"


/* Configurable options */

#define MAILFROMD_DAEMON        0
#define MAILFROMD_TEST          1
#define MAILFROMD_DELETE        2
#define MAILFROMD_LIST          3
#define MAILFROMD_EXPIRE        4
#define MAILFROMD_COMPACT       5    
#define MAILFROMD_SHOW_DEFAULTS 6

int mode = MAILFROMD_DAEMON; /* Default operation mode */
enum smtp_state test_state = smtp_state_envfrom; /* State for --test mode */
int need_config = 1;/* Set if the current mode requires reading the
		       configuration file */
char *config_file = DEFAULT_CONFIG_FILE;
int config_check;      /* Check config file syntax and exit */
int config_ldebug;     /* Enable tracing the lexical analyzer */
int config_ydebug;     /* Enable tracing the parser */
int config_dump_tree;  /* Dump created config tree to stdout */
int config_dump_code;  /* Dump disassembled code to stdout */
int do_transcript;     /* Enable session transript */
int do_trace;          /* Enable tracing configuration */
int debug_level;       /* Debugging level */
int prog_trace_option; /* Trace program execution */
int log_to_stderr;     /* Use stderr for logging */
char *smtp_domain;     /* Default SMTP domain for EHLO command */
char *portspec = DEFAULT_SOCKET; /* Communication socket specification */
int force_remove;      /* Remove local communication socket if it already
		          exists */
int foreground;        /* Stay in foreground */
char *postmaster_email = "";
unsigned long source_address = INADDR_ANY;  /* Source address for TCP
					       connections */
char *syslog_tag;      /* Tag to mark syslog entries with. */
char *pidfile = DEFAULT_PIDFILE;
char *user = DEFAULT_USER;       /* Switch to this user privileges after 
				    startup */
mu_list_t retain_groups;         /* List of group IDs to retain when switching
				    to users privileges. */

/* DBM-related options */
time_t negative_expire_interval = DEFAULT_EXPIRE_INTERVAL/2;
                                 /* Expire negative cache entries after this
				    number of seconds */
time_t expire_interval = 0;      /* When set, overrides all the three above
				    intervals */
int predict_next_option; 
double predict_rate;             /* Prediction rate for --list --format=rates*/

int ignore_failed_reads_option;  /* Ignore failed reads while compacting or
				    expiring */
int all_option;                  /* Process all databases */
char *time_format_string = "%c"; /* String to format the time stamps */
/* DBM-related options end */

int source_info_option;          /* Debug messages include source locations */

char *file_option;               /* File name for DB management commands */
struct db_format *format_option;

/* I/O settings */
unsigned io_timeout = 3;
unsigned io_attempts = 3;


static char *ctx_getsym(void *data, char *str);
static int ctx_setreply(void *data, char *code, char *xcode, char *message);
static void ctx_setheader(void *data, struct header_node *hdr);

/* Per-message data */
struct smtp_io_data; /* Declared later on */
struct message_data {
	eval_environ_t env;
	struct smtp_io_data *io;
	mu_list_t hdr;
	char *helostr;
	char msgid[64];
};

#define MLFIPRIV(ctx)  ((struct message_data*) smfi_getpriv(ctx))

static struct message_data *
priv_get(SMFICTX *ctx)
{
	struct message_data *md = MLFIPRIV(ctx);
	if (!md) {
		md = malloc(sizeof(*md));
		if (!md) 
			mu_error("not enough memory");
		else {
			md->env = create_environment(ctx,
						     ctx_getsym,
						     ctx_setreply,
						     ctx_setheader,
						     ctx);
			clear_rcpt_count(md->env);
			md->hdr = NULL;
			md->io = NULL;
			md->helostr = NULL;
			md->msgid[0] = 0;
			smfi_setpriv(ctx, md);
		}
	}
	if (!md->msgid[0]) {
		char *p = smfi_getsymval(ctx, "i");
		if (p) {
			size_t len = strlen(p);
			if (len > sizeof md->msgid - 3)
				len = sizeof md->msgid - 3;
			memcpy(md->msgid, p, len);
			md->msgid[len++] = ':';
			md->msgid[len++] = ' ';
			md->msgid[len] = 0;
		}
	}
	return md;
}

static int
priv_set_io(SMFICTX *ctx, struct smtp_io_data *io)
{
	struct message_data *md = priv_get(ctx);
	if (!md)
		return 1;
	md->io = io;
	return 0;
}

const char *
mailfromd_msgid(SMFICTX *ctx)
{
	struct message_data *md = priv_get(ctx);
	return md->msgid;
}

char *
mailfromd_timestr(time_t timestamp, char *timebuf, size_t bufsize)
{
	struct tm tm;
	gmtime_r(&timestamp, &tm);
	strftime(timebuf, bufsize, "%c", &tm);
	return timebuf;
}
	
int
priv_store_header_command(SMFICTX *ctx, struct header_node *node)
{
	struct message_data *md = priv_get(ctx);
	if (!md)
		return 1;
	if (!md->hdr)
		mu_list_create(&md->hdr);
	mu_list_append(md->hdr, node);
	return 0;
}


/* Run-time execution */

static char *
ctx_getsym(void *data, char *str)
{
	char *ret = smfi_getsymval(data, str);
	if (!ret) {
		struct message_data *md = priv_get(data);
		if (strcmp (str, "s") == 0)
			ret = md->helostr;
	}
	return ret;
}

static int
ctx_setreply(void *data, char *code, char *xcode, char *message)
{
	if (code) {
		if (!strchr(message, '\n'))
			return smfi_setreply(data, code, xcode, message);
		else {
			return sendmail_mlreply(data, code, xcode, message);
		}
	}
	return 0;
}

static void
ctx_setheader(void *data, struct header_node *hdr)
{
	priv_store_header_command(data, hdr);
}


/* Logging & debugging */

int
syslog_printer (int prio, const char *fmt, va_list ap)
{
#ifdef HAVE_VSYSLOG
	vsyslog (prio, fmt, ap);
#else
	char buf[128];
	vsnprintf (buf, sizeof buf, fmt, ap);
	syslog (prio, "%s", buf);
#endif
	return 0;
}

int
syslog_error_printer (const char *fmt, va_list ap)
{
	return syslog_printer(LOG_ERR, fmt, ap);
}

int
stderr_error_printer (const char *fmt, va_list ap)
{
	fprintf(stderr, "%s: ", program_invocation_short_name);
	vfprintf(stderr, fmt, ap);
	fputc ('\n', stderr);
	return 0;
}

void
vlogmsg(int prio, const char *fmt, va_list ap)
{
	if (log_to_stderr) {
		vfprintf(stderr, fmt, ap);
		fprintf(stderr, "\n");
	} else
		syslog_printer(prio, fmt, ap);
}

void
logmsg(int prio, const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vlogmsg(prio, fmt, ap);
	va_end(ap);
}

void
trace(const char *fmt, ...)
{
	if (do_trace) {
		va_list ap;
		va_start(ap, fmt);
		vlogmsg(LOG_INFO, fmt, ap);
		va_end(ap);
	}
}

void
debug_log(char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vlogmsg(LOG_DEBUG, fmt, ap);
	va_end(ap);
}

void
log_status(sfsistat status, SMFICTX *ctx)
{
	if (debug_level >= 1 && status != SMFIS_CONTINUE) {
		const char *str = sfsistat_str(status);
		if (str)
			logmsg(LOG_INFO,
			       "%s%s", mailfromd_msgid(ctx), str);
		else
			logmsg(LOG_INFO,
			       "%sstatus %d",
			       mailfromd_msgid(ctx), status);
	}
}


/* Sendmail class file support */

static mu_list_t domain_list;

static int
compare_string(const void *item, const void *value)
{
	return strcmp(item, value);
}

/* Read domains from sendmail-style domain file NAME and store them in
   DOMAIN_LIST */
void
read_domain_file(char *name)
{
	FILE *fp;
	char buf[256];
	char *p;
	
	fp = fopen(name, "r");
	if (!fp) {
		mu_error("cannot open file `%s': %s",
			 name, mu_strerror(errno));
		return;
	}

	if (!domain_list) {
		mu_list_create(&domain_list);
		mu_list_set_comparator(domain_list, compare_string);
	}

	while (p = fgets(buf, sizeof buf, fp)) {
		char *q;
		
		while (*p && isspace(*p))
			p++;
		if (*p == '#')
			continue;

		for (q = p + strlen(p) - 1; q > p && isspace(*q); q--)
			*q = 0;

		if (*p == 0)
			continue;
		
		mu_list_append(domain_list, strdup(p));
	}

	fclose(fp);
}

/* Return true if we relay domain NAME */
int
relayed_domain_p(char *name)
{
	char *p = name;

	while (p) {
		if (mu_list_locate(domain_list, p, NULL) == 0) {
			debug2(10,"%s is in relayed domain %s",
				  name, p);
			return 1;
		}
		p = strchr(p, '.');
		if (p)
			p++;
	}
	return 0;
}

/* Return true if CLIENT represents a host we relay. CLIENT is a dotted-quad
   IP address. */
int
host_in_relayed_domain_p(char *client)
{
	char hbuf[NI_MAXHOST];
	
	if (mu_list_is_empty(domain_list))
		return 0;

	if (resolve_ipstr(client, hbuf, sizeof(hbuf)))
		return 0;
	return relayed_domain_p(hbuf);
}


/* SMTP I/O functions */

#define SMTP_MAJOR(c) ((c)/100)

struct smtp_io_data {
	SMFICTX *ctx;        /* Milter context */
	mu_stream_t stream;  /* I/O stream */
	size_t send_off;     /* Send offset */
	size_t recv_off;     /* Receive offset */
	struct obstack stk;  /* Obstack for keeping commands/replies */
	char *command;       /* Last issued command */ 
	char *reply;         /* Last received reply */
	char *start;         /* First line of the reply, if it was multiline */
	char buf[128];       /* Input buffer */
	size_t level;        /* Number of bytes in buf */
	int code;            /* Reply code */
};

void
transcript(struct smtp_io_data *io, char *prefix, const char *msg)
{
	if (do_transcript) {
		int len = strlen(msg);
		if (msg[len-1] == '\n') {
			--len;
			if (len > 0 && msg[len-1] == '\r')
				--len;
		}
		if (len) 
			logmsg(LOG_INFO, "%s%s %*.*s",
			       mailfromd_msgid(io->ctx),
			       prefix, len, len, msg);
	}
}

int
smtp_io_data_init(struct smtp_io_data *dat, SMFICTX *ctx, mu_stream_t stream)
{
	obstack_init(&dat->stk);
	dat->send_off = dat->recv_off = 0;
	dat->stream = stream;
	dat->start = dat->command = dat->reply = NULL;
	dat->level = 0;
	dat->ctx = ctx;
	if (ctx && priv_set_io(ctx, dat)) /* save the private data */
		return 1;
	return 0;
}

void
smtp_io_data_destroy(struct smtp_io_data *dat)
{
	if (dat) {
		mu_stream_close (dat->stream);
		mu_stream_destroy(&dat->stream,
				  mu_stream_get_owner(dat->stream));
		obstack_free(&dat->stk, NULL);
		if (dat->ctx)
			priv_set_io(dat->ctx, NULL);
	}
}

int
smtp_stream_wait(mu_stream_t stream, int flags, int *attempt)
{
	int rc;
	int oflags = flags;
	struct timeval tv;
	while (*attempt < io_attempts) {
		tv.tv_sec = io_timeout;
		tv.tv_usec = 0;
		rc = mu_stream_wait(stream, &oflags, &tv);
		/* FIXME: smfi_progress */
		switch (rc) {
		case 0:
			if (flags & oflags)
				return 0;
			/* FALLTHROUGH */
		case EAGAIN:
		case EINPROGRESS:
			++*attempt;
			continue;

		default:
			return rc;
		}
	}
			
	return ETIMEDOUT;
}

int
smtp_wait(struct smtp_io_data *dat, int flags, int *attempt)
{
	return smtp_stream_wait(dat->stream, flags, attempt);
}

int
smtp_send(struct smtp_io_data *dat, char *command)
{
	int attempt = 0;
	size_t len = strlen(command);

	transcript(dat, "SEND:", command);
	do {
		size_t nb;
		int rc = mu_stream_write(dat->stream, command, len,
					 dat->send_off, &nb);
		if (rc == 0) {
			if (nb == 0) {
				mu_error("stream_write: wrote 0 bytes");
				return -1;
			}
			dat->send_off += nb;
			len -= nb;
			command += nb;
		} else if (rc == EAGAIN) {
			if (attempt < io_attempts) { 
				rc = smtp_wait(dat, MU_STREAM_READY_WR,
                                               &attempt);
				if (rc) {
					mu_error ("smtp_wait failed: %s",
						  mu_strerror(rc));
					return -1;
				}
				continue;
			} else {
				mu_error ("stream_write timed out");
				return -1;
			}
		} else {
			mu_error ("stream_write: %s", mu_strerror (rc));
			return -1;
		}
	} while (len > 0);
	return 0;
}

int
smtp_send2(struct smtp_io_data *dat, char *command, char *arg)
{
	obstack_grow(&dat->stk, command, strlen(command));
	if (arg)
		obstack_grow(&dat->stk, arg, strlen(arg));
	obstack_grow(&dat->stk, "\r\n", 2);
	obstack_1grow(&dat->stk, 0);
	dat->command = obstack_finish(&dat->stk);
	
	return smtp_send(dat, dat->command);
}

int
smtp_send3(struct smtp_io_data *dat, char *command, char *arg1, char *arg2)
{
	obstack_grow(&dat->stk, command, strlen(command));
	obstack_grow(&dat->stk, arg1, strlen(arg1));
	obstack_grow(&dat->stk, arg2, strlen(arg2));
	obstack_grow(&dat->stk, "\r\n", 2);
	obstack_1grow(&dat->stk, 0);
	dat->command = obstack_finish(&dat->stk);
	
	return smtp_send(dat, dat->command);
}

int
smtp_recvline(struct smtp_io_data *dat)
{
	int attempt = 0;
	for (;;) {
		char *p;
		
		if (dat->level == 0) {
			int rc = mu_stream_read(dat->stream,
						dat->buf, sizeof dat->buf,
						dat->recv_off, &dat->level);
			if (rc == 0) {
				if (dat->level == 0) {
					mu_error("stream_read: read 0 bytes");
					return -1;
				}
				dat->recv_off += dat->level;
			} else if (rc == EAGAIN) {
				if (attempt < io_attempts) {
					rc = smtp_wait(dat, 
						       MU_STREAM_READY_RD, 
						       &attempt);
					if (rc) {
						mu_error ("smtp_wait failed: %s",
                                                          mu_strerror(rc));
						return -1;
					}
					continue;
				} else {
					mu_error("stream_read timed out");
					return -1;
				}
			} else {
				mu_error ("stream_read: %s", mu_strerror (rc));
				return -1;
			}
		}
		attempt = 0;
		
		p = memchr(dat->buf, '\n', dat->level);
		if (!p) {
			obstack_grow(&dat->stk, dat->buf, dat->level);
			dat->level = 0;
			continue;
		} else {
			size_t len = p - dat->buf + 1;
			obstack_grow(&dat->stk, dat->buf, len);
			obstack_1grow(&dat->stk, 0);
			dat->reply = obstack_finish(&dat->stk);
			dat->level -= len;
			memmove(dat->buf, dat->buf + len, dat->level);
			break;
		}
	}
	return 0;
}

int
smtp_recv(struct smtp_io_data *dat)
{
	char *p;
	dat->start = NULL;
	do {
		int code;
		int rc = smtp_recvline(dat);
		if (rc)
			return -1;
		transcript(dat, "RECV:", dat->reply);
		code = strtoul (dat->reply, &p, 0);
		if (p - dat->reply != 3 || (*p != '-' && *p != ' ')) {
			mu_error("Unexpected reply from server: %s",
				 dat->reply);
			return -1;
		} else if (!dat->start) {
			dat->start = dat->reply;
			dat->code = code;
		} else if (dat->code != code) {
			mu_error("Unexpected reply code from server: %d",
				 code);
			return -1;
		}
	} while (*p == '-');
	obstack_1grow (&dat->stk, 0);
	obstack_finish (&dat->stk);
	return 0;
}

const char *
smtp_last_sent(struct smtp_io_data *dat)
{
	if (dat->command) {
		size_t len = strcspn(dat->command, "\r\n");
		dat->command[len] = 0;
		return dat->command;
	}
	return "";
}

const char *
smtp_last_received(struct smtp_io_data *dat)
{
	if (dat->reply) {
		size_t len = strcspn(dat->reply, "\r\n");
		dat->reply[len] = 0;
		return dat->reply;
	}
	return dat->reply ? dat->reply : "";
}



/* Milter-specific functions */

/* Verify whether EMAIL address is served by host CLIENT_ADDR. */
mf_status
check_on_host(SMFICTX *ctx, char *email, char *client_addr,
	      char *ehlo, char *mailfrom)
{
	int rc;
	mu_stream_t stream;
	struct smtp_io_data io;
	mf_status status = mf_success;
	int count = 0;
	
	debug2(10, "email = %s, client_addr = %s", email, client_addr);
	rc = mu_tcp_stream_create_with_source_ip (&stream, client_addr, 25,
						  source_address,
						  MU_STREAM_NONBLOCK);
	if (rc) {
		mu_error("%scannot connect to `%s': %s",
			 mailfromd_msgid(ctx), client_addr,
			 mu_strerror (rc));
		return mf_temp_failure;
	}

	while ((rc = mu_stream_open(stream))) {
		if ((rc == EAGAIN || rc == EINPROGRESS)
		    && count < io_attempts) {
			rc = smtp_stream_wait(stream, MU_STREAM_READY_WR,
					      &count);
			if (rc == 0)
				continue;
		}
		mu_error("%sstream_open: %s, %d",
			  mailfromd_msgid(ctx), mu_strerror(rc), count);
		mu_stream_destroy(&stream, mu_stream_get_owner(stream));
		if (rc == EAGAIN || rc == EINPROGRESS)
			return mf_temp_failure;
		else
			return mf_failure;
	}
	debug(100,"stream opened");
	
	if (smtp_io_data_init(&io, ctx, stream))
		status = mf_temp_failure;
	else {
		do {
			if (smtp_recv(&io)) {
				status = mf_temp_failure;
				break;
			} else if (SMTP_MAJOR(io.code) != 2) {
				status = SMTP_MAJOR(io.code) == 4 ?
					mf_temp_failure : mf_not_found;
				break;
			}
			smtp_send2(&io, "HELO ", ehlo);
			if (smtp_recv(&io)) {
				status = mf_temp_failure;
				break;
			} else if (SMTP_MAJOR(io.code) == 5) {
				/* Let's try EHLO, then */
				smtp_send2(&io, "EHLO ", ehlo);
				if (smtp_recv(&io)) {
					status = mf_not_found;
					break;
				}
			}
			if (SMTP_MAJOR(io.code) != 2) {
				status = SMTP_MAJOR(io.code) == 4 ?
					mf_temp_failure : mf_not_found;
				break;
			}
			smtp_send3(&io, "MAIL FROM: <", mailfrom, ">");
			if (smtp_recv(&io)) {
				status = mf_temp_failure;
				break;
			} else if (SMTP_MAJOR(io.code) != 2) {
				status = SMTP_MAJOR(io.code) == 4 ?
					mf_temp_failure : mf_not_found;
				break;
			}
			smtp_send3(&io, "RCPT TO: <", email, ">");
			if (smtp_recv(&io)) {
				status = mf_temp_failure;
				break;
			} else if (SMTP_MAJOR(io.code) != 2) {
				status = SMTP_MAJOR(io.code) == 4 ?
					mf_temp_failure : mf_not_found;
				break;
			}
		} while (0);

		debug4(1,
	             "%spoll exited with status: %s; sent \"%s\", got \"%s\"",
		       mailfromd_msgid(ctx), 
		       mf_status_str(status),
		       smtp_last_sent(&io),
		       smtp_last_received(&io));
		
		smtp_send2(&io, "QUIT", NULL);
		smtp_io_data_destroy(&io);
	}
	
	return status;
}

mf_status
check_mx_records(SMFICTX *ctx, char *email, char *client_addr,
		 char *ehlo, char *mailfrom)
{
	int i;
	mxbuf_t mxbuf;
	mf_status rc, mxstat = getmx(client_addr, mxbuf);

	switch (mxstat) {
	case mf_temp_failure:
		rc = mf_temp_failure;
		break;

	case mf_failure:
	case mf_not_found:
		rc = mf_not_found;
		break;

	case mf_success:
		debug1(2,"Checking MX servers for %s", email);
		rc = mf_not_found;
		for (i = 0; i < MAXMXCOUNT && mxbuf[i]; i++) {
			rc = check_on_host(ctx, email, mxbuf[i],
					   ehlo, mailfrom);
			if (mf_resolved(rc))
				break;
		}
		freemx(mxbuf);
		break;
		
	default:
		abort();
	}
	return rc;
}

/* Method "strict". Verifies whether EMAIL is understood either by
   host CLIENT_ADDR or one of MX servers of its domain */
mf_status
method_strict(SMFICTX *ctx, char *email, char *client_addr,
	      char *ehlo, char *mailfrom)
{
	mf_status rc = cache_get2(email, client_addr);

        if (email[0] == 0)
                return mf_success;

	if (!mf_resolved(rc)) {
		rc = check_on_host(ctx, email, client_addr, ehlo, mailfrom);
		if (!mf_resolved(rc)) {
			mf_status mx_stat;
			mx_stat = check_mx_records(ctx, email, client_addr,
						   ehlo, mailfrom);
			if (mf_resolved(mx_stat)
			    || mx_stat == mf_temp_failure)
				rc = mx_stat;
		}
		if (mf_resolved(rc))
			cache_insert2(email, client_addr, rc);
	}
	return rc;
}

/* Method "standard". Verifies whether EMAIL is understood by
   any of its MXs */
mf_status
method_standard(SMFICTX *ctx, char *email, char *ehlo, char *mailfrom)
{
	mf_status rc = cache_get(email);

        if (email[0] == 0)
                return mf_success;
	if (!mf_resolved(rc)) {	
		char *p = strchr(email, '@');
		if (p == NULL) {
			mu_error("Invalid address: %s", email);
			rc = mf_not_found;
		} else {
			rc = check_mx_records(ctx, email, p+1, ehlo, mailfrom);
			if (rc != mf_success) {
				mf_status host_stat;
				host_stat = check_on_host(ctx, email, p+1, 
							  ehlo, mailfrom);
				if (mf_resolved(host_stat)
				    || host_stat == mf_temp_failure)
					rc = host_stat;
			}
		}
		if (mf_resolved(rc))
			cache_insert(email, rc);
	}
	return rc;
}

mf_status
listens_on (const char *client_addr, int port)
{
	unsigned count = 0;
	mu_stream_t stream;
	int rc;
	
	if (port == 0)
		port = 25;

	debug2(10, "client_addr = %s, port = %d", client_addr, port);
	rc = mu_tcp_stream_create_with_source_ip (&stream, client_addr, 25,
						  source_address,
						  MU_STREAM_NONBLOCK);
	if (rc) {
		debug1 (10, "mu_tcp_stream_create_with_source_ip: %s", mu_strerror (rc));
		return mf_failure;
	}

	while ((rc = mu_stream_open(stream))) {
		if ((rc == EAGAIN || rc == EINPROGRESS)
		    && count < io_attempts) {
			rc = smtp_stream_wait(stream, MU_STREAM_READY_WR,
                                              &count);
			if (rc == 0)
				continue;
		}
		break;
	}
	mu_stream_destroy(&stream, mu_stream_get_owner(stream));

	debug3 (10, "mu_stream_open(%s): %s, %u", client_addr, mu_strerror(rc),
		count);

	return rc == 0 ? mf_success :
		(rc == EAGAIN || rc == EINPROGRESS) ?
		mf_temp_failure : mf_failure;
}

/* Cleanup functions */
void
filter_cleanup(SMFICTX *ctx)
{
	struct message_data *md = priv_get(ctx);
	debug(100,"cleaning up");
	if (md) {
		free(md->helostr);
		destroy_environment(md->env);
		smtp_io_data_destroy(md->io);
		mu_list_destroy(&md->hdr);
		free(md);
		smfi_setpriv(ctx, NULL);
	}
}


/* Milter interface functions */

int
xeval(eval_environ_t env, enum smtp_state tag)
{
	int rc;
	
	rc = eval_environment(env, entry_point[tag]);
	if (rc) {
		struct locus locus;
		const char *symbol = environment_get_null_symbol(env,
								 &locus);
		
		mu_error("Execution of the configuration program was not finished");
		if (symbol)
			mu_error("%s:%lu: Symbol `%s' is not defined",
				 locus.file,
				 (unsigned long) locus.line,
				 symbol);
	}
	return rc;
}

sfsistat
mlfi_eval(SMFICTX *ctx, enum smtp_state tag)
{
	int rc;
	sfsistat status;
	struct message_data *md = priv_get(ctx);

	rc = xeval(md->env, tag);
	if (rc == 0)
		status = environment_get_status(md->env);
	else {
		smfi_setreply(ctx, "410", NULL,
			      "Local configuration error; please try again later");
		status = SMFIS_TEMPFAIL;
	}
	log_status(status, ctx);
	return status;
}

sfsistat
mlfi_helo(SMFICTX *ctx, char *helohost)
{
	sfsistat status;
	struct message_data *md = priv_get(ctx);

	debug1(70, "Processing xxfi_helo: %s", helohost);
	md->helostr = strdup(helohost);
	env_init(md->env);
	env_push_string(md->env, helohost);
	env_make_frame(md->env);
	status = mlfi_eval(ctx, smtp_state_helo);
	env_leave_frame(md->env, 1);
	return status;
}

sfsistat
mlfi_envfrom(SMFICTX *ctx, char **argv)
{
	sfsistat status;
	struct message_data *md = priv_get(ctx);
	debug2(70, "Processing xxfi_envfrom: %s%s", argv[0],
	       argv[1] ? " (more)" : "");
	env_init(md->env);
	env_push_string(md->env, argv[0]);
	env_make_frame(md->env);
	status = mlfi_eval(ctx, smtp_state_envfrom);
	env_leave_frame(md->env, 1);
	return status;
}

sfsistat
mlfi_envrcpt(SMFICTX *ctx, char ** argv)
{
	sfsistat status;
        struct message_data *md = priv_get(ctx);
	
	debug3(70, "Processing xxfi_envrcpt: %s%s - %lu", argv[0],
	       argv[1] ? " (more)" : "",
	       get_rcpt_count(md->env));
	env_init(md->env);
	env_push_string(md->env, argv[0]);
	env_make_frame(md->env);
	incr_rcpt_count(md->env);
	status = mlfi_eval(ctx, smtp_state_envrcpt);
	env_leave_frame(md->env, 1);
	return status;
}

sfsistat
mlfi_header(SMFICTX *ctx, char *headerf, char *headerv)
{
	sfsistat status;
	struct message_data *md = priv_get(ctx);

	debug(10, "Processing xxfi_header:");
	env_init(md->env);
	env_push_string(md->env, headerv);
	env_push_string(md->env, headerf);
	env_make_frame(md->env);
	status = mlfi_eval(ctx, smtp_state_header);
	env_leave_frame(md->env, 2);
	return status;
}

sfsistat
mlfi_eoh(SMFICTX *ctx)
{
	sfsistat status;
        struct message_data *md = priv_get(ctx);
	debug(70, "Processing xxfi_eoh");
	env_init(md->env);
	env_make_frame(md->env);
	status = mlfi_eval(ctx, smtp_state_eoh);
	env_leave_frame(md->env, 0);
	return status;
}

sfsistat
mlfi_body(SMFICTX *ctx, unsigned char * bodyp, size_t len)
{
	sfsistat status;
	struct message_data *md = priv_get(ctx);
	debug1(70, "Processing xxfi_body: %lu", (unsigned long) len);
	env_init(md->env);
	env_push_number(md->env, len);
	env_push_string(md->env, bodyp);
	env_make_frame(md->env);
	status = mlfi_eval(ctx, smtp_state_body);
	env_leave_frame(md->env, 2);
	return status;
}

static int
run_header_command(void *item, void *data)
{
	struct header_node *hdr = item;
	SMFICTX *ctx = data;

	debug3(50,
	       "%s %s: %s",
	       header_command_str(hdr->opcode),
	       hdr->name, hdr->value);
	switch (hdr->opcode) {
	case header_add:
		smfi_addheader(ctx, hdr->name, hdr->value);
		break;
		
	case header_replace:
		smfi_chgheader(ctx, hdr->name, 1, hdr->value);
		break;

	case header_delete:
		smfi_chgheader(ctx, hdr->name, 1, NULL);
		break;
	}
	return 0;
}

	
sfsistat
mlfi_eom(SMFICTX *ctx)
{
        sfsistat status;
	struct message_data *md = priv_get(ctx);

	debug(70, "Processing xxfi_eom");

	env_init(md->env);
	env_make_frame(md->env);
	status = mlfi_eval(ctx, smtp_state_body);
	env_leave_frame(md->env, 0);
	if ((status == SMFIS_ACCEPT || status == SMFIS_CONTINUE) && md->hdr) {
		debug(50,"executing header commands");
		mu_list_do(md->hdr, run_header_command, ctx);
	}
	
	clear_rcpt_count(md->env);

	return status;
}

sfsistat
mlfi_abort(SMFICTX *ctx)
{
	filter_cleanup(ctx);
	return SMFIS_CONTINUE; /* FIXME? */
}		

sfsistat
mlfi_close(SMFICTX *ctx)
{
	filter_cleanup(ctx);
	return SMFIS_CONTINUE; 
}

struct smfiDesc smfilter =
{
	"MailfromFilter",	
	SMFI_VERSION,	
	SMFIF_ADDHDRS|SMFIF_CHGHDRS,
	NULL,	        /* connection info filter */
	NULL,	        /* SMTP HELO command filter */
	NULL,	        /* envelope sender filter */
	NULL,	        /* envelope recipient filter */
	NULL,	        /* header filter */
	NULL,           /* end of header */
	NULL,	        /* body block filter */
	mlfi_eom,	/* end of message */
	mlfi_abort,	/* message aborted */
	mlfi_close,	/* connection cleanup */
};

void
milter_enable_state(enum smtp_state state)
{
	switch (state) {
	case smtp_state_helo:
		smfilter.xxfi_helo = mlfi_helo;
		break;

	case smtp_state_envfrom:
		smfilter.xxfi_envfrom = mlfi_envfrom;
		break;
		
	case smtp_state_envrcpt:
		smfilter.xxfi_envrcpt = mlfi_envrcpt;
		break;
		
	case smtp_state_header:
		smfilter.xxfi_header = mlfi_header;
		break;

	case smtp_state_eoh:
		smfilter.xxfi_eoh = mlfi_eoh;
		break;
		
	case smtp_state_body:
		smfilter.xxfi_body = mlfi_body;
		break;

	default:
		break;
	}
}


/* Option handling.

   There are two classes of options: those that can be set using #pragma
   directive in the configuration file, and those that can be used only in
   command line.

   The latter are handled as usual. The former are processed in two stages:
   first, upon processing command line, their command line settings are
   verified and stored in struct option_cache below. Then, configuration
   file is parsed. Any valid `#pragma option' directives found there overwrite
   values in option_cache. Finally, process_options() is called that scans the
   cache and actually sets any options found there. */
   
void
set_ehlo(void *value)
{
	smtp_domain = value;
}

int
decode_debug(char *arg, int set)
{
	for (; *arg; arg++) {
		switch (*arg) {
		case 'd':
		case 'D':
			config_dump_tree = 1;
			break;

		case 'c':
		case 'C':
			config_dump_code = 1;
			break;
			
		case 'y':
		case 'Y':
			config_ydebug = 1;
			break;

		case 'l':
		case 'L':
			config_ldebug = 1;
			break;

		default:
			if (isdigit(*arg)) {
				if (set)
					debug_level = debug_level * 10 +
						      *arg - '0';
			} else {
				mu_error("Unknown debug flag: %c", *arg);
				return 1;
			}
		}
	}
	return 0;
}

void
set_debug(void *value)
{
	decode_debug(value, 1);
}

void
set_positive_expire(void *value)
{
#ifdef USE_DBM
	cache_format->expire_interval = *(time_t*) value;
	free(value);
#endif
}

void
set_negative_expire(void *value)
{
#ifdef USE_DBM
	negative_expire_interval = *(time_t*) value;
	free(value);
#endif
}

void
set_expire(void *value)
{
#ifdef USE_DBM
	expire_interval = *(time_t*) value;
	free(value);
#endif
}

void
set_rates_expire(void *value)
{
#ifdef USE_DBM
	rate_format->expire_interval = *(time_t*) value;
	free(value);
#endif
}

void
set_port(void *value)
{
	portspec = value;
}

void	
set_user(void *value)
{
	user = value;
}

void	
set_milter_timeout(void *value)
{
	time_t to = *(time_t*) value;
	free(value);
	if (smfi_settimeout(to) == MI_FAILURE) {
		mu_error("Invalid milter timeout: %lu", (unsigned long) to);
		exit(EX_USAGE);
	}
}

void
set_pidfile(void *value)
{
	pidfile = value;
}

void
set_mailfrom(void *value)
{
	postmaster_email = value;
}

void
set_retry(void *value)
{
	io_attempts = strtoul(value, NULL, 0);
}

void
set_timeout(void *value)
{
	io_timeout = *(time_t*) value;
	free(value);
}

static int
load_relay_file(void *item, void *data)
{
	read_domain_file(item);
	return 0;
}

void
set_relay(void *value)
{
	mu_list_do(value, load_relay_file, NULL);
}

void
set_source(void *value)
{
	source_address = *(unsigned long*)value;
}

void
set_group(void *value)
{
	/* do nothing */
}

void
set_source_info(void *value)
{
	source_info_option = (int) value;
}

void
set_lock_retry_count(void *value)
{
#ifdef USE_DBM
	mu_locker_set_default_retry_count(strtoul(value, NULL, 0));
#endif
}

void
set_lock_retry_timeout(void *value)
{
#ifdef USE_DBM
	mu_locker_set_default_retry_timeout(*(time_t*) value);
	free(value);
#endif
}

int
option_string(char *opt, void **pval, char *newval)
{
	if (*pval)
		free(*pval);
	*pval = strdup(newval);
	return 0;
}

int
option_number(char *opt, void **pval, char *newval)
{
	char *p;
	
	strtoul(newval, &p, 10);
	if (*p) {
		errno = EINVAL;
		return 1;
	}
	return option_string(opt, pval, newval);
}

int
option_boolean(char *opt, void **pval, char *newval)
{
	int val;

	if (strcmp (newval, "yes") == 0
	    || strcmp (newval, "true") == 0
	    || strcmp (newval, "t") == 0)
		val = 1;
	else if (strcmp (newval, "no") == 0
		 || strcmp (newval, "false") == 0
		 || strcmp (newval, "nil") == 0)
		val = 0;
	else {
		char *p;
		
		val = strtoul(newval, &p, 10);
		if (*p) {
			errno = EINVAL;
			return 1;
		}
	}
	
	*pval = (void*) val;
	return 0;
}

int
option_debug(char *opt, void **pval, char *newval)
{
	if (decode_debug(newval, 0)) {
		errno = EINVAL;
		return 1;
	}
	return option_string(opt, pval, newval);
}

int
option_time(char *opt, void **pval, char *newval)
{
	time_t interval;
	char *endp;
	if (parse_time_interval(newval, &interval, &endp)) {
		mu_error("%s: unrecognized time format (near `%s')",
			 opt, endp);
		return 1;
	}
	if (!*pval)
		*pval = emalloc(sizeof(time_t));
	*(time_t*) *pval = interval;
	return 0;
}

int
option_email(char *opt, void **pval, char *newval)
{
	if (newval[0] == 0 || strchr(newval, '@'))
		return option_string(opt, pval, newval);
	errno = EINVAL;
	return 1;
}

int
option_pidfile(char *opt, void **pval, char *newval)
{
	if (newval[0] != '/') {
		mu_error("Invalid pidfile name: must be absolute");
		return 1;
	}
	*pval = newval;
	return 0;
}

int
option_relay(char *opt, void **pval, char *newval)
{
	if (!*pval) 
		mu_list_create((mu_list_t*)pval);
	mu_list_append(*pval, strdup(newval));
	return 0;
}

int
option_source(char *opt, void **pval, char *newval)
{
	unsigned long address = inet_addr (newval);
	if (address == INADDR_NONE) {
		struct hostent *phe = gethostbyname (newval);
		if (!phe) {
			mu_error("Cannot resolve `%s'", newval);
			return 1;
		}
		address = *(((unsigned long **) phe->h_addr_list)[0]);
	}
	
	if (*pval == 0) {
		*pval = malloc (sizeof address);
		if (!*pval)
			return 1;
	}
	*(unsigned long*)*pval = address;
	return 0;
}

static int
gid_comp(const void *item, const void *data)
{
	return (gid_t) item != (gid_t) data;
}

int
option_group(char *opt, void **pval, char *newval)
{
	struct group *group = getgrnam(newval);
	if (group) {
		if (!retain_groups) {
			int rc = mu_list_create(&retain_groups);
			if (rc) { 
				mu_error("cannot create list: %s",
					 mu_strerror(rc));
				return 1;
			}
			mu_list_set_comparator(retain_groups, gid_comp);
		}
		mu_list_append(retain_groups, (void*)group->gr_gid);
	} else {
		mu_error("unknown group: %s", newval);
		return 1;
	}
	return 0;
}

struct option_cache {
	char *name;
	void *value;
	int (*handler)(char *opt, void **pval, char *newval);
	void (*set)(void *value);
	int cumulative;
} option_cache[] = {
	{ "ehlo", NULL, option_string, set_ehlo },
	{ "debug", NULL, option_debug, set_debug },
	{ "source-info", NULL, option_boolean, set_source_info },
	{ "expire-interval", NULL, option_time, set_expire },
	{ "positive-expire-interval", NULL, option_time, set_positive_expire },
	{ "negative-expire-interval", NULL, option_time, set_negative_expire },
	{ "rates-expire-interval", NULL, option_time, set_rates_expire },
	{ "port", NULL, option_string, set_port }, /* FIXME: option_port */
	{ "user", NULL, option_string, set_user },
	{ "group", NULL, option_group, set_group },
	{ "milter-timeout", NULL, option_time, set_milter_timeout },
	{ "pidfile", NULL, option_pidfile, set_pidfile },
	{ "mailfrom", NULL, option_email, set_mailfrom },
	{ "retry", NULL, option_number, set_retry },
	{ "timeout", NULL, option_time, set_timeout },
	{ "relay", NULL, option_relay, set_relay, 1 },
        { "source", NULL, option_source, set_source },
	{ "lock-retry-count", NULL, option_number, set_lock_retry_count },
	{ "lock-retry-timeout", NULL, option_time, set_lock_retry_timeout },
	{ NULL }
};

static struct option_cache *
find_option(char *name)
{
	struct option_cache *p;

	for (p = option_cache; p->name; p++)
		if (strcmp(p->name, name) == 0)
			return p;
	return NULL;
}

int
set_option(char *name, char *value, int override)
{
	struct option_cache *p = find_option(name);
	if (!p) {
		errno = ENOENT;
		return 1;
	}
	if (!p->cumulative && !override && p->value)
		return 0;
	return p->handler(name, &p->value, value);
}

void
process_options()
{
	struct option_cache *p;

	for (p = option_cache; p->name; p++)
		if (p->value)
			p->set(p->value);
}


/* Command line parsing */

const char *program_version = "mailfromd (" PACKAGE_STRING ")";
static char doc[] = N_("mailfromd -- MAIL FROM milter checker");
static char args_doc[] = "[var=value...]";

enum mailfromd_option {
	OPTION_ALL = 256,
	OPTION_COMPACT,
	OPTION_DAEMON,
	OPTION_DELETE,
	OPTION_DOMAIN_FILE,
	OPTION_DUMP_CODE,
	OPTION_DUMP_GRAMMAR_TRACE,
	OPTION_DUMP_LEX_TRACE,
	OPTION_DUMP_TREE,
	OPTION_EXPIRE,
	OPTION_FOREGROUND,
	OPTION_IGNORE_FAILED_READS,
	OPTION_LIST,
	OPTION_LOCK_RETRY_COUNT,
	OPTION_LOCK_RETRY_TIMEOUT,
	OPTION_MILTER_TIMEOUT, 
	OPTION_PIDFILE,
	OPTION_POSTMASTER_EMAIL,
	OPTION_PREDICT_NEXT,
	OPTION_RETRY,
	OPTION_SHOW_DEFAULTS,
	OPTION_SOURCE_INFO,
	OPTION_SYSLOG,
	OPTION_TIME_FORMAT,
	OPTION_TIMEOUT,
	OPTION_TRACE,
	OPTION_TRACE_PROGRAM,
};

static struct argp_option options[] = {
#define GRP 0
	{ NULL, 0, NULL, 0,
	  N_("Operation modifiers"), GRP },
#ifdef USE_DBM
	{ "delete", OPTION_DELETE, NULL, 0,
	  N_("Delete given entries from the database"), GRP+1 },
	{ "list", OPTION_LIST, NULL, 0,
	  N_("List cache (default) or rates database"), GRP+1 },
	{ "expire", OPTION_EXPIRE, NULL, 0,
	  N_("Delete expired entries from the cache"), GRP+1 },
	{ "compact", OPTION_COMPACT, NULL, 0,
	  N_("Compact the database"), GRP+1 },
#endif
	{ "test", 't', N_("state"), OPTION_ARG_OPTIONAL,
	  N_("Run in test mode"), GRP+1 },
	{ "lint", 'l', NULL, 0,
	  N_("Check syntax and exit"), GRP+1 },
	{ "syntax-check", 0, NULL, OPTION_ALIAS, NULL, GRP+1 },
	{ "show-defaults", OPTION_SHOW_DEFAULTS, NULL, 0,
	  N_("Show compilation defaults"), GRP+1 },
	{ "daemon", OPTION_DAEMON, NULL, 0,
	  N_("Run in daemon mode (default)"), GRP+1 },

#ifdef USE_DBM
#undef GRP
#define GRP 15	
	{ NULL, 0, NULL, 0,
	  N_("Database management options"), GRP },
	{ "file", 'f', N_("FILE"), 0,
	  N_("DB file name to operate upon (for --list, --delete, --expire)"),
	  GRP+1 },
	{ "format", 'H', N_("FORMAT"), 0,
	  N_("Format of the DB file to operate upon"), GRP+1 },
	{ "ignore-failed-reads", OPTION_IGNORE_FAILED_READS, NULL, 0,
	  N_("Ignore failed reads while compacting the database"), GRP+1 },
	{ "predict", OPTION_PREDICT_NEXT, N_("RATE"), 0,
	  N_("Predict when the user will be able to "
	     "send next message (use with --list --format=rate)"), GRP+1 },
	{ "lock-retry-count", OPTION_LOCK_RETRY_COUNT, N_("NUMBER"), 0,
	  N_("Set maximum number of attempts to acquire the lock"), GRP+1 },
	{ "lock-retry-timeout", OPTION_LOCK_RETRY_TIMEOUT, N_("TIME"), 0,
	  N_("Set timeout for acquiring the lockfile"), GRP+1 },
	{ "expire-interval", 'e', N_("NUMBER"), 0,
	  N_("Set cache expiration interval to NUMBER seconds"), GRP+1 },
	{ "all", OPTION_ALL, NULL, 0,
	  N_("With --compact or --expire: apply the operation to all "
	     "available databases"), GRP+1 },
	{ "time-format", OPTION_TIME_FORMAT, N_("FMT"), 0,
	  N_("Output timestamps using given format (default \"%c\")"), GRP+1 },
#endif
	
#undef GRP	
#define GRP 20
	{ NULL, 0, NULL, 0,
	  N_("General options"), GRP },
	{ "config-file", 'c', N_("FILE"), 0,
	  N_("Read configuration from FILE"), GRP+1 },
	{ "domain", 'D', N_("STRING"), 0,
	  N_("Set default SMTP domain"), GRP+1 },
	{ "ehlo", 0, NULL, OPTION_ALIAS, NULL, GRP+1 },
	{ "postmaster-email", OPTION_POSTMASTER_EMAIL, N_("EMAIL"), 0,
	  N_("Set postmaster email address"), GRP+1 },
	{ "mailfrom", 0, NULL, OPTION_ALIAS, NULL, GRP+1 },
	{ "port", 'p', N_("STRING"), 0,
	  N_("Set communication socket"), GRP+1 },
	{ "remove", 'r', NULL, 0,
	  N_("Force removing local socket file, if it already exists"),
	  GRP+1 },
	{ "foreground", OPTION_FOREGROUND, NULL, 0,
	  N_("Stay in foreground"), GRP+1 },
	{ "pidfile", OPTION_PIDFILE, N_("FILE"), 0,
	  N_("Set pidfile name"), GRP+1 },
	{ "user", 'u', N_("NAME"), 0,
	  N_("Switch to this user privileges after startup"), GRP+1 },
	{ "group", 'g', N_("NAME"), 0,
	  N_("Retain the supplementary group NAME when switching to user "
	     "privileges"),
	  GRP+1 },
	{ "source", 'S', N_("ADDRESS"), 0,
	  N_("Set source address for TCP connections"), GRP+1 },
#undef GRP

#undef GRP
#define GRP 30
	{ NULL, 0, NULL, 0,
	  N_("Timeout control"), GRP },
	{ "milter-timeout", OPTION_MILTER_TIMEOUT, N_("NUMBER"), 0,
	  N_("Set MTA connection timeout"), GRP+1 },
	{ "timeout", OPTION_TIMEOUT, N_("NUMBER"), 0,
	  N_("Set I/O operation timeout (seconds)"), GRP+1 },
	{ "retry", OPTION_RETRY, N_("NUMBER"), 0,
	  N_("Retry failed I/O operations NUMBER times"), GRP+1 },
	{ "relayed-domain-file", OPTION_DOMAIN_FILE, N_("FILE"), 0,
	  N_("Read relayed domains from FILE"), GRP+1 },
	
#undef GRP
#define GRP 40
	{ NULL, 0, NULL, 0,
	  N_("Informational and debugging options"), GRP },
	{ "transcript", 'X', NULL, 0,
	  N_("Enable transcript"), GRP+1 },
	{ "trace", OPTION_TRACE, NULL, 0,
	  N_("Enable tracing configuration"), GRP+1 },
	{ "trace-program", OPTION_TRACE_PROGRAM, NULL, 0,
	  N_("Enable program tracing"), GRP+1 },
	{ "debug", 'd', N_("LEVEL"), 0,
	  N_("Set debugging level"), GRP+1 },
	{ "dump-code", OPTION_DUMP_CODE, NULL, 0,
	  N_("Dump disassembled code"), GRP+1 },
	{ "dump-grammar-trace", OPTION_DUMP_GRAMMAR_TRACE, NULL, 0,
	  N_("Dump grammar traces when parsing the config"), GRP+1 },
	{ "dump-lex-trace", OPTION_DUMP_LEX_TRACE, NULL, 0,
	  N_("Dump lexical analizer traces when parsing the config"), GRP+1 },
	{ "dump-tree", OPTION_DUMP_TREE, NULL, 0,
	  N_("Dump parser tree"), GRP+1 },
	{ "stderr", 's', NULL, 0,
	  N_("Log to stderr"), GRP+1 },
	{ "syslog", OPTION_SYSLOG, NULL, 0,
	  N_("Log to syslog (default)"), GRP+1 },
	{ "log-tag", 'L', N_("STRING"), 0,
	  N_("Set the identifier used in syslog messages to STRING"), GRP+1 },
	{ "source-info", OPTION_SOURCE_INFO, NULL, 0,
	  N_("Debug messages include source information"), GRP+1 },
#undef GRP

#if 0
/*	This entry is to pacify `make check-docs'.  The --log-facility option
	is defined in libmailutils.
 */
	{ "log-facility", }
#endif
	{ NULL }
};

static int
validate_options()
{
#ifdef USE_DBM
	if (all_option &&
	    !(mode == MAILFROMD_COMPACT || mode == MAILFROMD_EXPIRE)) {
		mu_error("--all is meaningful only with --expire or --compact option");
		return 1;
	}
	if (all_option && format_option) 
		mu_error("--format is incompatible with --all");
#endif
	return 0;
}


static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
	switch (key) {
	case 'c':
		config_file = arg;
		break;
		
	case 'D':
		set_option("ehlo", arg, 1);
		break;

	case 'd':
		set_option("debug", arg, 1);
		break;

#ifdef USE_DBM
	case 'e':
		set_option("expire-interval", arg, 1);
		break;

	case 'f':
		file_option = arg;
		break;

	case 'H':
		format_option = db_format_lookup(arg);
		if (format_option == NULL)
			argp_error(state, "unknown DB format");
		break;

	case OPTION_ALL:
		all_option = 1;
		break;
		
	case OPTION_IGNORE_FAILED_READS:
		ignore_failed_reads_option = 1;
		break;

	case OPTION_LOCK_RETRY_COUNT:
		set_option("lock-retry-count", arg, 1);
		break;

	case OPTION_LOCK_RETRY_TIMEOUT:
		set_option("lock-retry-timeout", arg, 1);
		break;

	case OPTION_PREDICT_NEXT:
		if (convert_rate(arg, &predict_rate) == 0)
			predict_next_option = 1;
		mode = MAILFROMD_LIST;
		break;

	case OPTION_TIME_FORMAT:
		time_format_string = arg;
		break;
#endif
		
	case 'L':
		syslog_tag = arg;
		break;

	case 'l':
		config_check = 1;
		break;
		
	case 'p':
		set_option("port", arg, 1);
		break;
		
	case 'r':
		force_remove = 1;
		break;

	case 'S':
		set_option("source", arg, 1);
		break;
		
	case 's':
		log_to_stderr = 1;
		break;

	case 't':
		mode = MAILFROMD_TEST;
		if (arg) {
			test_state =  string_to_state(arg);
			if (test_state == smtp_state_none) 
				argp_error(state,
					   "Unknown smtp state tag: %s", arg);
		}
		log_to_stderr = 1;
		need_config = 1;
		break;

	case 'u':
		set_option("user", arg, 1);
		break;

	case 'g':
		set_option("group", arg, 1);
		break;
		
	case 'X':
		do_transcript = 1;
		break;
		
	case OPTION_COMPACT:
		need_config = 0;
		log_to_stderr = 1;
		mode = MAILFROMD_COMPACT;
		break;
		
	case OPTION_DAEMON:
		mode = MAILFROMD_DAEMON;
		need_config = 1;
		break;
		
	case OPTION_DELETE:
		need_config = 0;
		log_to_stderr = 1;
		mode = MAILFROMD_DELETE;
		break;
		
	case OPTION_DOMAIN_FILE:
		set_option("relay", arg, 1);
		break;

	case OPTION_DUMP_CODE:
		config_dump_code = 1;
		break;
		
	case OPTION_DUMP_GRAMMAR_TRACE:
		config_ydebug = 1;
		break;
		
	case OPTION_DUMP_LEX_TRACE:
		config_ldebug = 1;
		break;
		
	case OPTION_DUMP_TREE:
		config_dump_tree = 1;
		break;
		
	case OPTION_EXPIRE:
		need_config = 0;
		log_to_stderr = 1;
		mode = MAILFROMD_EXPIRE;
		break;

	case OPTION_FOREGROUND:
		foreground = 1;
		break;

	case OPTION_LIST:
		need_config = 0;
		log_to_stderr = 1;
		mode = MAILFROMD_LIST;
		break;

	case OPTION_MILTER_TIMEOUT:
		set_option("milter-timeout", arg, 1);
		break;

	case OPTION_PIDFILE:
		set_option("pidfile", arg, 1);
		break;

	case OPTION_POSTMASTER_EMAIL:
		set_option("mailfrom", arg, 1);
		break;

	case OPTION_RETRY:
		set_option("retry", arg, 1);
		break;
		
	case OPTION_SHOW_DEFAULTS:
		mode = MAILFROMD_SHOW_DEFAULTS;
		need_config = 0;
		break;

	case OPTION_SOURCE_INFO:
		set_option("source-info", "yes", 1);
		break;
		
	case OPTION_SYSLOG:
		log_to_stderr = 0;
		break;

	case OPTION_TIMEOUT:
		set_option("timeout", arg, 1);
		break;

	case OPTION_TRACE:
		do_trace = 1;
		break;

	case OPTION_TRACE_PROGRAM:
		prog_trace_option = 1;
		break;
		
	case ARGP_KEY_FINI:
		if (validate_options())
			exit(EX_USAGE);

		if (!syslog_tag)
			syslog_tag = program_invocation_short_name;

		if (!format_option)
			format_option = db_format_lookup("cache");
		break;
		
	default:
		return ARGP_ERR_UNKNOWN;
	}
	return 0;
}

static const char *capa[] = {
	"auth",
	"license",
	"logging",
	NULL
};

static struct argp argp = {
	options,
	parse_opt,
	args_doc,
	doc,
	NULL,
	NULL,
	NULL
};


/* Auxiliary functions */

/* Check whether local socket P exists. Act in accordance with the settings
   of --remove command line option. */
static void
check_local_portspec(char *p)
{
	struct stat st;
	
	if (stat(p, &st)) {
		if (errno == ENOENT)
			return; /* That's OK. The socket must not exist */
		mu_error("cannot stat file `%s': %s",
			 p, mu_strerror(errno));
		exit(EX_SOFTWARE);
	} else if (!force_remove) {
		mu_error("File `%s' already exist. Remove it or use --remove option.", p);
		exit(EX_SOFTWARE);
	}

	if (unlink (p)) {
		mu_error("cannot unlink file `%s': %s", p, mu_strerror(errno));
		exit(EX_SOFTWARE);
	}
}

void
check_portspec(char *portspec)
{
	static struct spectab {
		char *proto;
		void (*check)(char *);
	} *sp, spectab[] = { 
		{ "unix",  check_local_portspec }, 
		{ "local", check_local_portspec },
		{ "inet",  NULL },
		{ "inet6", NULL },
 	};
	
	if (strchr(portspec, ':') == NULL) {
		mu_error("Missing protocol in port specification");
		exit(1);
	}

	for (sp = spectab; sp < spectab + sizeof(spectab)/sizeof(spectab[0]);
	     sp++) {
		int len = strlen(sp->proto);
		if (strlen(portspec) > len+1
		    && strncasecmp(portspec, sp->proto, len) == 0
		    && portspec[len] == ':') {
			if (sp->check)
				sp->check(portspec + len + 1);
			return;
		}
	}
	mu_error("Unknown port specification: %s", portspec);
}

/* Switch to the given UID/GID */
int
switch_to_privs(uid_t uid, gid_t gid)
{
	int rc = 0;
	gid_t *emptygidset;
	size_t size = 1, j = 1;
	mu_iterator_t itr;

	/* Create a lis of supplementary groups */
	mu_list_count (retain_groups, &size);
	size++;
	emptygidset = emalloc (size * sizeof emptygidset[0]);
	emptygidset[0] = gid ? gid : getegid();

	if (mu_list_get_iterator(retain_groups, &itr) == 0) {
		for (mu_iterator_first (itr);
		     !mu_iterator_is_done (itr); mu_iterator_next (itr)) 
			mu_iterator_current (itr, (void **)(emptygidset + j++));
		mu_iterator_destroy(&itr);
	}

	/* Reset group permissions */
	if (geteuid() == 0 && setgroups(j, emptygidset)) {
		mu_error("setgroups(1, %lu) failed: %s",
		         (unsigned long) emptygidset[0],
			 mu_strerror(errno));
		rc = 1;
	}
	free(emptygidset);
	
	/* Switch to the user's gid. On some OSes the effective gid must
	   be reset first */

#if defined(HAVE_SETEGID)
	if ((rc = setegid(gid)) < 0)
		mu_error("setegid(%lu) failed: %s",
			(unsigned long) gid, mu_strerror(errno));
#elif defined(HAVE_SETREGID)
	if ((rc = setregid(gid, gid)) < 0)
		mu_error("setregid(%lu,%lu) failed: %s",
		         (unsigned long) gid, (unsigned long) gid,
			 mu_strerror(errno));
#elif defined(HAVE_SETRESGID)
	if ((rc = setresgid(gid, gid, gid)) < 0)
		mu_error("setresgid(%lu,%lu,%lu) failed: %s",
		         (unsigned long) gid,
		         (unsigned long) gid,
		         (unsigned long) gid,
			 mu_strerror(errno));
#endif

	if (rc == 0 && gid != 0) {
		if ((rc = setgid(gid)) < 0 && getegid() != gid) 
			mu_error("setgid(%lu) failed: %s",
				 (unsigned long) gid, mu_strerror(errno));
		if (rc == 0 && getegid() != gid) {
			mu_error("cannot set effective gid to %lu",
			         (unsigned long) gid);
			rc = 1;
		}
	}

	/* Now reset uid */
	if (rc == 0 && uid != 0) {
		uid_t euid;

		if (setuid(uid)
		    || geteuid() != uid
		    || (getuid() != uid
			&& (geteuid() == 0 || getuid() == 0))) {
			
#if defined(HAVE_SETREUID)
			if (geteuid() != uid) {
				if (setreuid(uid, -1) < 0) { 
					mu_error("setreuid(%lu,-1) failed",
					         (unsigned long) uid,
						 mu_strerror(errno));
					rc = 1;
				}
				if (setuid(uid) < 0) {
					mu_error("second setuid(%lu) failed",
					         (unsigned long) uid,
						 mu_strerror(errno));
					rc = 1;
				}
			} else
#endif
				{
					mu_error("setuid(%lu) failed",
					         (unsigned long) uid,
						 mu_strerror(errno));
					rc = 1;
				}
		}
	
		euid = geteuid();
		if (uid != 0 && setuid(0) == 0) {
			mu_error("seteuid(0) succeeded when it should not");
			rc = 1;
		} else if (uid != euid && setuid(euid) == 0) {
			mu_error("cannot drop non-root setuid privileges");
			rc = 1;
		}

	}

	return rc;
}

/* Check whether pidfile NAME exists and if so, whether its PID is still
   active. Exit if it is. */
void
check_pidfile(char *name)
{
	unsigned long pid;
	FILE *fp = fopen(name, "r");
	if (!fp) {
		if (errno == ENOENT)
			return;
		mu_error("Cannot open pidfile `%s': %s",
			 name, mu_strerror(errno));
		exit(EX_TEMPFAIL);
	}
	if (fscanf(fp, "%lu", &pid) != 1) {
		mu_error("Cannot get pid from pidfile `%s'", name);
	} else {
		if (kill(pid, 0) == 0) {
			mu_error("%s appears to run with pid %lu. If it does not, remove `%s' and retry.",
				 program_invocation_short_name,
				 pid,
				 name);
			exit(EX_USAGE);
		}
	}
	fclose(fp);
	if (unlink(pidfile)) {
		mu_error("Cannot unlink pidfile `%s'",
			 name, mu_strerror(errno));
		exit(EX_USAGE);
	}
}

static char *license =
    "   This is mailfromd filter.\n"
    "   Copyright (C) 2005, 2006 Sergey Poznyakoff\n"
    "\n"
    "   ls is free software; you can redistribute it and/or modify\n"
    "   it under the terms of the GNU General Public License as published by\n"
    "   the Free Software Foundation; either version 2 of the License, or\n"
    "   (at your option) any later version.\n"
    "\n"
    "   This program is distributed in the hope that it will be useful,\n"
    "   but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
    "   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
    "   GNU General Public License for more details.\n"
    "\n"
    "   You should have received a copy of the GNU General Public License along\n"
    "   with this program; if not, write to the Free Software Foundation,\n"
    "   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA\n\n";

static void
priv_setup()
{
	if (getuid() == 0) {
		if (!user) {
			mu_error("When running as root, --user option is mandatory.");
			exit(EX_USAGE);
		} else {
			struct passwd *pw = getpwnam(user);
			if (!pw) {
				mu_error("no such user: %s", user);
				exit(EX_SOFTWARE);
			}
			if (pw && switch_to_privs(pw->pw_uid, pw->pw_gid))
				exit(EX_SOFTWARE);
		}
	}
}

void
mailfromd_daemon()
{
	int rc;

	priv_setup();

	if (!foreground)
		check_pidfile(pidfile);

	check_portspec(portspec);
	
	if (smfi_setconn(portspec) == MI_FAILURE) {
		mu_error("smfi_setconn: %s", mu_strerror(errno));
		exit(EX_SOFTWARE);
	}
	
	if (smfi_register(smfilter) == MI_FAILURE) {
		mu_error("smfi_register failed");
		exit(EX_UNAVAILABLE);
	}

 	if (!foreground) {
		if (daemon(0, 0) == -1) {
			mu_error("cannot become a daemon: %s",
				 mu_strerror(errno));
			exit(EX_OSERR);
		}
		umask(0117);
		mu_daemon_create_pidfile(pidfile);
		rc = smfi_main();
		mu_error("OK. REMOVING PIDFILE");
		mu_daemon_remove_pidfile();
	} else {
		umask(0117);
		rc = smfi_main();
		mu_error("smfi_main failed: rc=%d errno=%s",
			 rc, strerror (errno));
	}
	exit(rc);
}

void
assert_db_format()
{
	if (!format_option) {
		mu_error("operation is not applicable");
		exit(EX_USAGE);
	}
}

char *
get_db_name()
{
	assert_db_format();
	if (file_option)
		return file_option;
	if (format_option->dbname)
		return format_option->dbname;
	mu_error("Database file name is not given");
	exit(EX_USAGE);
}

void
db_format_setup()
{
#ifdef USE_DBM
	cache_format = db_format_install(cache_format);
	rate_format = db_format_install(rate_format);
	greylist_format = db_format_install(greylist_format);

	mu_locker_set_default_flags (MU_LOCKER_KERNEL|MU_LOCKER_RETRY,
				     mu_locker_assign);
	mu_locker_set_default_retry_timeout (1);
	mu_locker_set_default_retry_count (3);
#endif
}

void
mailfromd_delete(int argc, char **argv)
{
	int i;
	char *name = get_db_name();

	for (i = 0; i < argc; i++) {
		debug2(50,"deleting %s info for %s",
		       format_option->name, argv[i]);
		db_delete(name, argv[i]);
	}
}

void
mailfromd_list(int argc, char **argv)
{
	int rc;
	char *name = get_db_name();
	
	if (!format_option->print_item) {
		/* Should not happen */
		mu_error("list is not applicable to this DB format");
		exit(EX_USAGE);
	}		

	rc = 0;
	if (argc) {
		int i;
		for (i = 0; i < argc; i++) 
			rc += db_list_item(name, argv[i],
					   format_option->print_item);
	} else
		rc = db_list(name, format_option->print_item);
	exit(rc != 0);
}

static int
db_proc_enumerator(void *sym, void *data)
{
	struct db_format *fmt = sym;
	int (*func)(char *name, db_expire_t exp) = data;
	
	if (fmt->expire) 
		func(fmt->dbname, fmt->expire);
	return 0;
}

void
mailfromd_expire()
{
#ifdef USE_DBM
	priv_setup();
	if (all_option) 
		symbol_enumerate(SYM_DB_FORMAT, db_proc_enumerator, db_expire);
	else {
		if (!format_option->expire) {
			mu_error("expire is not applicable to this DB format");
			exit(EX_USAGE);
		}		
		exit(db_expire(get_db_name(), format_option->expire) != 0);
	}
#endif
}	

void
mailfromd_compact()
{
#ifdef USE_DBM
	priv_setup();
	if (all_option) 
		symbol_enumerate(SYM_DB_FORMAT, db_proc_enumerator,
				 db_compact);
	else {
		if (!format_option->expire) {
			mu_error("compact is not applicable to this DB format");
			exit(EX_USAGE);
		}
		exit(db_compact(get_db_name(), format_option->expire) != 0);
	}
#endif
}

#ifdef USE_DBM
static int
db_format_enumerator(void *sym, void *data)
{
	struct db_format *fmt = sym;
	
	printf("%s database:      %s\n", fmt->name, fmt->dbname);
	printf("%s expiration:    %lu\n", fmt->name, fmt->expire_interval);
	if (strcmp(fmt->name, "cache") == 0)
		printf("%s negative expiration: %lu\n", fmt->name,
		       negative_expire_interval);
	return 0;
}
#endif

void
mailfromd_show_defaults()
{
	printf("version:             %s\n", VERSION);
	printf("configuration file:  %s\n", DEFAULT_CONFIG_FILE);
	printf("user:                %s\n", DEFAULT_USER);
	printf("statedir:            %s\n", MAILFROMSTATEDIR);
	printf("socket:              %s\n", DEFAULT_SOCKET);
	printf("pidfile:             %s\n", DEFAULT_PIDFILE);
#ifdef USE_DBM
	printf("database format:     ");
#if defined WITH_GDBM
	printf("GDBM");
#elif defined WITH_BDB
	printf("Berkeley DB %d.x", WITH_BDB);
#endif
	printf("\n");

	symbol_enumerate(SYM_DB_FORMAT, db_format_enumerator, NULL);
#else
	printf("cache database:      %s\n", "DISABLED");
#endif
}	

int
main(int argc, char **argv)
{
	int index;

	mu_license_text = license;
	MU_AUTH_REGISTER_ALL_MODULES();
	if (!program_invocation_short_name)
		program_invocation_short_name = argv[0];
	mu_error_set_print(stderr_error_printer);
	builtin_setup();
	db_format_setup();
	mu_argp_init (program_version, "<" PACKAGE_BUGREPORT ">");
	
	mu_argp_parse (&argp, &argc, &argv, 0, capa, &index, NULL);

	argv += index;
	argc -= index;

	if (need_config
	    && parse_config(config_file, config_ydebug, config_ldebug))
		exit(EX_CONFIG);
	process_options();

	/* Process some special options */
#ifdef USE_DBM
	if (expire_interval) 
		cache_format->expire_interval = negative_expire_interval =
			rate_format->expire_interval = expire_interval;
#endif

	builtin_post_setup();

	if (debug_level >= 100)
		prog_trace_option = 1;
		
	if (config_dump_tree)
		print_config();
	if (config_dump_code)
		print_code();
	
	/* Set up default values */
	if (!log_to_stderr) {
		openlog(syslog_tag, LOG_PID, log_facility);
		mu_error_set_print(syslog_error_printer);
	} 

	if (config_check || config_dump_code || config_dump_tree
	    || config_ldebug || config_ydebug)
		exit(EX_OK);
	
	switch (mode) {
	case MAILFROMD_COMPACT:
		mailfromd_compact();
		break;
			
	case MAILFROMD_DAEMON:
		if (argc > 0) {
			mu_error("too many arguments");
			exit(EX_USAGE);
		}
		mailfromd_daemon();
		break;
		
	case MAILFROMD_TEST:
		mailfromd_test(argc, argv);
		break;
		
	case MAILFROMD_DELETE:
		mailfromd_delete(argc, argv);
		break;

	case MAILFROMD_LIST:
		mailfromd_list(argc, argv);
		break;

	case MAILFROMD_EXPIRE:
		mailfromd_expire();
		break;
		
	case MAILFROMD_SHOW_DEFAULTS:
		mailfromd_show_defaults();
		break;
	}

	exit(EX_OK);
}
