/* This file is part of mailfrom filter.
   Copyright (C) 2006 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <setjmp.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <assert.h>
#include <fnmatch.h>
#include <syslog.h>
#include "mailfrom.h"

#define SP(s) ((s) ? (s) : "''")


/* Code generation support */
#define CODE_INITIAL    128
#define CODE_INCREMENT  128

static prog_counter_t pc, pmax;
static instr_t *prog;
size_t stack_size = 512;

void
code_init()
{
	pmax = CODE_INITIAL;
	prog = malloc(pmax*sizeof prog[0]);
	if (!prog) {
		mu_error("Not enough memory to compile the program");
		exit(1);
	}
}

prog_counter_t
code_get_counter()
{
	return pc;
}

prog_counter_t
code_reserve(size_t count)
{
	prog_counter_t ret = pc;
	if (pc + count > pmax) {
		size_t incr = (pc + count - pmax + CODE_INCREMENT - 1)
 			       / CODE_INCREMENT;
		pmax += incr * CODE_INCREMENT;
		prog = realloc(prog, pmax*sizeof prog[0]);
		if (!prog) {
			mu_error("Not enough memory to compile the program");
			exit(1);
		}
	}
	pc += count;
	return ret;
}

prog_counter_t
code_instr(const instr_t ptr)
{
	if (pc >= pmax) {
		pmax += CODE_INCREMENT;
		prog = realloc(prog, pmax*sizeof prog[0]);
		if (!prog) {
			mu_error("Not enough memory to compile the program");
			exit(1);
		}
	}
	prog[pc] = ptr;
	return pc++;
}

prog_counter_t
code_immediate(const void *ptr)
{
	return code_instr(ptr);
}

void
code_put(prog_counter_t pos, void *ptr)
{
	assert(pos < pmax);
	prog[pos] = (instr_t) ptr;
}

instr_t
code_peek(prog_counter_t pos)
{
	assert(pos < pmax);
	return prog[pos];
}


/* Run-time evaluation */

typedef void *STKVAL;

struct eval_environ {
	prog_counter_t pc;     /* Program counter */

	STKVAL *stack;         /* Stack space */
	prog_counter_t tos;    /* Top of stack: 0 <= tos <= toh */
	prog_counter_t toh;    /* Top of heap:  tos <= toh <= stack_size */
	prog_counter_t base;   /* Base pointer */
	STKVAL reg;            /* General purpose register */
	STKVAL *var;           /* Data space: %<name> variables */
	
	struct locus locus;    /* Program locus corresponding to PC */

	/* Sendmail interaction data: */
	SMFICTX *ctx;          /* Milter Context */
	void *data;            /* MTA symbol table */
	/* methods to access the latter */
	char *(*getsym)(void *data, char *str);
	int (*setreply)(void *data, char *code, char *xcode, char *message);
	void (*setheader)(void *data, struct header_node *hdr);

	/* Non-local exits */
	prog_counter_t catch[mf_status_count];
	
	/* Transaction cache: */
	mu_list_t dict;        /* DNS lookup cache */
	
	/* Exit information */
	char *null_symbol;     /* Non resolved symbol that caused the abort */
	sfsistat status;       /* Program exit status */
	jmp_buf x;             /* Return point */
};

/* Auxiliary functions */
#define advance_pc(env, cnt) ((env)->pc += cnt)
#define adjust_stack(env, cnt) ((env)->tos -= cnt)

void
prog_trace(eval_environ_t env, const char *fmt, ...)
{
	char buf[512];
	va_list ap;
	
	va_start(ap, fmt);
	vsnprintf(buf, sizeof buf, fmt, ap);
	va_end(ap);
	logmsg(LOG_DEBUG, "%4lu: %s:%lu: %s",
	       (unsigned long) env->pc,
	       env->locus.file, (unsigned long) env->locus.line,
	       buf);
}	

void
runtime_warning(eval_environ_t env, const char *fmt, ...)
{
	int n;
	va_list ap;
	char buf[512];
	
	n = snprintf(buf, sizeof buf, "RUNTIME WARNING near %s:%lu: ",
		     env->locus.file, (unsigned long) env->locus.line);
	va_start(ap, fmt);
	vsnprintf(buf + n, sizeof buf - n, fmt, ap);
	va_end(ap);
	mu_error(buf);
}
	      
void
runtime_error(eval_environ_t env, const char *fmt, ...)
{
	int n;
	va_list ap;
	char buf[512];
	
	n = snprintf(buf, sizeof buf, "RUNTIME ERROR near %s:%lu: ",
		     env->locus.file, (unsigned long) env->locus.line);
	va_start(ap, fmt);
	vsnprintf(buf + n, sizeof buf - n, fmt, ap);
	va_end(ap);
	mu_error(buf);
	env->status = SMFIS_TEMPFAIL; /* FIXME */
	longjmp(env->x, 1);
}
	      
void *
get_immediate(eval_environ_t env, unsigned n)
{
	return prog[env->pc + n + 1];
}

STKVAL
get_arg(eval_environ_t env, unsigned n)
{
	return env->stack[env->tos - n - 1];
}

void
push(eval_environ_t env, STKVAL val)
{
	if (env->tos >= env->toh) {
		debug2(100, "tos=%lu, toh=%lu",
		       (unsigned long) env->tos,
		       (unsigned long) env->toh);
		runtime_error(env, "Out of stack space; increase #pragma stacksize");
	}
	env->stack[env->tos++] = val;
}

STKVAL
pop(eval_environ_t env)
{
	if (env->tos == 0)
		runtime_error(env, "Stack underflow");
	return env->stack[--env->tos];
}

#define B2STACK(s) (s + sizeof(STKVAL) - 1) / sizeof(STKVAL)

STKVAL
heap_reserve(eval_environ_t env, size_t size)
{
	size_t words = B2STACK(size);
	
        if (env->toh - words <= env->tos) {
                /* Heap overrun: */
                /* gc(); */
                if (env->tos - words <= env->tos) 
                        runtime_error(env,
				      "Heap overrun; increase #pragma stacksize");
        }
        env->toh -= words;
        return env->stack + env->toh--;
}

void
pushs(eval_environ_t env, char *s)
{
	push(env, (STKVAL) strcpy(heap_reserve(env, strlen(s) + 1), s));
}

/* Auxiliary instructions */
void
instr_locus(eval_environ_t env)
{
	env->locus.file = get_immediate(env, 0);
	env->locus.line = (size_t) get_immediate(env, 1);
	if (prog_trace_option)
		prog_trace(env, "LOCUS");
	advance_pc(env, 2);
}

void
dump_locus(prog_counter_t i, prog_counter_t *ret __attribute__((__unused__)))
{
	printf("\"%s\" %lu", (char*) prog[i],
	       (unsigned long) (size_t) prog[i+1]);
}

/* Stack manipulation instructions */
void
instr_xchg(eval_environ_t env)
{
	prog_counter_t p = env->tos - 1;
	STKVAL tmp = env->stack[p];
	env->stack[p] = env->stack[p - 1];
	env->stack[p - 1] = tmp;
	if (prog_trace_option)
		prog_trace(env, "XCHG");
}

void
instr_dup(eval_environ_t env)
{
	if (prog_trace_option)
		prog_trace(env, "DUP");
	push(env, get_arg(env, 0));
}

void
instr_pop(eval_environ_t env)
{
	if (prog_trace_option)
		prog_trace(env, "POP");
	pop(env);
}

void
instr_push_immediate(eval_environ_t env)
{
	if (prog_trace_option)
		prog_trace(env, "PUSH_IMMEDIATE %p", get_immediate(env, 0));
	push(env, get_immediate(env, 0));
	advance_pc(env, 1);
}

void
dump_push_immediate(prog_counter_t i,
		    prog_counter_t *ret __attribute__((__unused__)))
{
	printf("%p", prog[i]); /* FIXME: Bad. Should have a way for determining
				  operand type */
}

/* Type conversion instructions */
void
instr_ston(eval_environ_t env)
{
	char *s = pop(env);
	char *p;
	long v = strtol(s, &p, 0);

	if (prog_trace_option)
		prog_trace(env, "STON %s", s);
	if (*p) {
		if (env_catch(env, mf_ston_conv) == 0)
			return;
		runtime_error(env,
		    "Cannot convert stack value to string (stopped at %8.8s)",
			      p);
	}
	push(env, (STKVAL) v);
}

void
instr_ntos(eval_environ_t env)
{
	long v = (long) get_arg(env, 0);
	char buf[NUMERIC_BUFSIZE_BOUND];
	
	if (prog_trace_option)
		prog_trace(env, "NTOS %ld", v);
	adjust_stack(env, 1);
	
	snprintf(buf, sizeof buf, "%ld", v);
	pushs(env, buf);
}

/* Evaluation instructions */
void
instr_cmp(eval_environ_t env)
{
	long l = (long) get_arg(env, 1);
	long r = (long) get_arg(env, 0);
	adjust_stack(env, 2);
	if (prog_trace_option)
		prog_trace(env, "CMP %ld %ld", l, r);
	push(env, (STKVAL) (l == r));
}

void
instr_symbol(eval_environ_t env)
{
	char *symbol = get_immediate(env, 0);
	char *s = env->getsym(env->data, symbol);
	if (prog_trace_option)
		prog_trace(env, "SYMBOL %s", symbol);
	if (!s) {
		if (prog_trace_option)
			prog_trace(env, "%s is not defined", symbol);
		env->null_symbol = symbol;
		longjmp(env->x, 1);
	} else if (env->null_symbol && strcmp(env->null_symbol, symbol) == 0) 
		env->null_symbol = NULL;
	
	if (prog_trace_option)
		prog_trace(env, "%s dereferenced to %s", symbol, s);
		
	push(env, (STKVAL) s);
	advance_pc(env, 1);
}

void
dump_symbol(prog_counter_t i, prog_counter_t *ret __attribute__((__unused__)))
{
	printf("%s", (char *) prog[i]);
}

/* Comparation instructions */
void
instr_eqn(eval_environ_t env)
{
	long a = (long) get_arg(env, 1);
	long b = (long) get_arg(env, 0);
	if (prog_trace_option)
		prog_trace(env, "EQN %ld %ld", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (a == b));
}

void
instr_eqs(eval_environ_t env)
{
	char *a = (char*) get_arg(env, 1);
	char *b = (char*) get_arg(env, 0);
	if (prog_trace_option)
		prog_trace(env, "EQS %s %s", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (strcmp(a, b) == 0));
}

void
instr_nen(eval_environ_t env)
{
	long a = (long) get_arg(env, 1);
	long b = (long) get_arg(env, 0);
	if (prog_trace_option)
		prog_trace(env, "NEN %ld %ld", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (a != b));
}
	
void
instr_nes(eval_environ_t env)
{
	char *a = (char*) get_arg(env, 1);
	char *b = (char*) get_arg(env, 0);
	if (prog_trace_option)
		prog_trace(env, "NES %s %s", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) strcmp(a, b));
}

void
instr_ltn(eval_environ_t env)
{
	long a = (long) get_arg(env, 1);
	long b = (long) get_arg(env, 0);
	if (prog_trace_option)
		prog_trace(env, "LTN %ld %ld", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (a < b));
}

void
instr_lts(eval_environ_t env)
{
	char *a = (char*) get_arg(env, 1);
	char *b = (char*) get_arg(env, 0);
	if (prog_trace_option)
		prog_trace(env, "LTS %s %s", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (strcmp(a, b) < 0));
}

void
instr_len(eval_environ_t env)
{
	long a = (long) get_arg(env, 1);
	long b = (long) get_arg(env, 0);
	if (prog_trace_option)
		prog_trace(env, "LEN %ld %ld", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (a <= b));
}

void
instr_les(eval_environ_t env)
{
	char *a = (char*) get_arg(env, 1);
	char *b = (char*) get_arg(env, 0);
	if (prog_trace_option)
		prog_trace(env, "LES %s %s", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (strcmp(a, b) <= 0));
}

void
instr_gtn(eval_environ_t env)
{
	long a = (long) get_arg(env, 1);
	long b = (long) get_arg(env, 0);
	if (prog_trace_option)
		prog_trace(env, "GTN %ld %ld", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (a > b));
}

void
instr_gts(eval_environ_t env)
{
	char *a = (char*) get_arg(env, 1);
	char *b = (char*) get_arg(env, 0);
	if (prog_trace_option)
		prog_trace(env, "GTS %s %s", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (strcmp(a, b) > 0));
}

void
instr_gen(eval_environ_t env)
{
	long a = (long) get_arg(env, 1);
	long b = (long) get_arg(env, 0);
	if (prog_trace_option)
		prog_trace(env, "GEN %ld %ld", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (a >= b));
}

void
instr_ges(eval_environ_t env)
{
	char *a = (char*) get_arg(env, 1);
	char *b = (char*) get_arg(env, 0);
	if (prog_trace_option)
		prog_trace(env, "GES %s %s", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (strcmp(a, b) >= 0));
}

/* Jump instructions */
void
instr_bz(eval_environ_t env)
{
	long v = (long) get_arg(env, 0);
	long off = (long) get_immediate(env, 0);

	if (prog_trace_option)
		prog_trace(env, "BZ %ld (%ld)", off, v);
	adjust_stack(env, 1);
	if (v == 0)
		advance_pc(env, off);
	advance_pc(env, 1);
}

void
dump_branch (prog_counter_t i, prog_counter_t *ret __attribute__((__unused__)))
{
	printf("%ld (%ld)", (long) prog[i], i + (long) prog[i] + 1);
}

void
instr_bnz(eval_environ_t env)
{
	long v = (long) get_arg(env, 0);
	long off = (long) get_immediate(env, 0);

	if (prog_trace_option)
		prog_trace(env, "BNZ %ld (%ld)", off, v);
	adjust_stack(env, 1);
	if (v != 0)
		advance_pc(env, off);
	advance_pc(env, 1);
}

void
instr_jmp(eval_environ_t env)
{
	long off = (long) get_immediate(env, 0);
	if (prog_trace_option)
		prog_trace(env, "JMP %ld", off);
	advance_pc(env, off+1);
}

/* Boolean instructions */
void
instr_not(eval_environ_t env)
{
	long v = (long) get_arg(env, 0);
	if (prog_trace_option)
		prog_trace(env, "NOT %ld", v);
	adjust_stack(env, 1);
	push(env, (STKVAL) !v);
}

/* Arithmetic instructions */
void
instr_add(eval_environ_t env)
{
	long a = (long) get_arg(env, 1);
	long b = (long) get_arg(env, 0);
	adjust_stack(env, 2);
	if (prog_trace_option)
		prog_trace(env, "ADD %ld %ld", a, b); 
	push(env, (STKVAL) (a + b));
}

void
instr_sub(eval_environ_t env)
{
	long a = (long) get_arg(env, 1);
	long b = (long) get_arg(env, 0);
	adjust_stack(env, 2);
	if (prog_trace_option)
                prog_trace(env, "SUB %ld %ld", a, b); 
	push(env, (STKVAL) (a - b));
}

void
instr_mul(eval_environ_t env)
{
	long a = (long) get_arg(env, 1);
	long b = (long) get_arg(env, 0);
	adjust_stack(env, 2);
	if (prog_trace_option)
		prog_trace(env, "MUL %ld %ld", a, b); 
	push(env, (STKVAL) (a * b));
}

void
instr_div(eval_environ_t env)
{
	long a = (long) get_arg(env, 1);
	long b = (long) get_arg(env, 0);
	adjust_stack(env, 2);
	if (prog_trace_option)
		prog_trace(env, "DIV %ld %ld", a, b);
	if (b == 0) {
		if (env_catch(env, mf_divzero) == 0)
			return;
		runtime_error(env, "Division by zero");
	}
	push(env, (STKVAL) (a / b));
}

void
instr_neg(eval_environ_t env)
{
	long v = (long) get_arg(env, 0);
	if (prog_trace_option)
		prog_trace(env, "NEG %ld", v);
	adjust_stack(env, 1);
	push(env, (STKVAL) -v);
}

/* Matching and Regular expression instructions */
void
instr_regex(eval_environ_t env)
{
	int v;
	long flag = (long)get_arg(env, 0);
	regex_t *re = get_arg(env, 1);
	char *string = get_arg(env, 2);
	adjust_stack(env, 3);
	
	if (prog_trace_option)
		prog_trace(env, "REGEX %p %s", re, string);
	v = regexec(re, string, 0, NULL, 0);
	if (flag) {
		regfree(re);
		free(re);
	}
	push(env, (STKVAL) (v == 0));
}

void
instr_regcomp(eval_environ_t env)
{
	int v;
	regex_t *re;
	char *expr = get_arg(env, 0);
	adjust_stack(env, 1);

	if (prog_trace_option)
		prog_trace(env, "REGCOMP %s", expr);
	re = malloc(sizeof *re);
	if (!re)
		runtime_error(env, "Not enough memory");
	v = regcomp(re, expr, regex_flags);
	if (v) {
		char errbuf[512];
		regerror(v, re, errbuf, sizeof(errbuf));
		free(re);
		if (env_catch(env, mf_regcomp) == 0)
			return;
		runtime_error(env, "compiling regex `%s': %s",
			      expr,
			      errbuf);
	}	
	push(env, (STKVAL) re);
}

void
instr_fnmatch(eval_environ_t env)
{
	char *string = get_arg(env, 1);
	char *pattern = get_arg(env, 0);
	adjust_stack(env, 2);
	if (prog_trace_option)
                prog_trace(env, "FNMATCH %s %s", string, pattern);
	push(env, (STKVAL) (fnmatch (pattern, string, 0) == 0));
}

static int
mx_match(eval_environ_t env, char *string,
	 int (*matcher)(const char *name, void *data), void *data)
{
	int rc = 0;
	mxbuf_t mxbuf;
	mf_status mxstat;

	char *p = strchr(string, '@');
	if (p)
		p++;
	else
		p = string;
	mxstat = getmx(p, mxbuf);
	rc = 0;
	if (mxstat == mf_success) {
		int i;
		
		for (i = 0; i < MAXMXCOUNT && mxbuf[i]; i++) {
			if (matcher(mxbuf[i], data)) {
				rc = 1;
				break;
			}
		}
		freemx(mxbuf);
	} else
		env_catch(env, mxstat);
	return rc;
}

static int
fn_matcher(const char *string, void *data)
{
	return fnmatch (data, string, 0) == 0;
}

void
instr_fnmatch_mx(eval_environ_t env)
{
	char *string = get_arg(env, 1);
	char *pattern = get_arg(env, 0);
	adjust_stack(env, 2);

	if (prog_trace_option)
		prog_trace(env, "FNMATCH,MX %s %s", string, pattern);
	push(env, (STKVAL) mx_match(env, string, fn_matcher, pattern));
}

static int
regex_matcher(const char *string, void *data)
{
	return regexec(data, string, 0, NULL, 0) == 0;
}

void
instr_regex_mx(eval_environ_t env)
{
	int rc;
	long flag = (long)get_arg(env, 0);
	regex_t *re = get_arg(env, 1);
	char *string = get_arg(env, 2);
	adjust_stack(env, 3);
	
	if (prog_trace_option)
		prog_trace(env, "REGEX,MX %p %s", re, string);
	
	rc = mx_match(env, string, regex_matcher, re);
	
	if (flag) {
		regfree(re);
		free(re);
	}
	push(env, (STKVAL) rc);
}
	
/* Mail filter specific instructions */

void
instr_next(eval_environ_t env)
{
	if (prog_trace_option)
		prog_trace(env, "NEXT", env->locus.file, env->locus.line);
	trace("%s%s:%lu: next",
	      mailfromd_msgid(env->ctx),
	      env->locus.file, env->locus.line);
}

void
instr_result(eval_environ_t env)
{
	sfsistat status = (sfsistat) get_immediate(env, 0);
	char *code = (char *) get_immediate(env, 1);
	char *xcode = (char *) get_immediate(env, 2);
	char *message = (char *) get_arg(env, 0);

	if (prog_trace_option)
		prog_trace(env, "RESULT %d %s %s %s",
			   status,
			   SP(code),
			   SP(xcode),
			   SP(message));
	
	trace("%s%s:%lu: %s %s %s %s",
	      mailfromd_msgid(env->ctx),
	      env->locus.file, env->locus.line,
	      sfsistat_str(status),
	      SP(code),
	      SP(xcode),
	      SP(message));
	
	env->status = status;
	env->setreply(env->data, code, xcode, message);
	advance_pc(env, 3);
	adjust_stack(env, 1);
}

void
dump_result(prog_counter_t i, prog_counter_t *ret __attribute__((__unused__)))
{
	printf("%s %s %s",
	       sfsistat_str((sfsistat)prog[i]),
	       SP((char*) prog[i+1]),
	       SP((char*) prog[i+2]));
}

void
instr_header(eval_environ_t env)
{
	struct header_node *hdr = emalloc (sizeof(*hdr));

	hdr->opcode = (enum header_opcode) get_immediate(env, 0);
	hdr->name = (char *) get_immediate(env, 1);
	hdr->value = (char *) get_immediate(env, 2);
	advance_pc(env, 3);
	
	if (prog_trace_option)
		prog_trace(env, "HEADER %d %s %s",
			   hdr->opcode, hdr->name, hdr->value);
	
	trace("%s%s:%lu: %s %s %s",
	      mailfromd_msgid(env->ctx),
	      env->locus.file, env->locus.line,
	      header_command_str(hdr->opcode),
	      hdr->name, hdr->value);
	
	env->setheader(env->data, hdr);
}

void
dump_header(prog_counter_t i, prog_counter_t *ret __attribute__((__unused__)))
{
	printf("%s %s %s",
	       header_command_str((enum header_opcode) prog[i]),
	       SP((char*) prog[i+1]), SP((char*) prog[i+2]));
}

void
instr_builtin(eval_environ_t env)
{
	char *name = get_immediate(env, 0);
	void (*handler)(eval_environ_t) = get_immediate(env, 1);

	if (prog_trace_option)
		prog_trace(env, "BUILTIN %s", name);
	advance_pc(env, 2);
	handler(env);
}

void
dump_builtin(prog_counter_t i, prog_counter_t *ret __attribute__((__unused__)))
{
	printf("%s ", SP((char*) prog[i]));
}

void
instr_concat(eval_environ_t env)
{
	char *left = get_arg(env, 1);
	char *right = get_arg(env, 0);
	char *res = heap_reserve(env, strlen(left) + strlen(right) + 1);

	strcat(strcpy(res, left), right);
	adjust_stack(env, 2);
	if (prog_trace_option)
		prog_trace(env, "CONCAT ('%s','%s')='%s'", left, right, res);
	push(env, (STKVAL) res);
}

void
instr_variable(eval_environ_t env)
{
	size_t off = (size_t) get_immediate(env, 0);
	advance_pc(env, 1);
	if (prog_trace_option)
		prog_trace(env, "VARIABLE %lu", (unsigned long) env->var[off]);
	push(env, (STKVAL) env->var[off]);
}

void
dump_variable(prog_counter_t i,
	      prog_counter_t *ret __attribute__((__unused__)))
{
	printf("%lu ", (unsigned long) prog[i]);
}

void
instr_asgn(eval_environ_t env)
{
	STKVAL val = get_arg(env, 0);
	size_t off = (size_t) get_immediate(env, 0);
	advance_pc(env, 1);
	adjust_stack(env, 1);
	if (prog_trace_option)
		prog_trace(env, "ASGN %lu %lu",
			   (unsigned long) off, (unsigned long) val);
	env->var[off] = val;
}

void
instr_svalidate(eval_environ_t env)
{
	STKVAL val = get_arg(env, 0);
	if (prog_trace_option)
		prog_trace(env, "SVALIDATE %p", val);
	if (val == 0) {
		static char nullstr[] = "";
		env->stack[env->tos - 1] = (STKVAL)nullstr;
	}
}		

void
instr_arg(eval_environ_t env)
{
	size_t n = (size_t) get_immediate(env, 0);
	advance_pc(env, 1);
	if (prog_trace_option)
		prog_trace(env, "ARG %lu = %p", 
			   (unsigned long) n, env->stack[env->base - 2 - n]);
	push(env, (STKVAL) env->stack[env->base - 2 - n]);
}

void
instr_catch(eval_environ_t env)
{
	long off = (long) get_immediate(env, 0);
	/* int count = (long) get_immediate(env, 1); -- Count is not used */
	unsigned mask = (unsigned) get_immediate(env, 2);
	int i;
	prog_counter_t entry = env->pc + 4;
	
	if (prog_trace_option) 
		prog_trace(env, "CATCH %ld, %ld", entry, off);

	for (i = 0; i < mf_status_count; i++)
		if (mask & EXMASK(i)) {
			if (prog_trace_option)
				prog_trace(env, "CATCH TARGET: %s",
					   mf_status_str((mf_status)i));
			env->catch[i] = entry;
		}

	advance_pc(env, off);
}

void
dump_catch(prog_counter_t i, prog_counter_t *reti)
{
	unsigned mask = (unsigned) prog[i+2];
	printf("%ld (%ld)", (long) prog[i], i + (long) prog[i]);
	printf("; Targets: ");

	for (i = 0; i < mf_status_count; i++)
		if (mask & EXMASK(i)) 
			printf("%s ", mf_status_str((mf_status)i));
}

void
instr_echo(eval_environ_t env)
{
	char *str = (char*) get_arg(env, 0);
	logmsg(LOG_INFO, "%s", str);	
}

void
instr_return(eval_environ_t env)
{
	if (prog_trace_option)
		prog_trace(env, "RETURN");
	env_leave_frame(env, 0);
	env->pc--;
}

void
instr_retcatch(eval_environ_t env)
{
	prog_counter_t pc = env->pc;
	if (prog_trace_option)
		prog_trace(env, "RETCATCH");
	env_leave_frame(env, 0);
	env->pc = pc;
}

void
instr_saveex(eval_environ_t env)
{
	unsigned exmask = (unsigned) get_immediate(env, 0);
	int i;
	advance_pc(env, 1);
	if (prog_trace_option)
		prog_trace(env, "SAVEEX %x", exmask);
	push(env, (STKVAL) exmask);
	for (i = 0; i < mf_status_count; i++) {
		if (EXMASK(i) & exmask) 
			push(env, (STKVAL) env->catch[i]);
	}
}

static void
dump_saveex(prog_counter_t i, prog_counter_t *ret)
{
	printf("%x", (unsigned)prog[i]);
}

void
instr_restex(eval_environ_t env)
{
	unsigned exmask = (unsigned)env->stack[env->base];
	int i;
	for (i = 0; i < mf_status_count; i++) {
		if (EXMASK(i) & exmask)
			env->catch[i] = (prog_counter_t)
				           env->stack[env->base+i+1];
	}
}

void
instr_adjust(eval_environ_t env)
{
	long nargs = (long) get_immediate(env, 0);
	if (prog_trace_option)
		prog_trace(env, "ADJUST %ld", nargs);
	adjust_stack(env, nargs);
	advance_pc(env, 1);
}

void
instr_popreg(eval_environ_t env)
{
	env->reg = pop(env);
	if (prog_trace_option)
		prog_trace(env, "POPREG %p", env->reg);
}

void
instr_pushreg(eval_environ_t env)
{
	if (prog_trace_option)
		prog_trace(env, "PUSHREG %p", env->reg);
	push(env, env->reg);
}

void
instr_funcall(eval_environ_t env)
{
	char *name = get_immediate(env, 0);
	prog_counter_t pc = (prog_counter_t) get_immediate(env, 1);
	advance_pc(env, 2);
	if (prog_trace_option)
		prog_trace(env, "FUNCALL %s (%lu)", name, (unsigned long)pc);
	env_make_frame(env);
	env->pc = pc-1;
}

void
dump_funcall(prog_counter_t i, prog_counter_t *ret __attribute__((__unused__)))
{
	printf("%s (%lu)", SP((char*) prog[i]), (unsigned long)prog[i+1]);
}

/* Opcode:

   xlat N x y x1 y1 ... xN yN

   Synopsis:

   Scan the table until xI==REG is found or the table is exhausted.
   If found, replace REG with yI and return 0. Otherwise, return 1
*/
void
instr_xlat(eval_environ_t env)
{
	long i, count = (long) get_immediate(env, 0);
	instr_t *tab = prog + env->pc + 2;
	count *= 2;
	advance_pc(env, count+1);
        if (prog_trace_option)
		prog_trace(env, "XLAT %ld", count);
	for (i = 0; i < count; i+=2) {
		if ((long) tab[i] == (long) env->reg) {
			env->reg = (STKVAL) tab[i+1];
			push(env, (STKVAL)0);
			return;
		}
	}
	push(env, (STKVAL)1);
}

void
instr_xlats(eval_environ_t env)
{
	long i, count = (long) get_immediate(env, 0);
	instr_t *tab = prog + env->pc + 2;
	char *str = (char*) env->reg;
	count *= 2;
	advance_pc(env, count+1);
        if (prog_trace_option)
		prog_trace(env, "XLATS %ld", count);
	for (i = 0; i < count; i+=2) {
		if (strcmp((char*) tab[i], str) == 0) {
			env->reg = (STKVAL) tab[i+1];
			push(env, (STKVAL)0);
			return;
		}
	}
	push(env, (STKVAL)1);
}

void
dump_xlat(prog_counter_t i, prog_counter_t *ret)
{
	long j, count = (long) prog[i];
	count *= 2;
	*ret = i + count;
	printf("%ld ", (long) prog[i]);
	i++;
	for (j = 0; j < count; j+=2) 
		printf("(%ld %ld) ", (long)prog[i+j], (long)prog[i+j+1]);
}

void
dump_xlats(prog_counter_t i, prog_counter_t *ret)
{
	long j, count = (long) prog[i];
	count *= 2;
	*ret = i + count;
	printf("%ld ", (long) prog[i]);
	i++;
	for (j = 0; j < count; j+=2) 
		printf("(%s %ld) ", (char*)prog[i+j], (long)prog[i+j+1]);
}

void
instr_jreg(eval_environ_t env)
{
	env->pc += (prog_counter_t)env->reg - 1;
	if (prog_trace_option)
		prog_trace(env, "JREG %ld (%lu)",
			   (long)((prog_counter_t)env->reg - 1),
			   (unsigned long)env->pc);
}


struct optab {
	char *name;
	instr_t instr;
	void (*dump) (prog_counter_t i, prog_counter_t *p);
	long length;
} optab[] = {
	{ "XCHG",           instr_xchg,           NULL,                  0 },
        { "STON",           instr_ston,           NULL,                  0 },
        { "NTOS",           instr_ntos,           NULL,                  0 },
        { "LOCUS",          instr_locus,          dump_locus,            2 },
        { "PUSH_IMMEDIATE", instr_push_immediate, dump_push_immediate,   1 },
        { "SYMBOL",         instr_symbol,         dump_symbol,           1 },
        { "BZ",             instr_bz,             dump_branch,           1 },
        { "BNZ",            instr_bnz,            dump_branch,           1 },
        { "JMP",            instr_jmp,            dump_branch,           1 },
        { "DUP",            instr_dup,            NULL,                  0 },
        { "CMP",            instr_cmp,            NULL,                  0 },
        { "POP",            instr_pop,            NULL,                  0 },
        { "REGEX",          instr_regex,          NULL,                  0 },
        { "REGEX,MX",       instr_regex_mx,       NULL,                  0 },
        { "REGCOMP",        instr_regcomp,        NULL,                  0 },
        { "FNMATCH",        instr_fnmatch,        NULL,                  0 },
        { "FNMATCH,MX",     instr_fnmatch_mx,     NULL,                  0 },
        { "NOT",            instr_not,            NULL,                  0 },
        { "NEXT",           instr_next,           NULL,                  0 },
        { "RESULT",         instr_result,         dump_result,           3 },
        { "HEADER",         instr_header,         dump_header,           3 },
        { "EQN",            instr_eqn,            NULL,                  0 },
        { "EQS",            instr_eqs,            NULL,                  0 },
        { "NEN",            instr_nen,            NULL,                  0 },
        { "NES",            instr_nes,            NULL,                  0 },
	{ "LTN",            instr_ltn,            NULL,                  0 },
	{ "LTS",            instr_lts,            NULL,                  0 },
	{ "LEN",            instr_len,            NULL,                  0 },
	{ "LES",            instr_les,            NULL,                  0 },
	{ "GTN",            instr_gtn,            NULL,                  0 },
	{ "GTS",            instr_gts,            NULL,                  0 },
	{ "GEN",            instr_gen,            NULL,                  0 },
	{ "GES",            instr_ges,            NULL,                  0 },
	{ "ADD",            instr_add,            NULL,                  0 },
	{ "SUB",            instr_sub,            NULL,                  0 },
	{ "MUL",            instr_mul,            NULL,                  0 },
	{ "DIV",            instr_div,            NULL,                  0 },
	{ "BUILTIN",        instr_builtin,        dump_builtin,          2 },
	{ "CONCAT",         instr_concat,         NULL,                  0 },
	{ "SET",            instr_asgn,           dump_variable,         1 },
	{ "VARIABLE",       instr_variable,       dump_variable,         1 },
	{ "SVALIDATE",      instr_svalidate,      NULL,                  0 },
	{ "ARG",            instr_arg,            dump_variable,         1 },
	{ "CATCH",          instr_catch,          dump_catch,            3 },
	{ "ECHO",           instr_echo,           NULL,                  0 },
	{ "RETURN",         instr_return,         NULL,                  0 },
	{ "RETCATCH",       instr_retcatch,       NULL,                  0 }, 
	{ "ADJUST",         instr_adjust,         dump_variable,         1 },
	{ "POPREG",         instr_popreg,         NULL,                  0 },
	{ "PUSHREG",        instr_pushreg,        NULL,                  0 },
	{ "FUNCALL",        instr_funcall,        dump_funcall,          2 },
	{ "JREG",           instr_jreg,           NULL,                  0 },
	{ "XLAT",           instr_xlat,           dump_xlat,             0 },
	{ "XLATS",          instr_xlats,          dump_xlats,            0 },
	{ "SAVEEX",         instr_saveex,         dump_saveex,           1 },
	{ "RESTEX",         instr_restex,         NULL,                  0 },
	{ NULL }
};

struct optab *
find_op(instr_t instr)
{
	struct optab *op;
	for (op = optab; op->instr; op++)
		if (op->instr == instr)
			return op;
	return NULL;
}

void
dump_code(prog_counter_t start, prog_counter_t end)
{
	prog_counter_t i;
	if (end == 0)
		end = pc;
	for (i = start; i < end; i++) {
		printf("%4lu: ", i);
		if (prog[i]) {
			struct optab *op = find_op (prog[i]);
			if (!op) {
				printf("Unknown function. Aborting dump\n");
				exit (1);
			}
			printf("%s ", op->name);
			if (op->dump)
				op->dump (i+1, &i);
			i += op->length;
		} else
			printf("STOP");
		putchar('\n');
	}
}


struct declvar {
	struct declvar *next;
	struct variable *var;
	STKVAL value;
};

static struct declvar *declvar;

static struct declvar *
find_declared_variable(const char *name)
{
	struct declvar *dv;
	for (dv = declvar; dv; dv = dv->next)
		if (strcmp(dv->var->name, name) == 0)
			return dv;
	return NULL;
}

static void
declare_variable(struct variable *var, STKVAL value)
{
	struct declvar *dv = emalloc(sizeof *dv);
	dv->next = declvar;
	declvar = dv;
	dv->var = var;
	dv->value = value;
}

void
declare_numeric_variable(struct variable *var, long val)
{
	declare_variable(var, (STKVAL) val);
	var->type = rettype_number;
}

void
declare_string_variable(struct variable *var, char *val)
{
	declare_variable(var, (STKVAL) val);
	var->type = rettype_string;
}


void
env_init(eval_environ_t env)
{
	int i;

	/* Initialize status and registers */
	env->status = SMFIS_CONTINUE;
	env->tos = 0;
	env->toh = stack_size - 1;
	env->base = 0;
	env->reg = 0;
	
	/* Initialize catch functions */
	for (i = 0; i < mf_status_count; i++)
		env->catch[i] = 0;
}

/* Initialize declared variables */
void
env_decl_vars(eval_environ_t env)
{
	struct declvar *dv;

	for (dv = declvar; dv; dv = dv->next)
		env->var[dv->var->off] = dv->value;
}

void
env_make_frame(eval_environ_t env)
{
	push(env, (STKVAL) (env->pc + 1));
	push(env, (STKVAL) env->base);
	env->base = env->tos;
}

void
env_leave_frame(eval_environ_t env, int nargs)
{
	env->tos = env->base;
	env->base = (prog_counter_t) pop(env);
	env->pc = (prog_counter_t) pop(env);
	adjust_stack(env, nargs);
}
	
void
env_push_string(eval_environ_t env, char *arg)
{
	push(env, (STKVAL) arg);
}

void
env_push_number(eval_environ_t env, long arg)
{
	push(env, (STKVAL) arg);
}

int
env_set_variable(eval_environ_t env, char *ident, char *value)
{
	char *end;
	struct variable *var = variable_lookup(ident);
	if (!var) {
		mu_error("variable %s is not declared");
		return 1;
	}
	switch (var->type) {
	case rettype_string:
		env->var[var->off] = (STKVAL) strdup(value);
		break;

	case rettype_number:
		env->var[var->off] = (STKVAL) strtol(value, &end, 0);
		if (*end) {
			mu_error("%s is not a number", value);
			return 1;
		}
		break;
		
	default:
		mu_error("type of %s is unknown", ident);
		return 1;
	}
	return 0;
}

int
eval_environment(eval_environ_t env, prog_counter_t start)
{
	if (setjmp(env->x)) 
		return 1;

	for (env->pc = start; prog[env->pc]; env->pc++) {
		if (env->pc >= pc)
			runtime_error(env, "pc out of range");
		(*(prog[env->pc]))(env);
	}
	return 0;
}

int
env_catch(eval_environ_t env, mf_status status)
{
	prog_counter_t pc = env->catch[status];
	if (pc) {
		/* Reset the exception to avoid recursion. We will never have
		   to restore it again, since the catch handler either returns
		   from the function, in which case the exception mask will be
		   restored, or stops the execution with an action, in which
		   case exceptions are not relevant anymore. */
		env->catch[status] = 0;
		/* Fixup the program counter */
		env->pc = pc - 1;
		/* Generate normal entry frame */
		env_push_number(env, status);
		env_make_frame(env);
		return 0;
	}
	return 1;
}

sfsistat
environment_get_status(eval_environ_t env)
{
	return env->status;
}

const char *
environment_get_null_symbol(eval_environ_t env, struct locus *locus)
{
	*locus = env->locus;
	return env->null_symbol;
}


/* Rudimentary dictionary support */
struct dict_entry {
	char *name;
	char *value;
};

static int
name_comp(const void *item, const void *data)
{
	const struct dict_entry *ent = item;
	const struct dict_entry *p = data;
	return strcmp(ent->name, p->name);
}

void
name_destroy(void *item)
{
	const struct dict_entry *ent = item;
	free(ent->name);
	free(ent->value);
}

void
dict_init(mu_list_t *dict)
{
	mu_list_create(dict);
	mu_list_set_comparator(*dict, name_comp);
	mu_list_set_destroy_item (*dict, name_destroy);
}

char *
dict_install(mu_list_t dict, char *name, char *value)
{
	struct dict_entry ent, *p;

	ent.name = name;

	if (mu_list_locate(dict, &ent, (void**)&p) == 0) {
		free(p->value);
		p->value = strdup(value);
	} else {
		p = malloc(sizeof(*p));
		if (!p)
			abort();
		p->name = strdup(name);
		p->value = strdup(value);
		mu_list_append(dict, p);
	}
	return p->value;
}

void
dict_destroy(mu_list_t *dict)
{
	mu_list_destroy(dict);
}

char *
dict_getsym(void *data, char *str)
{
	struct dict_entry ent, *p;
	char *tmp = NULL;
	char *val = NULL;
	
	if (str[0] == '{') {
		int len = strlen(str);
		if (str[len-1] == '}') {
			tmp = malloc(len-1);
			if (!tmp)
				abort();
			memcpy(tmp, str+1, len-2);
			tmp[len-2] = 0;
			str = tmp;
		}
	}
	
	ent.name = str;
	if (mu_list_locate(data, &ent, (void**)&p) == 0)
		val = p->value;
	free(tmp);
	return val;
}

eval_environ_t
create_environment(SMFICTX *ctx,
		   char *(*getsym)(void *data, char *str),
		   int (*setreply)(void *data, char *code, char *xcode,
				   char *message),
		   void (*setheader)(void *data, struct header_node *hdr),
		   void *data)
{
	struct eval_environ *env = calloc(1, sizeof *env);

	if (!env) {
		mu_error("Not enough memory");
		exit(1);
	}

	env->stack = calloc(stack_size, sizeof env->stack[0]);
	if (!env->stack) {
		mu_error("Not enough memory");
		exit(1);
	}

	env->var = calloc(variable_count, sizeof env->var[0]);
	
	env->ctx = ctx;
	env->data = data;
	env->getsym = getsym;
	env->setreply = setreply;
	env->setheader = setheader;
	env->status = SMFIS_CONTINUE;
	dict_init(&env->dict);
	env_decl_vars(env);
	
	return env;
}

void
destroy_environment(eval_environ_t env)
{
	free(env->stack);
	free(env->var);
	dict_destroy(&env->dict);
	free(env);
}


/* Builtins */

/* Locations of the internal variables in the data segment */

size_t greylist_seconds_left_loc;
size_t rcpt_count_loc;
struct variable *ehlo_domain_var;
struct variable *mailfrom_address_var;

/* Functions to access %rcpt_count */
unsigned long
get_rcpt_count(eval_environ_t env)
{
	return (long) env->var[rcpt_count_loc];
}

void
clear_rcpt_count(eval_environ_t env)
{
	env->var[rcpt_count_loc] = 0;
}

void
incr_rcpt_count(eval_environ_t env)
{
	env->var[rcpt_count_loc]++;
}


/* Built-in functions */

void
bi_toupper(eval_environ_t env)
{
	char *string = get_arg(env, 0);
	char *p;
	for (p = string; *p; p++)
		*p = toupper(*p);
	push(env, string);
}

void
bi_tolower(eval_environ_t env)
{
	char *string = get_arg(env, 0);
	char *p;
	for (p = string; *p; p++)
		*p = tolower(*p);
	push(env, string);
}


void
bi_hostname(eval_environ_t env)
{
	char *s;
	char *string = get_arg(env, 0);
	adjust_stack(env, 1);
	
	if (prog_trace_option)
		prog_trace(env, "HOSTNAME %s", string);

	s = dict_getsym(env->dict, string);
	if (!s) {
		char hbuf[NI_MAXHOST];
		if (resolve_ipstr(string, hbuf, sizeof(hbuf)) == 0) {
			s = dict_install(env->dict, string, hbuf);
		} else
			s = string;
	}
	push(env, s);	
}

void
bi_resolve(eval_environ_t env)
{
	char *s;
	struct hostent *host;
	char *string = get_arg(env, 0);

	adjust_stack(env, 1);
	
	if (prog_trace_option)
		prog_trace(env, "RESOLVE %s", string);

	s = dict_getsym(env->dict, string);
	if (!s) {
		host = gethostbyname(string);
		if (!host)
			s = "0";
		else if (host->h_length == 4) {
			struct in_addr addr;
			addr.s_addr = *(unsigned long*)host->h_addr;
			s = inet_ntoa(addr);
			s = dict_install(env->dict, string, s);
		} else
			s = "resolved";
	}
	push(env, s);	
}

void
bi_hasmx(eval_environ_t env)
{
	mxbuf_t mxbuf;
	mf_status mxstat;
	char *string = get_arg(env, 0);

	adjust_stack(env, 1);
	
	if (prog_trace_option)
		prog_trace(env, "HASMX %s", string);

	mxstat = getmx(string, mxbuf);
	
	if (mxstat != mf_success) {
		if (env_catch(env, mxstat) == 0)
			return;
	} else
		freemx(mxbuf);

	push(env, (STKVAL)(mxstat == mf_success));
}

void
bi_relayed(eval_environ_t env)
{
	char *s = get_arg(env, 0);
	adjust_stack(env, 1);
	if (prog_trace_option)
		prog_trace(env, "RELAYED %s", s);	
	push(env, (STKVAL) relayed_domain_p(s));
}

/* Args:
   stdpoll(email, ehlo, mailfrom)
*/
void
bi_stdpoll(eval_environ_t env)
{
	char *email = get_arg(env, 2);
	char *ehlo = get_arg(env, 1);
	char *mailfrom = get_arg(env, 0);
	mf_status rc;

	if (prog_trace_option)
		prog_trace(env, "STDPOLL %s %s %s", email, ehlo, mailfrom);
	
	adjust_stack(env, 3);
	
	trace("%s%s:%lu: STANDARD POLL FOR %s FROM %s AS <%s>",
	      mailfromd_msgid(env->ctx),
	      env->locus.file, env->locus.line,
	      email, ehlo, mailfrom);
	rc = method_standard(env->ctx,
			     email, ehlo, mailfrom);
	if (!mf_resolved (rc)) {
		if (env_catch(env, rc) == 0)
			return;
		else
			runtime_error(env, "unhandled exception %s",
				      mf_status_str(rc));
	}
			
	push(env, (STKVAL) rc);	
}

/* Args:
   strictpoll(host, email, ehlo, mailfrom)
*/
void
bi_strictpoll(eval_environ_t env)
{
	char *email = get_arg(env, 3);
	char *ehlo = get_arg(env, 2);
	char *mailfrom = get_arg(env, 1);
	char *host = get_arg(env, 0);
	mf_status rc;

	if (prog_trace_option)
		prog_trace(env, "STRICTPOLL %s %s %s %s",
			   host, email, ehlo, mailfrom);
	
	adjust_stack(env, 4);
	
	trace("%s%s:%lu: STRICT POLL FOR %s HOST %s FROM %s AS <%s>",
	      mailfromd_msgid(env->ctx),
	      env->locus.file, env->locus.line,
	      email, host,
	      ehlo, mailfrom);

	rc = method_strict(env->ctx, email, host, ehlo, mailfrom);
	
	if (!mf_resolved (rc)) {
		if (env_catch(env, rc) == 0)
			return;
		else
			runtime_error(env, "unhandled exception %s",
				      mf_status_str(rc));
	}
	
	push(env, (STKVAL) rc);	
}

void
bi__pollhost(eval_environ_t env)
{
	char *host = get_arg(env, 3);
	char *email = get_arg(env, 2);
	char *ehlo = get_arg(env, 1);
	char *mailfrom = get_arg(env, 0);
	mf_status rc;

	if (prog_trace_option)
		prog_trace(env, "POLLHOST %s %s %s %s",
			   host, email, ehlo, mailfrom);
	
	adjust_stack(env, 4);
	
	trace("%s%s:%lu: POLLHOST %s FOR %s FROM %s AS <%s>",
	      mailfromd_msgid(env->ctx),
	      env->locus.file, env->locus.line,
	      email, host,
	      ehlo, mailfrom);

	rc = check_on_host(env->ctx, email, host, ehlo, mailfrom);

	if (!mf_resolved (rc)) {
		if (env_catch(env, rc) == 0)
			return;
		else
			runtime_error(env, "unhandled exception %s",
				      mf_status_str(rc));
	}
	push(env, (STKVAL) rc);	
}
	
void
bi__pollmx(eval_environ_t env)
{
	char *domain = get_arg(env, 3);
	char *email = get_arg(env, 2);
	char *ehlo = get_arg(env, 1);
	char *mailfrom = get_arg(env, 0);
	mf_status rc;

	if (prog_trace_option)
		prog_trace(env, "POLLMX %s %s %s %s",
			   domain, email, ehlo, mailfrom);
	
	adjust_stack(env, 4);
	
	trace("%s%s:%lu: POLLMX %s FOR %s FROM %s AS <%s>",
	      mailfromd_msgid(env->ctx),
	      env->locus.file, env->locus.line,
	      domain, email, ehlo, mailfrom);

	rc = check_mx_records(env->ctx, email, domain, ehlo, mailfrom);

	if (!mf_resolved (rc)) {
		if (env_catch(env, rc) == 0)
			return;
		else
			runtime_error(env, "unhandled exception %s",
				      mf_status_str(rc));
	}
	push(env, (STKVAL) rc);	
}
	

/* FIXME: Retained for a moment for backward compatibility */
void
bi_numrcpt(eval_environ_t env)
{
	long n = (long) get_arg(env, 0);
	if (prog_trace_option)
		prog_trace(env, "NUMRCPT(%ld), %lu",
			   n, (long) env->var[rcpt_count_loc]);
	adjust_stack(env, 1);
	push(env, (STKVAL)((long) env->var[rcpt_count_loc] > n));
}
	
void
bi_listens(eval_environ_t env)
{
	char *s = get_arg(env, 0);
	adjust_stack(env, 1);
	if (prog_trace_option)
		prog_trace(env, "LISTENS %s", s);
	
	push(env, (STKVAL)(listens_on (s, 25) == mf_success));
}

static int
valid_user_p(const char *name)
{
	int rc;
	struct mu_auth_data *auth_data;

	auth_data = mu_get_auth_by_name(name);
	rc = auth_data != NULL;
	mu_auth_data_free(auth_data);
	debug2(10, "Checking user %s: %s", name,
	       rc ? "true" : "false");
	return rc;
}

void
bi_validuser(eval_environ_t env)
{
	char *name = get_arg(env, 0);
	adjust_stack(env, 1);
	if (prog_trace_option)
		prog_trace(env, "VALIDUSER %s", name);
	
	push(env, (STKVAL) valid_user_p(name));
}

void
bi_match_cidr(eval_environ_t env)
{
	char *ipstr = get_arg(env, 1);
	char *cidrstr = get_arg(env, 0);
	char *p;
	struct in_addr addr, cidr;
	char netbuf[16], *netstr;
	unsigned long netmask;
	int result = 0;
	
	adjust_stack(env, 2);

	if (prog_trace_option)
		prog_trace(env, "MATCH_CIDR %s %s" , ipstr, cidrstr);
	do {
		if (inet_aton(ipstr, &addr) == 0) {
			if (env_catch(env, mf_invip) == 0)
				return;
			runtime_error(env, "invalid IP address (%s)",
				      ipstr);
		}
		p = strchr(cidrstr, '/');
		if (!p) {
			netstr = cidrstr;
			netmask = 0xfffffffful;
		} else {
			size_t len = p - cidrstr;
			unsigned long n;
			
			if (len >= sizeof netbuf) {
				if (env_catch(env, mf_invcidr) == 0)
					return;
				runtime_error(env, "invalid CIDR (%s)",
					      cidrstr);
			} 
			memcpy(netbuf, cidrstr, len);
			netbuf[len] = 0;
			netstr = netbuf;
			n = strtoul(p + 1, &p, 10);
			if (*p || n > 32) {
				if (env_catch(env, mf_invcidr) == 0)
					return;
				runtime_error(env,
					      "invalid CIDR (netmask of %s)",
					      cidrstr);
			}
			n = 32 - n;
			if (n == 32)
				netmask = 0;
			else
				netmask = (0xfffffffful >> n) << n;
		}
		if (inet_aton(netstr, &cidr) == 0) {
			if (env_catch(env, mf_invcidr) == 0)
				return;
			runtime_warning(env,
					"invalid CIDR (network part of %s)",
					cidrstr);
			break;
		}

		addr.s_addr = ntohl(addr.s_addr);
		cidr.s_addr = ntohl(cidr.s_addr);
		debug5(100, "%s:%lu: addr=%lx net=%lx mask=%lx" ,
		       env->locus.file, env->locus.line,
		       addr.s_addr, cidr.s_addr, netmask);
		result = (addr.s_addr & netmask) == cidr.s_addr;
	} while (0);
	push(env, (STKVAL) result);
}

void
bi_domainpart(eval_environ_t env)
{
	char *str = get_arg(env, 0);
	char *p = strchr(str, '@');
	adjust_stack(env, 1);
	if (prog_trace_option)
		prog_trace(env, "DOMAINPART %s" , str);
	pushs(env, p ? p+1 : str);
}

void
bi_localpart(eval_environ_t env)
{
	char *str = get_arg(env, 0);
	char *p = strchr(str, '@');
	adjust_stack(env, 1);
	if (prog_trace_option)
		prog_trace(env, "LOCALPART %s" , str);

	if (p) {
		size_t size = p - str;
		char *string_space = heap_reserve(env, size);
		memcpy(string_space, str, size);
		string_space[size] = 0;
		push(env, string_space);
	} else
		pushs(env, str);
}



void
bi_interval(eval_environ_t env)
{
       time_t t;
       char *endp;
       char *str = get_arg(env, 0);
       adjust_stack(env, 1);
       if (prog_trace_option)
	       prog_trace(env, "INTERVAL %s", str);
       if (parse_time_interval(str, &t, &endp)) {
	       if (env_catch(env, mf_invtime) == 0)
		       return;
               runtime_error(env,
                             "unrecognized time format (near `%s')",
                             endp);
       }
       push(env, (STKVAL) t);
}

void
bi_rate(eval_environ_t env)
{
	double rate;
	char *key = get_arg(env, 1);
	long interval = (long) get_arg(env, 0);
	long lrate;
		
	adjust_stack(env, 2);

	if (get_rate(key, &rate) != mf_success) {
		if (env_catch(env, mf_dbfailure) == 0)
			return;
		runtime_error(env, "cannot get rate for %s", key);
	}
	lrate = rate * interval;
	
	if (prog_trace_option)
		prog_trace(env, "RATE %s %ld = %ld", key, interval, lrate);
	
	push(env, (STKVAL) lrate);
}

#ifdef USE_DBM

static int
dbmap_lookup_p(eval_environ_t env, char *dbname, const char *email)
{
	int rc;
	char *name;
	DBM_FILE db;
	DBM_DATUM key;
	DBM_DATUM contents;

	if (mu_dbm_open(dbname, &db, MU_STREAM_READ, 0)) {
		if (env_catch(env, mf_dbfailure) == 0)
			return -1;
		else
			runtime_error(env, "mu_dbm_open(%s) failed: %s", dbname,
				      mu_strerror(errno));
	}
	name = strdup(email);

	memset (&key, 0, sizeof key);
	memset (&contents, 0, sizeof contents);
	MU_DATUM_PTR (key) = name;
	MU_DATUM_SIZE (key) = strlen(name)+1;
	rc = mu_dbm_fetch(&db, key, &contents) == 0;
	debug2(10, "Checking alias %s: %s", name,
	       rc ? "true" : "false");
	mu_dbm_datum_free(&contents);
	mu_dbm_close(&db);
	return rc;
}

void
bi_dbmap(eval_environ_t env)
{
	int rc;
	char *dbname = get_arg(env, 1);
	char *key = get_arg(env, 0);
	adjust_stack(env, 2);
	if (prog_trace_option)
		prog_trace(env, "DBMAP %s %s" , dbname, key);
	rc = dbmap_lookup_p(env, dbname, key);
	if (rc != -1)
		push(env, (STKVAL) rc);
}

static void
greylist_print_item(const char *key, size_t size, const void *content)
{
	time_t timestamp = *(time_t*) content;
	size--; /* Size includes the trailing nul */
	printf("%*.*s ", size, size, key);
	format_time_str(stdout, timestamp);
	putchar('\n');
}

static struct db_format greylist_format_struct = {
	"greylist",
	DEFAULT_GREYLIST_DATABASE,
	DEFAULT_EXPIRE_INTERVAL,
	greylist_print_item,
	NULL
};

struct db_format *greylist_format = &greylist_format_struct;

/* greylist(key, interval)

   Returns true if the key is greylisted, false if it's OK to
   deliver mail.
 */
void
bi_greylist(eval_environ_t env)
{
	int rc;
	DBM_FILE db;
	DBM_DATUM key;
	DBM_DATUM contents;
	int readonly;
	time_t now;
	char *email = get_arg(env, 1);
	long interval = (long) get_arg(env, 0);

	adjust_stack(env, 4);

	if (prog_trace_option)
		prog_trace(env, "GREYLIST %s %ld %ld", email, interval);

	if (mu_dbm_open(greylist_format->dbname, &db, MU_STREAM_RDWR, 0600)) {
		if (env_catch(env, mf_dbfailure) == 0)
			return;
		runtime_error(env, "mu_dbm_open(%s) failed: %s",
			      greylist_format->dbname, mu_strerror(errno));
	}
	readonly = mu_dbm_lock(&db);
	
	memset(&key, 0, sizeof key);
	memset(&contents, 0, sizeof contents);
	MU_DATUM_PTR(key) = email;
	MU_DATUM_SIZE(key) = strlen(email)+1;

	time(&now);
	if (mu_dbm_fetch(&db, key, &contents) == 0) {
		time_t timestamp, diff;

		if (MU_DATUM_SIZE(contents) != sizeof timestamp) {
			if (env_catch(env, mf_dbfailure) == 0)
				return;
			runtime_error(env,
				    "greylist database %s has wrong data size",
				      greylist_format->dbname);
		}
		
		timestamp = *(time_t*) MU_DATUM_PTR(contents);
		diff = now - timestamp;
	
		__DBG(20) {
			char timebuf[32];
			debug_log("%s entered greylist database on %s, "
				  "%ld seconds ago",
				  email,
				  mailfromd_timestr(timestamp, timebuf,
						    sizeof timebuf),
				  (long) diff);
		}

		if (diff < interval) {
			diff = interval - diff;
			env->var[greylist_seconds_left_loc] = (STKVAL) diff;
			
			debug2(20, "%s still greylisted (for %lu sec.)",
			       email,
			       (unsigned long) diff);
			rc = 1;
		} else if (diff > greylist_format->expire_interval) {
			debug1(20, "greylist record for %s expired",
			       email);
			if (!readonly) {
				memcpy(MU_DATUM_PTR(contents),
				       &now, sizeof now);
				if (mu_dbm_insert(&db, key, contents, 1))
					mu_error("Cannot insert datum in "
						 "greylist database %s",
						 greylist_format->dbname);
			} else
				debug(20, "database opened in readonly mode: "
				      "not updating");
			rc = 1;
		} else {
			debug1(20, "%s finished greylisting period",
			       email);
			rc = 0;
		}
		mu_dbm_datum_free(&contents);
	} else if (!readonly) {
		debug1(20, "greylisting %s", email);
		env->var[greylist_seconds_left_loc] = (STKVAL) interval;
		MU_DATUM_PTR(contents) = (void*)&now;
		MU_DATUM_SIZE(contents) = sizeof now;
		if (mu_dbm_insert(&db, key, contents, 1))
			mu_error("Cannot insert datum in greylist "
				 "database %s",
				 greylist_format->dbname);
		rc = 1;
	} else
		rc = 0;

	mu_dbm_unlock(&db);
	mu_dbm_close(&db);

	push(env, (STKVAL) rc);
}
#endif

void
builtin_setup ()
{
	va_builtin_install("toupper", bi_toupper, rettype_string,
			   1, rettype_string);
	va_builtin_install("tolower", bi_tolower, rettype_string,
			   1, rettype_string);
	va_builtin_install("domainpart", bi_domainpart, rettype_string,
			   1, rettype_string);
	va_builtin_install("localpart", bi_localpart, rettype_string,
			   1, rettype_string);
	va_builtin_install("hostname", bi_hostname, rettype_string,
			   1, rettype_string);
	va_builtin_install("resolve", bi_resolve, rettype_string,
			   1, rettype_string);
	va_builtin_install("hasmx", bi_hasmx, rettype_number,
			   1, rettype_string);
	va_builtin_install("relayed", bi_relayed, rettype_number,
			   1, rettype_string);
	va_builtin_install("listens", bi_listens, rettype_number,
			   1, rettype_string);
	va_builtin_install("numrcpt", bi_numrcpt, rettype_number,
			   1, rettype_number);
	va_builtin_install("validuser", bi_validuser, rettype_number,
			   1, rettype_string);
	va_builtin_install("match_cidr", bi_match_cidr, rettype_number,
			   2, rettype_string, rettype_string);
	va_builtin_install("interval", bi_interval, rettype_number,
                          1, rettype_string);
	va_builtin_install("stdpoll", bi_stdpoll, rettype_number,
			   3, rettype_string, rettype_string, rettype_string);
	va_builtin_install("strictpoll", bi_strictpoll, rettype_number,
			   4, rettype_string, rettype_string, rettype_string,
			   rettype_string);
	va_builtin_install("_pollhost", bi__pollhost, rettype_number,
			   4, rettype_string, rettype_string, rettype_string,
			   rettype_string);
	va_builtin_install("_pollmx", bi__pollmx, rettype_number,
			   4, rettype_string, rettype_string, rettype_string,
			   rettype_string);
#ifdef USE_DBM
	va_builtin_install("dbmap", bi_dbmap, rettype_number,
			   2, rettype_string, rettype_string);	
	va_builtin_install("greylist", bi_greylist, rettype_number,
			   2, rettype_string, rettype_number);
	va_builtin_install("rate", bi_rate, rettype_number,
			   2, rettype_string, rettype_number);
#endif
	greylist_seconds_left_loc =
		builtin_variable_install("greylist_seconds_left",
					 rettype_number)->off;
	rcpt_count_loc =
		builtin_variable_install("rcpt_count", rettype_number)->off;
	ehlo_domain_var =
		builtin_variable_install("ehlo_domain", rettype_string);
	mailfrom_address_var =
		builtin_variable_install("mailfrom_address", rettype_string);
}

void
builtin_post_setup()
{
	if (!smtp_domain)
		mu_get_host_name(&smtp_domain);
	if (!find_declared_variable(ehlo_domain_var->name))
		declare_string_variable(ehlo_domain_var, smtp_domain);
	if (!find_declared_variable(mailfrom_address_var->name))
		declare_string_variable(mailfrom_address_var,
					postmaster_email);
}
				



static int
function_counter(void *sym, void *data)
{
	size_t *p = data;
	++*p;
	return 0;
}

struct entry_point {
	int ishandler;
	prog_counter_t pc;
	union {
		enum smtp_state tag;
		const char *name;
	} v;
};

struct enum_data {
	int i;
	struct entry_point *base;
};

static int
function_lister(void *sym, void *data)
{
	struct function *f = sym;
	struct enum_data *d = data;
	d->base[d->i].ishandler = 0;
	d->base[d->i].v.name = f->name;
	d->base[d->i].pc = f->entry;
	d->i++;
	return 0;
}

static int
comp_pc(const void *a, const void *b)
{
	const struct entry_point *ap = a;
	const struct entry_point *bp = b;

	if (ap->pc < bp->pc)
		return -1;
	else if (ap->pc > bp->pc)
		return 1;
	return 0;
}

void
print_code()
{
	enum smtp_state tag;
	struct entry_point *ep; /* Entry points */
	size_t epcount = 0; /* Number of entry points */
	size_t i;
	struct enum_data d;
	prog_counter_t ip;
	
	/* Count all available entry points: */
	for (tag = smtp_state_first; tag < smtp_state_count; tag++)
		if (entry_point[tag])
			epcount++;
	symbol_enumerate(SYM_FUNC, function_counter, &epcount);
	
	ep = emalloc((epcount+1)*sizeof *ep);

        /* Fill in entry points array */
	i = 0;
	for (tag = smtp_state_first; tag < smtp_state_count; tag++)
		if (entry_point[tag]) {
			ep[i].ishandler = 1;
			ep[i].pc = entry_point[tag];
			ep[i].v.tag = tag;
			i++;
		}

	d.base = ep;
	d.i = i;
	symbol_enumerate(SYM_FUNC, function_lister, &d);

	/* Sort the array */
	qsort(ep, epcount, sizeof ep[0], comp_pc);
	ep[epcount].pc = pc;
	
	/* Actually print the code */

	i = 0;
	while (i < epcount) {
		if (ep[i].ishandler)	
			printf("HANDLER %s\n", state_to_string(ep[i].v.tag));
		else
			printf("FUNCTION %s\n", ep[i].v.name);
		for (ip = ep[i++].pc; ip < ep[i].pc; ip++) { 
			printf("%4lu: ", ip);
			if (prog[ip]) {
				struct optab *op = find_op (prog[ip]);
				if (!op) {
					printf("Unknown function. Aborting dump\n");
					exit (1);
				}
				printf("%s ", op->name);
				if (op->dump)
					op->dump (ip+1, &ip);
				ip += op->length;
			} else
				printf("STOP");
			putchar('\n');
		}
	}
	free(ep);
}
